/**************************************************************************
*   Copyright (C) 2009 by Ian Reinhart Geiser                             *
*   geiseri@kde.org                                                       *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include "taskitembutton.h"

#include <KDE/KGlobalSettings>
#include <KDE/KIconLoader>
#include <taskmanager/taskactions.h>
#include <taskmanager/taskitem.h>
#include <taskmanager/taskgroup.h>


#include <QHBoxLayout>
#include <QMouseEvent>
#include <QApplication>
#include <QFont>
#include <QLabel>
#include <QPainter>
#include <QFontMetrics>
#include <QStylePainter>
#include <QStyleOptionButton>
#include <QMenu>

#include <kdebug.h>

TaskItemButton::TaskItemButton( TaskManager::AbstractGroupableItem *ptr, TaskManager::GroupManager *mgr, QWidget *parent )
    : KPushButton( parent ), m_ptr(ptr), m_mgr(mgr)
{
    setStyleSheet("text-align:left;");
    setSizePolicy(QSizePolicy( QSizePolicy::Expanding, QSizePolicy::Fixed ));

    TaskManager::TaskItem *taskitem = qobject_cast<TaskManager::TaskItem*>(ptr);
    
    connect( ptr, SIGNAL(changed(::TaskManager::TaskChanges)), this, SLOT(updateTask()));
    connect( taskitem->task().data(), SIGNAL(activated()), this, SLOT(updateTask()));
    connect( taskitem->task().data(), SIGNAL(deactivated()), this, SLOT(updateTask()));

    setFont( KGlobalSettings::taskbarFont() );

    updateTask();
}

void TaskItemButton::setActive( bool act )
{
    setDown(act);
}

void TaskItemButton::mousePressEvent ( QMouseEvent * event )
{
    if( event->button() == Qt::RightButton )
    {

        TaskManager::BasicMenu *menu = 0;
        if( !m_ptr->isGroupItem() )
            menu = new TaskManager::BasicMenu(this, qobject_cast<TaskManager::TaskItem *>(m_ptr), m_mgr );
        else
            menu = new TaskManager::BasicMenu(this, qobject_cast<TaskManager::TaskGroup *>(m_ptr), m_mgr );
        menu->exec( mapToGlobal(event->pos()));
        delete menu;

    }
    else if( event->button() == Qt::LeftButton )
    {
        if( m_ptr->isActive() )
        {
            if( !m_ptr->isGroupItem() )
            {
                TaskManager::TaskItem *item = qobject_cast<TaskManager::TaskItem *>(m_ptr);
                item->task()->setIconified (true);
            }
        }
        else
        {
            if( !m_ptr->isGroupItem() )
            {
                TaskManager::TaskItem *item = qobject_cast<TaskManager::TaskItem *>(m_ptr);
                item->task()->activateRaiseOrIconify();
            }
	    setActive( m_ptr->isActive() );
        }
    }
}

QSize TaskItemButton::sizeHint() const
{
    int maxWidth = KGlobalSettings::desktopGeometry( this ).width() * 3 / 4;
    QFontMetrics fm(fontMetrics());
    int textWidth = fm.width( m_ptr->name() );
    if (textWidth > maxWidth)
    {
         textWidth = maxWidth;
    }
    return QSize(textWidth, KPushButton::sizeHint().height());
}

QSize TaskItemButton::minimumSizeHint() const
{
    QSize sz = KPushButton::minimumSizeHint();
    sz.setWidth(-1);
    return sz;
}

void TaskItemButton::updateTask()
{
    kDebug() << "Update text" << m_ptr->name();
    bool gotStaticIcon = false;
    QFontMetrics fm( font() );
    QIcon icn = m_ptr->icon();
    setIcon( icn );
    setText( squeezeTextToButton(m_ptr->name(), height() ) );

    setActive( m_ptr->isActive() );

    if( m_ptr->demandsAttention () )
    {
        kDebug() << m_ptr->name() << "demands attention";
    }
    
    update();
}

void TaskItemButton::resizeEvent( QResizeEvent *event )
{
    updateTask();
}

QString TaskItemButton::squeezeTextToButton( const QString &text, int iconSz )
{
    int padding = 8;
    QFontMetrics fm(fontMetrics());
    int buttonWidth = width() - ( iconSz + padding);
    QString squeezedText = text;
    int textWidth = fm.width(text);
    if ( textWidth > buttonWidth)
    {
         squeezedText = fm.elidedText(text, Qt::ElideMiddle, buttonWidth );
    }
    return squeezedText;
}
