/*
 * KDE Composite Manager, 
 * Copyright (C) 2003-2005 Fredrik Höglund <fredrik@kde.org>
 *
 * Based on xcompmgr,
 * Copyright (C) 2003 Keith Packard <keithp@keithp.com>
 *
 * Internal structure and naming style inspired by KWin,
 * Copyright (C) 1999, 2000 Matthias Ettrich <ettrich@kde.org>
 * Copyright (C) 2003 Lubos Lunak <l.lunak@kde.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Fredrik Höglund not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Fredrik Höglund makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * FREDRIK HÖGLUND DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL FREDRIK HÖGLUND BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */


#include "shadowpainter.h"
#include "utils.h"

#include <cmath>
#include <cstdlib>

#include "settings.h"

//#include <iostream>

ShadowPainter *ShadowPainter::sSelf = NULL;
extern Display *dpy;


TileSet::TileSet( double radius )
	: mRadius( radius )
{
	// Create the gaussian kernel
	// -------------------------------------------------------------
	mSize = int( std::ceil( mRadius * 3 ) + 1 ) & ~1;

	int center  = mSize / 2;
	int len     = mSize * mSize;

	double *kernel = new double[ len ];

	double total = 0;
	int ki = 0;
	for ( int y = 0; y < mSize; y++ ) {
		for ( int x = 0; x < mSize; x++ )
		{
			double val = gaussian( x - center, y - center );
			kernel[ki++] = val;
			total += val;
		}
	}

	for ( int i = 0; i < len; i++ )
		kernel[i] /= total;


	// Create the shadow image
	// ------------------------------------------------------------
	int width = int( mSize * 3 );
	int height = width;

	// Allocate image data
	unsigned char *data = (unsigned char *)malloc( width * height );

	// Center fill
	unsigned char pixel = sum( kernel, center, center );
	memset( data, pixel, width * height );

	// Corners
	for ( int y = 0; y < mSize; y++ ) {
		for ( int x = 0; x < mSize; x++ )
		{
			unsigned char pixel = sum( kernel, x - center, y - center );
			data[y * width + x] = pixel;
			data[(height - y - 1) * width + x] = pixel;
			data[(height - y - 1) * width + (width - x - 1)] = pixel;
			data[y * width + (width - x - 1)] = pixel;
		}
	}

	// Top / bottom
	for ( int y = 0; y < mSize; y++ ) {
		unsigned char pixel = sum( kernel, center, y - center );
		int i = y * width + mSize;
		int j = (height - y - 1) * width + mSize;
		for ( int x = mSize; x < width - mSize; x++ ) {
			data[i++] = pixel;
			data[j++] = pixel;
		}
	}

	// Left / Right
	for ( int x = 0; x < mSize; x++ ) {
		unsigned char pixel = sum( kernel, x - center, center );
		for ( int y = mSize; y < height - mSize; y++ ) {
			data[y * width + x] = pixel;
			data[y * width + (width - x - 1)] = pixel;
		}
	}

	delete [] kernel;

	
	// Convert the image to a pixmap
	// ------------------------------------------------------------
	XImage *image = XCreateImage( dpy, DefaultVisual(dpy, DefaultScreen(dpy)),
								  8, ZPixmap, 0, (char*)data, width, height, 8, width );
	Pixmap pixmap = XCreatePixmap( dpy, DefaultRootWindow(dpy), width, height, 8 );
	GC gc = XCreateGC( dpy, pixmap, 0, 0 );
	XPutImage( dpy, pixmap, gc, image, 0, 0, 0, 0, width, height );
	XFreeGC( dpy, gc );
	XDestroyImage( image );	

	XRenderPictFormat *format = XRenderFindStandardFormat( dpy, PictStandardA8 );
	Picture picture = XRenderCreatePicture( dpy, pixmap, format, 0, 0 );
	XFreePixmap( dpy, pixmap );


	// Cut the mTiles from the pixmap
	// ------------------------------------------------------------
	mTiles[TopLeft]     = createPicture( picture, 0, 0, mSize, mSize );
	mTiles[Top]         = createPicture( picture, mSize, 0, mSize, mSize, Repeating );
	mTiles[TopRight]    = createPicture( picture, width - mSize, 0, mSize, mSize );

	mTiles[Left]        = createPicture( picture, 0, mSize, mSize, mSize, Repeating );
	mTiles[Center]      = createPicture( picture, mSize, mSize, mSize, mSize, Repeating );
	mTiles[Right]       = createPicture( picture, width - mSize, mSize, mSize, mSize, Repeating );

	mTiles[BottomLeft]  = createPicture( picture, 0, height - mSize, mSize, mSize);
	mTiles[Bottom]      = createPicture( picture, mSize, height - mSize, mSize, mSize, Repeating );
	mTiles[BottomRight] = createPicture( picture, width - mSize, height - mSize, mSize, mSize );

	XRenderFreePicture( dpy, picture );
}


TileSet::~TileSet()
{
	for ( int i = 0; i < NTiles; i++ )
		releasePicture( mTiles[i] );
}


double TileSet::gaussian( double x, double y )
{
	return ( (1 / sqrt (2 * M_PI * mRadius))
			* std::exp( -(x*x + y*y) / (2 * mRadius * mRadius)) );
}


unsigned char TileSet::sum( double *kernel, int x, int y )
{
	int center = mSize / 2;

	int x_start = qMax( center - x, 0 );
	int x_end   = qMin( mSize + center - x, mSize );

	int y_start = qMax( center - y, 0 );
	int y_end   = qMin( mSize + center - y, mSize );

	double val = 0;
	for ( int y = y_start; y < y_end; y++ ) {
		int ki = y * mSize + x_start;
		for ( int x = x_start; x < x_end; x++ )
			val += kernel[ ki++ ];
	}

	if ( val > 1 )
		val = 1;

	return uchar( val * 255.0 );
}


Picture TileSet::createPicture( Picture source, int x, int y, int width, int height, uint flags )
{
	Picture picture = ::createPicture( width, height, PictStandardA8, flags );
	XRenderComposite( dpy, PictOpSrc, source, None, picture, x, y, 0, 0, 0, 0, width, height );
	return picture;
}



// ----------------------------------------------------------------------------



ShadowPainter *ShadowPainter::self()
{
	if ( !sSelf )
		sSelf = new ShadowPainter;
	return sSelf;
}


ShadowPainter::ShadowPainter()
	: mSource( None ), mLastOpacity( -1 )
{
	mCache = new QCache<int, TileSet>( 16 );
}


ShadowPainter::~ShadowPainter()
{
	XRenderFreePicture( dpy, mSource );
	delete mCache;
}


void ShadowPainter::setupForOpacity( double opacity )
{
	if ( !mSource ) {
		XRenderPictFormat *format = XRenderFindStandardFormat( dpy, PictStandardARGB32 );
		XRenderPictureAttributes pa;
		pa.repeat = true; 

		Pixmap pixmap = XCreatePixmap( dpy, DefaultRootWindow(dpy), 1, 1, 32 );
		mSource = XRenderCreatePicture( dpy, pixmap, format, CPRepeat, &pa );
		XFreePixmap( dpy, pixmap );
	} 

	if ( int(opacity * 0xff) != int(mLastOpacity * 0xff) ) {
		XRenderColor col = { 0, 0, 0, int( 0xffff * opacity ) };
		XRenderFillRectangle( dpy, PictOpSrc, mSource, &col, 0, 0, 1, 1 );
	}

	mLastOpacity = opacity;
}


TileSet *ShadowPainter::findTileSet( double radius )
{
	int r = int( std::floor( radius ) );	
	TileSet *tiles = mCache->object( r );

	if ( !tiles ) {
		//std::cout << "creating tileset for radius: " << radius << " (" << r << ")\n";
		tiles = new TileSet( radius );
		mCache->insert( r, tiles );
	}

	return tiles;
}

QRect ShadowPainter::geometry( const QRect &r, double radius )
{
	int size = findTileSet( radius )->size();
	return QRect( r.x() - size / 2, r.y() - size / 2, r.width() + size, r.height() + size );
}


void ShadowPainter::draw( Picture dest, const QRect &r, double radius, double windowOpacity )
{
	setupForOpacity( Settings::shadowOpacity() * windowOpacity );

	TileSet &tiles = *findTileSet( radius );
	int size = tiles.size();

	int lw = (r.width() > size * 2) ? size : r.width() / 2;
	int th = (r.height() > size * 2) ? size : r.height() / 2;
	int rw = qMin( r.width() - lw, size );
	int bh = qMin( r.height() - th, size );
	int sx = size - rw; // Only for right column
	int sy = size - bh; // Only for bottom row


	// Top row
	// ------------------------------------------------------------------------
	XRenderComposite( dpy, PictOpOver, mSource, tiles[TopLeft], dest,
					  0, 0, 0, 0, r.x(), r.y(), lw, th );

	if ( r.width() > size * 2 )
		XRenderComposite( dpy, PictOpOver, mSource, tiles[Top], dest,
						  0, 0, 0, 0, r.x() + size, r.y(), r.width() - size * 2, th );

	XRenderComposite( dpy, PictOpOver, mSource, tiles[TopRight], dest,
					  sx, 0, 0, 0, r.right() - rw + 1, r.y(), rw, th );

	// Center row
	// ------------------------------------------------------------------------
	if ( r.height() > size * 2 ) {
		XRenderComposite( dpy, PictOpOver, mSource, tiles[Left], dest,
						  0, 0, 0, 0, r.x(), r.y() + size, lw, r.height() - size * 2 );

		if ( r.width() > size * 2 )
			XRenderComposite( dpy, PictOpOver, mSource, tiles[Center], dest,
							  0, 0, 0, 0, r.x() + size, r.y() + size,
							  r.width() - size * 2, r.height() - size * 2 );

		XRenderComposite( dpy, PictOpOver, mSource, tiles[Right], dest,
						  sx,0, 0,0, r.right() - rw + 1, r.y() + size,
						  rw, r.height() - size * 2 );
	}

	// Bottom row
	// ------------------------------------------------------------------------
	XRenderComposite( dpy, PictOpOver, mSource, tiles[BottomLeft], dest,
					  0, sy, 0, 0, r.x(), r.bottom() - bh + 1, lw, bh );

	if ( r.width() > size * 2 )
		XRenderComposite( dpy, PictOpOver, mSource, tiles[Bottom], dest,
						  0, sy, 0, 0, r.x() + size, r.bottom() - bh + 1,
						  r.width() - size * 2, bh );

	XRenderComposite( dpy, PictOpOver, mSource, tiles[BottomRight], dest,
					  sx, sy, 0, 0, r.right() - rw + 1, r.bottom() - bh + 1, rw, bh );
}


void ShadowPainter::draw( Picture dest, const QRect &r, Picture shadow, double windowOpacity )
{
	setupForOpacity( Settings::shadowOpacity() * windowOpacity );
	XRenderComposite( dpy, PictOpOver, mSource, shadow, dest, 0, 0, 0, 0,
					  r.x(), r.y(), r.width(), r.height() );
}


Picture ShadowPainter::createShadow( const QSize &s, double radius )
{
	int width = s.width(), height = s.height();
	Picture shadow = createPicture( width, height, PictStandardA8 );

	TileSet &tiles = *findTileSet( radius );
	int size = tiles.size();

	int lw = (width > size * 2) ? size : width / 2;
	int th = (height > size * 2) ? size : height / 2;
	int rw = qMin( width - lw, size );
	int bh = qMin( height - th, size );
	int sx = size - rw; // Only for right column
	int sy = size - bh; // Only for bottom row

	// Top row
	// ------------------------------------------------------------------------
	XRenderComposite( dpy, PictOpSrc, tiles[TopLeft], None, shadow,
					  0, 0, 0, 0, 0, 0, lw, th );

	if ( width > size * 2 )
		XRenderComposite( dpy, PictOpSrc, tiles[Top], None, shadow,
						  0, 0, 0, 0, size, 0, width - size * 2, th );

	XRenderComposite( dpy, PictOpSrc, tiles[TopRight], None, shadow,
					  sx, 0, 0, 0, width - rw, 0, rw, th );

	// Center row
	// ------------------------------------------------------------------------
	if ( height > size * 2 ) {
		XRenderComposite( dpy, PictOpSrc, tiles[Left], None, shadow,
						  0, 0, 0, 0, 0, size, lw, height - size * 2 );

		if ( width > size * 2 )
			XRenderComposite( dpy, PictOpSrc, tiles[Center], None, shadow,
							  0, 0, 0, 0, size, size, width - size * 2, height - size * 2 );

		XRenderComposite( dpy, PictOpSrc, tiles[Right], None, shadow,
						  sx, 0, 0, 0, width - rw, size, rw, height - size * 2 );
	}

	// Bottom row
	// ------------------------------------------------------------------------
	XRenderComposite( dpy, PictOpSrc, tiles[BottomLeft], None, shadow,
					  0, sy, 0, 0, 0, height - bh, lw, bh );

	if ( width > size * 2 )
		XRenderComposite( dpy, PictOpSrc, tiles[Bottom], None, shadow,
						  0, sy, 0, 0, size, height - bh, width - size * 2, bh );

	XRenderComposite( dpy, PictOpSrc, tiles[BottomRight], None, shadow,
					  sx, sy, 0, 0, width - rw, height - bh, rw, bh );

	return shadow;
}




