/*
    This file is part of KOrganizer.
    Copyright (c) 2002 Cornelius Schumacher <schumacher@kde.org>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#include "engine.h"

#include "appstuff.h"
#include "prefs.h"

#include <qcstring.h>
#include <qdom.h>
#include <qfileinfo.h>

#include <kapplication.h>
#include <kdebug.h>
#include <kio/job.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kstandarddirs.h>
#include <kmessagebox.h>

using namespace KNS;

struct Engine::Private
{
    bool mIgnoreInstallResult;
    AppStuff *mNewStuff;
};

Engine::Engine( AppStuff *newStuff, const QString &type,
                QWidget *parentWidget ) :
  mParentWidget( parentWidget ),
  d(new Private), mType( type )
{
  d->mNewStuff = newStuff;
  mProviderLoader = new ProviderLoader( mParentWidget );
  connect( mProviderLoader,
           SIGNAL( providersLoaded( Provider::List * ) ),
           SLOT( getMetaInformation( Provider::List * ) ) );
}

Engine::~Engine()
{
  clearEntries();

  delete d;
  delete mProviderLoader;
}

void Engine::clearEntries()
{
  Entry::List::ConstIterator it;
  Entry::List::ConstIterator end(mEntries.end() );
  for( it = mEntries.begin(); it != end; ++it ) {
    delete *it;
  }
  mEntries.clear();
}

void Engine::retrieveListing()
{
  kdDebug(5850) << "Engine::retrieveListing()" << endl;

  mProviderLoader->load( mType, mProviderList );
}

void Engine::getMetaInformation( Provider::List *providers )
{
  kdDebug() << "Engine::getMetaInformation()" << endl;

  if ( providers->isEmpty() ) {
    QString msg = mProviderLoader->lastError();
    if ( msg.isEmpty() ) msg = i18n("No providers found.");
    KMessageBox::sorry( mParentWidget, msg,
      i18n("Error retrieving provider list") );
    emit retrievalFailed( i18n("No Providers") );
    return;
  }

  clearEntries();

  mNewStuffJobData.clear();

  Provider *p;
  for ( p = providers->first(); p; p = providers->next() ) {
    kdDebug() << "Provider: " << p->downloadUrl() << endl;
    if ( p->downloadUrl().isEmpty() ) continue;

    KURL url = createStuffUrl( p->downloadUrl() );
    kdDebug() << "Download URL: " << url << endl;

    KIO::TransferJob *job = KIO::get( url );
    connect( job, SIGNAL( result( KIO::Job * ) ),
             SLOT( slotNewStuffJobResult( KIO::Job * ) ) );
    connect( job, SIGNAL( data( KIO::Job *, const QByteArray & ) ),
             SLOT( slotNewStuffJobData( KIO::Job *, const QByteArray & ) ) );

    mNewStuffJobData.insert( job, "" );
    mProviderJobs[ job ] = p;
  }
}

KURL Engine::createStuffUrl( const KURL &provider )
{
  KURL url = provider;

  QString path = mType + "/" + Prefs::distributionName() + "/" +
    Prefs::distributionVersion() + "/" + Prefs::systemArchitecture() +
    "/knewstuff.xml";

  url.setPath( url.path( 1 ) + path );
  return url;
}

void Engine::slotNewStuffJobData( KIO::Job *job, const QByteArray &data )
{
  if ( data.isEmpty() ) return;

//  kdDebug(5850) << "Engine:slotNewStuffJobData()" << endl;

  QCString str( data, data.size() + 1 );

  mNewStuffJobData[ job ].append( QString::fromUtf8( str ) );
}

void Engine::slotNewStuffJobResult( KIO::Job *job )
{
  kdDebug() << "Engine::slotNewStuffResult() done" << endl;

  if ( job->error() ) {
    kdDebug(5850) << "Error downloading new stuff descriptions." << endl;
    job->showErrorDialog( mParentWidget );
  } else {
    QString knewstuffDoc = mNewStuffJobData[ job ];

#if 0
    kdDebug(5850) << "---START---" << endl << knewstuffDoc << "---END---" << endl;
#endif

    QDomDocument doc;
    if ( !doc.setContent( knewstuffDoc ) ) {
      kdDebug(5850) << "Error parsing knewstuff.xml." << endl;
      return;
    } else {
      QDomElement knewstuff = doc.documentElement();

      if ( knewstuff.isNull() ) {
        kdDebug(5850) << "No document in knewstuffproviders.xml." << endl;
      } else {
        QDomNode p;
        for ( p = knewstuff.firstChild(); !p.isNull(); p = p.nextSibling() ) {
          QDomElement stuff = p.toElement();
          if ( stuff.tagName() != "stuff" ) continue;
          if ( stuff.attribute("type", mType) != mType ) continue;

          Entry *entry = new Entry( stuff );
          mEntries.append( entry );

          emit entryRetrieved( entry );

#if 0
          kdDebug(5850) << "KNEWSTUFF: " << entry->name() << endl;

          kdDebug(5850) << "  SUMMARY: " << entry->summary() << endl;
          kdDebug(5850) << "  VERSION: " << entry->version() << endl;
          kdDebug(5850) << "  RELEASEDATE: " << entry->releaseDate().toString() << endl;
          kdDebug(5850) << "  RATING: " << entry->rating() << endl;

          kdDebug(5850) << "  LANGS: " << entry->langs().join(", ") << endl;
#endif
        }
      }
    }
  }

  mNewStuffJobData.remove( job );
  mProviderJobs.remove( job );

  if ( mNewStuffJobData.count() == 0 ) {
    emit retrievalFinished();
  }

  kdDebug() << "Engine::slotNewStuffResult() done" << endl;
}

void Engine::download( Entry *entry )
{
  kdDebug(5850) << "Engine::download()" << entry->name() << endl;

  KURL source = entry->payload();
  mDownloadDestination = d->mNewStuff->downloadDestination( entry );

  if ( mDownloadDestination.isEmpty() ) {
    kdDebug(5850) << "Empty downloadDestination. Cancelling download." << endl;
    return;
  }

  KURL destination = KURL( mDownloadDestination );

  kdDebug(5850) << "  SOURCE: " << source.url() << endl;
  kdDebug(5850) << "  DESTINATION: " << destination.url() << endl;

  KIO::FileCopyJob *job = KIO::file_copy( source, destination, -1, true );
  connect( job, SIGNAL( result( KIO::Job * ) ),
           SLOT( slotDownloadJobResult( KIO::Job * ) ) );
}

void Engine::slotDownloadJobResult( KIO::Job *job )
{
  if ( job->error() ) {
    kdDebug(5850) << "Error downloading new stuff payload." << endl;
//    job->showErrorDialog( mParentWidget );
    emit downloadFailed( job->errorString() );
    return;
  }

  emit downloadFinished( mDownloadDestination );

#if 0
  if ( d->mNewStuff->install( mDownloadDestination ) ) {
    if ( !d->mIgnoreInstallResult ) {
      KMessageBox::information( mParentWidget,
                                i18n("Successfully installed hot new stuff.") );
    }
  } else
    if ( !d->mIgnoreInstallResult ){
      KMessageBox::error( mParentWidget,
                          i18n("Failed to install hot new stuff.") );
  }
#endif
}

void Engine::ignoreInstallResult(bool ignore)
{
  d->mIgnoreInstallResult = ignore;
}

KNS::Entry::List Engine::entriesForName( const QString &name )
{
  Entry::List result;

  Entry::List::ConstIterator it;
  Entry::List::ConstIterator end( mEntries.end() );
  for( it = mEntries.begin(); it != end; ++it ) {
    if ( (*it)->name() == name ) result.append( *it );
  }

  return result;
}

#include "engine.moc"

