/*
    This file is part of the Nepomuk KDE project.
    Copyright (C) 2010  Vishesh Handa <handa.vish@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "identificationthread.h"
#include "mergerequest.h"
#include "changelog.h"
#include "nie.h"

#include <QtCore/QMutexLocker>
#include <QtCore/QHashIterator>
#include <QtCore/QFile>

#include <Soprano/Model>
#include <Soprano/QueryResultIterator>
#include <Soprano/Vocabulary/RDF>

#include <Nepomuk/ResourceManager>
#include <Nepomuk/Resource>
#include <Nepomuk/Variant>

#include <KDebug>

Nepomuk::IdentificationThread::IdentificationThread(QObject* parent): QThread(parent)
{
    start();
}


Nepomuk::IdentificationThread::~IdentificationThread()
{
    //stop();
    //quit();
}


int Nepomuk::IdentificationThread::process(const Nepomuk::Sync::SyncFile& sf)
{
    m_queueMutex.lock();

    Sync::IdentificationRequest* ip = new Sync::IdentificationRequest( sf, ResourceManager::instance()->mainModel(), this );
    int id = ip->id();
    m_queue.enqueue( ip );

    m_queueMutex.unlock();
    m_queueWaiter.wakeAll();

    return id;
}


void Nepomuk::IdentificationThread::stop()
{
    m_stopped = true;
    m_queueWaiter.wakeAll();
}



void Nepomuk::IdentificationThread::run()
{
    m_stopped = false;

    while( !m_stopped ) {

        // lock for initial iteration
        m_queueMutex.lock();

        while( !m_queue.isEmpty() ) {

            Sync::IdentificationRequest* ip = m_queue.dequeue();

            // unlock after queue utilization
            m_queueMutex.unlock();

            ip->load();

            connect( ip, SIGNAL(identified(int,QUrl,QUrl)),
                     this, SIGNAL(identified(int,QUrl,QUrl)) );
            connect( ip, SIGNAL(notIdentified(int,QList<Soprano::Statement>)),
                     this, SIGNAL(notIdentified(int,QList<Soprano::Statement>)) );

            ip->identifyAll();

            Sync::ChangeLog log = ip->convert();
            if( !log.empty() ) {
                kDebug() << "sending ChangeLog";
                emit processed( log );
            }

            if( !ip->done() ) {
                QMutexLocker lock( &m_processMutex );
                m_processes[ ip->id() ] = ip;
            }

            disconnect( ip, SIGNAL(identified(int,QUrl,QUrl)),
                     this, SIGNAL(identified(int,QUrl,QUrl)) );
            disconnect( ip, SIGNAL(notIdentified(int,QList<Soprano::Statement>)),
                     this, SIGNAL(notIdentified(int,QList<Soprano::Statement>)) );

            if( ip->done() ) {
                m_processes.remove( ip->id() );
                delete ip;
            }

            m_queueMutex.lock();
        }

        // wait for more input
        kDebug() << "Waiting...";
        m_queueWaiter.wait( &m_queueMutex );
        m_queueMutex.unlock();
        kDebug() << "Woke up.";
    }
}


bool Nepomuk::IdentificationThread::identify(int id, const QUrl& oldUri, const QUrl& newUri)
{
    kDebug() << newUri;
    // Lock the mutex and all
    QMutexLocker lock ( &m_processMutex );

    QHash<int, Sync::IdentificationRequest*>::iterator it = m_processes.find( id );
    if( it == m_processes.end() )
        return false;

    Sync::IdentificationRequest* ip = *it;

    if ( oldUri.scheme() != QLatin1String("nepomuk") )
        return false;

    if( newUri.scheme() == QLatin1String("nepomuk") ) {
        ip->provideResourceUri( oldUri, newUri);
    }
    else if( newUri.scheme() == "file" ) {
        ip->provideNieUrl( oldUri, newUri );
    }

    m_queueMutex.lock();
    m_queue.enqueue( ip );
    m_queueMutex.unlock();
    m_queueWaiter.wakeAll();

    return true;
}


bool Nepomuk::IdentificationThread::ignore(int id, const QUrl& url, bool ignoreSub)
{
    // Lock the mutex and all
    QMutexLocker lock ( &m_processMutex );

    QHash<int, Sync::IdentificationRequest*>::iterator it = m_processes.find( id );
    if( it == m_processes.end() )
        return false;

    Sync::IdentificationRequest* ip = *it;
    return ip->ignore( url, ignoreSub );
}

#include "identificationthread.moc"
