/*
    This file is part of the Nepomuk KDE project.
    Copyright (C) 2010  Vishesh Handa <handa.vish@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "backupstorage.h"
#include "tools.h"
#include "syncfile.h"
#include "changelog.h"

#include <QtCore/QUrl>
#include <QtCore/QDateTime>
#include <QtCore/QFile>

#include <Soprano/QueryResultIterator>
#include <Soprano/Model>

#include <Nepomuk/ResourceManager>

#include <KStandardDirs>
#include <KTempDir>
#include <KTemporaryFile>
#include <KZip>
#include <KDebug>
#include <KConfig>
#include <KConfigGroup>


Nepomuk::BackupStorage::BackupStorage(const QString& url)
{
    if( url.isEmpty() )
        m_url = KStandardDirs::locateLocal( "data", "nepomuk/backupsync/backup" );
    else
        m_url = url;
}


Nepomuk::BackupStorage::~BackupStorage()
{
}


void Nepomuk::BackupStorage::add(const Nepomuk::ChangeLog& log)
{
    KTempDir dir;
    QUrl url( dir.name() + "syncFile" );
    SyncFile sf( log );
    sf.save( url );

    add( url );
}


void Nepomuk::BackupStorage::add(const QUrl& syncFileUrl)
{
    KZip zipFile( m_url );
    zipFile.open( QIODevice::ReadWrite );

    QString fileName = QDateTime::currentDateTime()/*.toUTC()*/.toString( ChangeLog::dateTimeFormat() );
    zipFile.addLocalFile( syncFileUrl.toLocalFile(), fileName );
}


void Nepomuk::BackupStorage::create()
{
    //FIXME: The time stamps should be changed to UTC if UTC time is being used.
    add( ChangeLog( getBackupStatements() ) );
}


bool Nepomuk::BackupStorage::exists() const
{
    KZip zipFile( m_url );
    zipFile.open( QIODevice::ReadOnly );

    return !zipFile.directory()->entries().empty();
}


QList<Nepomuk::SyncFile> Nepomuk::BackupStorage::syncFiles() const
{
    KTempDir tempDir;

    KZip zipFile( m_url );
    zipFile.open( QIODevice::ReadOnly );
    zipFile.directory()->copyTo( tempDir.name() );
    QStringList entries = zipFile.directory()->entries();
    zipFile.close();

    entries.sort();

    QList<SyncFile> list;
    foreach( const QString & name, entries ) {
        QUrl url( tempDir.name() + name );
        SyncFile sf( url );

        list.append( sf );
    }
    return list;
}


QDateTime Nepomuk::BackupStorage::lastBackup() const
{
    KZip zipFile( m_url );
    zipFile.open( QIODevice::ReadOnly );
    QStringList entries = zipFile.directory()->entries();
    zipFile.close();

    QDateTime maxDt;
    foreach( const QString & name, entries ) {
        kDebug() << name;
        QDateTime dateTime = QDateTime::fromString( name, ChangeLog::dateTimeFormat() );
        if( maxDt < dateTime )
            maxDt = dateTime;
    }

    return maxDt;
}


QList<QDateTime> Nepomuk::BackupStorage::syncFileDates() const
{
    KZip zipFile( m_url );
    zipFile.open( QIODevice::ReadOnly );
    QStringList entries = zipFile.directory()->entries();
    zipFile.close();

    entries.sort();

    QList<QDateTime> list;
    foreach( const QString & name, entries ) {
        list.append( QDateTime::fromString( name, ChangeLog::dateTimeFormat() ) );
    }
    return list;
}


Nepomuk::SyncFile Nepomuk::BackupStorage::syncFile(const QDateTime& dateTime) const
{
    KZip zipFile( m_url );
    zipFile.open( QIODevice::ReadOnly );

    const KArchiveFile * file = static_cast<const KArchiveFile *>( zipFile.directory()->entry( dateTime.toString( ChangeLog::dateTimeFormat() ) )  );

    KTemporaryFile tempFile;
    tempFile.open();
    tempFile.close();

    file->copyTo( tempFile.fileName() );
    return SyncFile( QUrl( tempFile.fileName() ) );
}


void Nepomuk::BackupStorage::remove()
{
    kDebug() << "Removing previous backup";
    QFile::remove( m_url );
}
