/*
 *   Copyright (C) 2010 Ivan Cukic <ivan.cukic(at)kde.org>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2,
 *   or (at your option) any later version, as published by the Free
 *   Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "ActivityDirector.h"
#include "ActivityDirector_p.h"

#include <QUuid>
#include <QDBusConnection>

#include <KConfig>
#include <KConfigGroup>
#include <KCrash>
#include <KUrl>
#include <KDebug>

#include <KWindowInfo>
#include <KWindowSystem>

#include <Nepomuk/ResourceManager>
#include <Nepomuk/Resource>

#include "activitydirectoradaptor.h"
#include "EventProcessor.h"

#include "config-features.h"

#ifdef HAVE_NEPOMUK
    #define NEPOMUK_RUNNING (Nepomuk::ResourceManager::instance()->initialized())
#else
    #define NEPOMUK_RUNNING false
#endif

#define ActivityDirectorServicePath "org.kde.ActivityDirector"
#define ACTIVITIES_PROTOCOL "activities://"

// Private

ActivityDirectorPrivate::ActivityDirectorPrivate(ActivityDirector * parent)
    : config("activitymanagerrc"), q(parent)
{
    // Initializing config
    connect(&configSyncTimer, SIGNAL(timeout()),
             this, SLOT(configSync()));

    configSyncTimer.setSingleShot(true);
    configSyncTimer.setInterval(2 * 60 * 1000);

    foreach(const QString & activity, activitiesConfig().keyList()) {
        activities[activity] = ActivityDirector::Stopped;
    }

    foreach(const QString & activity, mainConfig().readEntry("runningActivities", activities.keys())) {
        if (activities.contains(activity)) {
            activities[activity] = ActivityDirector::Running;
        }
    }

    currentActivity = mainConfig().readEntry("currentActivity", QString());

    connect(KWindowSystem::self(), SIGNAL(windowClosed(WId)),
            this, SLOT(windowClosed(WId)));
}

ActivityDirectorPrivate::~ActivityDirectorPrivate()
{
    configSync();
}

void ActivityDirectorPrivate::windowClosed(WId windowId)
{
    if (!resourcesForWindow.contains(windowId)) {
        return;
    }

    foreach(const KUrl & uri, resourcesForWindow[windowId]) {
        q->NotifyResourceClosed(windowId, uri.url());
    }
}

void ActivityDirectorPrivate::setActivityState(const QString & id, ActivityDirector::State state)
{
    if (activities[id] == state) return;

    /**
     * Treating 'Starting' as 'Running', and 'Stopping' as 'Stopped'
     * as far as the config file is concerned
     */
    bool configNeedsUpdating = ((activities[id] & 4) != (state & 4));

    activities[id] == state;

    switch (state) {
        case ActivityDirector::Running:
            emit q->ActivityStarted(id);
            break;

        case ActivityDirector::Stopped:
            emit q->ActivityStopped(id);
            break;

        default:
            break;
    }

    emit q->ActivityStateChanged(id, state);

    if (configNeedsUpdating) {
        mainConfig().writeEntry("runningActivities",
                activities.keys(ActivityDirector::Running) +
                activities.keys(ActivityDirector::Starting));
        scheduleConfigSync();
    }
}

KConfigGroup ActivityDirectorPrivate::activitiesConfig()
{
    return KConfigGroup(&config, "activities");
}

KConfigGroup ActivityDirectorPrivate::mainConfig()
{
    return KConfigGroup(&config, "main");
}

void ActivityDirectorPrivate::ensureCurrentActivityIsRunning()
{
    QStringList runningActivities = q->ListActivities(ActivityDirector::Running);

    if (!runningActivities.contains(currentActivity)) {
        if (runningActivities.size() > 0) {
            currentActivity = runningActivities.first();
        } else {
            currentActivity.clear();
        }
    }
}

bool ActivityDirectorPrivate::setCurrentActivity(const QString & id)
{
    if (id.isEmpty()) {
        currentActivity.clear();

    } else {
        if (!activities.contains(id)) {
            return false;
        }

        q->StartActivity(id);

        currentActivity = id;
        mainConfig().writeEntry("currentActivity", id);

        scheduleConfigSync();
    }

    emit q->CurrentActivityChanged(id);
    return true;
}

QString ActivityDirectorPrivate::activityName(const QString & id)
{
    return activitiesConfig().readEntry(id, QString());
}

void ActivityDirectorPrivate::scheduleConfigSync()
{
    if (!configSyncTimer.isActive()) {
        configSyncTimer.start();
    }
}

void ActivityDirectorPrivate::configSync()
{
    configSyncTimer.stop();
    config.sync();
}

#ifdef HAVE_NEPOMUK
Nepomuk::Resource ActivityDirectorPrivate::activityResource(const QString & id) const
{
    return Nepomuk::Resource(KUrl(ACTIVITIES_PROTOCOL + id));
}
#endif // HAVE_NEPOMUK


// Main

ActivityDirector::ActivityDirector()
    : d(new ActivityDirectorPrivate(this))
{
    // KCrash::setFlags(KCrash::AutoRestart);

    QDBusConnection dbus = QDBusConnection::sessionBus();
    if (!dbus.interface()->registerService(ActivityDirectorServicePath)) {
        kError() << "already running, quitting";

        QCoreApplication::quit();
        return;
    }

    new ActivityDirectorAdaptor(this);
    dbus.registerObject("/ActivityDirector", this);

    // Initializing Nepomuk
#ifdef HAVE_NEPOMUK
    Nepomuk::ResourceManager::instance()->init();
#endif // HAVE_NEPOMUK

    // TODO: Sync activities in nepomuk with currently existing ones

    // ensureCurrentActivityIsRunning();

    EventProcessor::self()->start();
}

ActivityDirector::~ActivityDirector()
{
    delete d;
}

void ActivityDirector::Start()
{
    // doing absolutely nothing
}

void ActivityDirector::Stop()
{
    QCoreApplication::quit();
}


// workspace activities control

QString ActivityDirector::CurrentActivity() const
{
    return d->currentActivity;
}

bool ActivityDirector::SetCurrentActivity(const QString & id)
{
    if (id.isEmpty()) {
        return false;
    }

    return d->setCurrentActivity(id);
}

QString ActivityDirector::AddActivity(const QString & name)
{
    QString id;

    // Ensuring a new Uuid. The loop should usually end after only
    // one iteration
    QStringList existingActivities = d->activities.keys();
    while (id.isEmpty() || existingActivities.contains(id)) {
        id = QUuid::createUuid();
        id.replace(QRegExp("[{}]"), QString());
    }

    d->setActivityState(id, Invalid);

    SetActivityName(id, name);

    StartActivity(id);

    emit ActivityAdded(id);

    d->configSync();
    return id;
}

void ActivityDirector::RemoveActivity(const QString & id)
{
    if (d->activities.size() < 2 ||
            !d->activities.contains(id)) {
        return;
    }

    // If the activity is running, stash it
    StopActivity(id);

    d->setActivityState(id, Invalid);

    // Removing the activity
    d->activities.remove(id);
    d->activitiesConfig().deleteEntry(id);

    // If the removed activity was the current one,
    // set another activity as current
    if (d->currentActivity == id) {
        d->ensureCurrentActivityIsRunning();
    }

    emit ActivityRemoved(id);
    d->configSync();
}

void ActivityDirector::StartActivity(const QString & id)
{
    if (!d->activities.contains(id) ||
            d->activities[id] == Running) {
        return;
    }

    d->setActivityState(id, Starting);

    // TODO: notify ksmserver of the event
    // and move the lines below to the ksmserver response
    // handler

    d->setActivityState(id, Running);

    emit ActivityStarted(id);
    d->configSync();
}

void ActivityDirector::StopActivity(const QString & id)
{
    if (!d->activities.contains(id) ||
            d->activities[id] == Stopped) {
        return;
    }

    d->setActivityState(id, Stopping);

    // TODO: notify ksmserver of the event
    // and move the lines below to the ksmserver response
    // handler

    d->ensureCurrentActivityIsRunning();

    d->setActivityState(id, Stopped);
    emit ActivityStopped(id);
    d->configSync();
}

int ActivityDirector::ActivityState(const QString & id) const
{
    if (!d->activities.contains(id)) {
        return Invalid;
    }

    else return d->activities[id];
}

QStringList ActivityDirector::ListActivities() const
{
    return d->activities.keys();
}

QStringList ActivityDirector::ListActivities(int state) const
{
    return d->activities.keys((State)state);
}

QString ActivityDirector::ActivityName(const QString & id) const
{
    return d->activityName(id);
}

void ActivityDirector::SetActivityName(const QString & id, const QString & name)
{
    if (!d->activities.contains(id)) {
        return;
    }

    d->activitiesConfig().writeEntry(id, name);

    if (NEPOMUK_RUNNING) {
        d->activityResource(id).setLabel(name);
    }

    d->scheduleConfigSync();
    emit ActivityChanged(id);
}

QString ActivityDirector::ActivityDescription(const QString & id) const
{
    if (!NEPOMUK_RUNNING || !d->activities.contains(id)) {
        return QString();
    }

    return d->activityResource(id).description();
}

void ActivityDirector::SetActivityDescription(const QString & id, const QString & description)
{
    if (!NEPOMUK_RUNNING || !d->activities.contains(id)) {
        return;
    }

    d->activityResource(id).setDescription(description);

    emit ActivityChanged(id);
}

QString ActivityDirector::ActivityIcon(const QString & id) const
{
    if (!NEPOMUK_RUNNING || !d->activities.contains(id)) {
        return QString();
    }

    QStringList symbols = d->activityResource(id).symbols();

    if (symbols.isEmpty()) {
        return QString();
    } else {
        return symbols.first();
    }
}

void ActivityDirector::SetActivityIcon(const QString & id, const QString & icon)
{
    if (!NEPOMUK_RUNNING || !d->activities.contains(id)) {
        return;
    }

    d->activityResource(id).setSymbols(QStringList() << icon);

    emit ActivityChanged(id);
}


// Resource related mothods

void ActivityDirector::NotifyResourceAccessed(const QString & application, const QString & uri)
{
    EventProcessor::addEvent(application, uri, EventProcessor::Accessed);
}

void ActivityDirector::NotifyResourceClosed(uint windowId, const QString & uri)
{
    d->resourcesForWindow[windowId].remove(KUrl(uri));

    EventProcessor::addEvent(d->applicationForWindow[windowId], uri, EventProcessor::Closed);

    if (d->resourcesForWindow[windowId].size() == 0) {
        d->resourcesForWindow.remove(windowId);
        d->applicationForWindow.remove(windowId);
    }
}

void ActivityDirector::NotifyResourceModified(uint windowId, const QString & uri)
{
    EventProcessor::addEvent(d->applicationForWindow[windowId], uri, EventProcessor::Modified);
}

void ActivityDirector::NotifyResourceOpened(const QString & application, uint windowId, const QString & uri)
{
    if (!d->applicationForWindow.contains(windowId)) {
        d->applicationForWindow[windowId] = application;
    }

    KUrl kuri(uri);
    d->resourcesForWindow[windowId] << kuri;
    d->activitiesForUrl[kuri] << CurrentActivity();

    EventProcessor::addEvent(application, uri, EventProcessor::Opened);
}

QStringList ActivityDirector::ActivitiesForResource(const QString & uri) const
{
    return d->activitiesForUrl.value(uri).toList();
}


