/*  This file is part of the KDE semantic clipboard
    Copyright (C) 2008 Tobias Wolf <twolf@access.unizh.ch>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301, USA.
*/

#include "sparqlmappingelementtest.h"

#include "autoptr.h"
#include "sparqlgraphtransformer.h"
#include "testdata.h"

#include <QtCore/QSet>
#include <QtCore/QString>
#include <QtCore/QUrl>
#include <QtTest/QtTest>

#include <qtest_kde.h>
#include <kstandarddirs.h>
#include <kzip.h>

using namespace SemClip;

void SparqlMappingElementTest::init()
{
    const QSet< QUrl > ontologies = TestData::exampleOntologies();
    m_mapping = new SparqlMappingElement( ontologies.values().at( 0 ), ontologies.values().at( 1 ), false );
    m_content = new SparqlMappingElement::Content();
    m_content->prefixes = TestData::examplePrefixMap();
    m_content->queries = TestData::exampleMappingQueries();
}

void SparqlMappingElementTest::cleanup()
{
    delete m_mapping;
    delete m_content;
}

void SparqlMappingElementTest::testContentToZip()
{
    const QString filePath = KStandardDirs::locateLocal( "data", TestData::exampleMappingFileName() );
    KZip output( filePath );
    output.open( QIODevice::WriteOnly );
    QVERIFY( SparqlMappingElement::contentToZip( &output, m_content ) );
}

void SparqlMappingElementTest::testContentFromZip()
{
    const QString filePath = KStandardDirs::locateLocal( "data", TestData::exampleMappingFileName() );
    KZip input( filePath );
    input.open( QIODevice::ReadOnly );
    QVERIFY( SparqlMappingElement::contentFromZip( &input, m_content ) );
    QCOMPARE( m_content->prefixes, TestData::examplePrefixMap() );
    QCOMPARE( m_content->queries, TestData::exampleMappingQueries() );
}

void SparqlMappingElementTest::testCreateTransformer()
{
    AutoPtr< GraphTransformer > transformer( m_mapping->createTransformer() );
    QVERIFY( !transformer.isNull() );
}

void SparqlMappingElementTest::testGenerateFileName()
{
    const QString fileName =
        SparqlMappingElement::generateFileName( m_mapping->sourceUri(), m_mapping->targetUri(), m_content );
    QCOMPARE( fileName, TestData::exampleMappingFileName() );
}

void SparqlMappingElementTest::testSetFileName()
{
    m_mapping->setFileName( "/" + TestData::exampleMappingFileName() );
    QCOMPARE( m_mapping->fileName(), TestData::exampleMappingFileName() );
}

void SparqlMappingElementTest::testSetDirectory()
{
    m_mapping->setDirectory( QString( "." ) );
    QCOMPARE( m_mapping->directory(), QString( "./" ) );
}

void SparqlMappingElementTest::testLoadContent()
{
    // test loading with an empty file name
    QVERIFY( !m_mapping->loadContent() );

    // test loading with an invalid file name
    m_mapping->setFileName( "whatever" );
    QVERIFY( !m_mapping->loadContent() );

    // test loading with a valid file name
    m_mapping->setFileName( TestData::exampleMappingFileName() );
    QVERIFY( m_mapping->loadContent() );
}

// depends on loadContent() working
void SparqlMappingElementTest::testContent()
{
    // test with a mapping element with an empty file name
    QVERIFY( !m_mapping->content() );

    // test with a mapping element with a valid file name
    m_mapping->setFileName( TestData::exampleMappingFileName() );
    QVERIFY( m_mapping->content() );
}

// depends on content() working
void SparqlMappingElementTest::testSetContent()
{
    QVERIFY( !m_mapping->content() );

    // test setting not null
    m_mapping->setContent( m_content );
    QVERIFY( m_mapping->content() );

    // test setting to not null (will delete the previous content, m_content)
    m_mapping->setContent( 0 );
    m_content = 0;
    QVERIFY( !m_mapping->content() );
}

// depends on loadContent() and setContent() working
void SparqlMappingElementTest::testSaveContent()
{
    // test with empty content
    m_mapping->setContent( 0 );
    QVERIFY( !m_mapping->saveContent() );

    // test with unmodified content
    m_mapping->setFileName( TestData::exampleMappingFileName() );
    m_mapping->loadContent();
    QVERIFY( m_mapping->saveContent() );

    // test with modified content
    m_mapping->setContent( m_content );
    m_content = 0;
    QVERIFY( m_mapping->saveContent() );
}

// depends on saveContent() and setContent() working
void SparqlMappingElementTest::testEraseContent()
{
    // test without persistent mapping content
    m_mapping->setFileName( "whatever.zip" );
    QVERIFY( !m_mapping->eraseContent() );

    // test with persistent mapping content
    m_mapping->setFileName( TestData::exampleMappingFileName() );
    QVERIFY( m_mapping->eraseContent() );
}

QTEST_KDEMAIN( SparqlMappingElementTest, NoGUI )

#include "sparqlmappingelementtest.moc"
