/*
  Copyright (C) 2010 Michael Leupold <lemma@confuego.org>

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <gcrypt.h>
#include <stdlib.h>

#define MIN(a, b) (a < b ? a : b)

static gcry_md_hd_t _handle;
static int hash_len;

static void hash_chunk(unsigned char *outputbuf, const unsigned char *data, int size)
{
   int i;

   gcry_md_reset(_handle);
   gcry_md_write(_handle, data, size);
   for (i = 0; i < 2000; ++i) {
      memcpy(outputbuf, gcry_md_read(_handle, 0), hash_len);
      gcry_md_reset(_handle);
      gcry_md_write(_handle, outputbuf, hash_len);
   }
}

void password_to_hash(unsigned char *hash, int *hashsz,
                      const unsigned char *password, int passwordsz)
{
   /* initialize */
   if (gcry_md_open(&_handle, GCRY_MD_SHA1, 0) != GPG_ERR_NO_ERROR) {
      *hashsz = 0;
      return;
   }
   hash_len = gcry_md_get_algo_dlen(GCRY_MD_SHA1);

   int buflength;
   
   unsigned char outputbuf1[hash_len];
   unsigned char outputbuf2[hash_len];
   unsigned char outputbuf3[hash_len];
   unsigned char outputbuf4[hash_len];
   
   buflength = MIN(passwordsz, 16);
   hash_chunk(outputbuf1, password, buflength);
   
   if (passwordsz > 16) {
      buflength = MIN(passwordsz - 16, 16);
      hash_chunk(outputbuf2, password + 16, buflength);

      if (passwordsz > 32) {
         buflength = MIN(passwordsz - 32, 16);
         hash_chunk(outputbuf3, password + 32, buflength);

         if (passwordsz > 48) {
            buflength = passwordsz - 48;
            hash_chunk(outputbuf4, password + 48, buflength);

            /* split 14/14/14/14 */
            *hashsz = 56;
            memcpy(hash     , outputbuf1, 14);
            memcpy(hash + 14, outputbuf2, 14);
            memcpy(hash + 28, outputbuf3, 24);
            memcpy(hash + 42, outputbuf4, 14);
         } else {
            /* split 20/20/16 */
            *hashsz = 56;
            memcpy(hash     , outputbuf1, 20);
            memcpy(hash + 20, outputbuf2, 20);
            memcpy(hash + 40, outputbuf3, 16);
         }
      } else {
         /* split 20/20 */
         *hashsz = 40;
         memcpy(hash     , outputbuf1, 20);
         memcpy(hash + 20, outputbuf2, 20);
      }
   } else {
      /* use entire first block */
      *hashsz = 20;
      memcpy(hash, outputbuf1, 20);
   }

   /* clean-up */
   gcry_md_close(_handle);
}
