/*
  Copyright (C) 2009 Pierre Ducroquet <pinaraf@pinaraf.info>
  Copyright (C) 2010 Michael Leupold <lemma@confuego.org>

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "opener.h"

#include <QtCore/QCoreApplication>
#include <KComponentData>
#include <KAboutData>
#include <KCmdLineArgs>
#include <KCmdLineOptions>
#include <KConfig>
#include <KConfigGroup>
#include <KDebug>
#include <KToolInvocation>

#include <QtCore/QTextStream>
#include <QtCrypto/QtCrypto>

#include <unistd.h>

static bool isWalletEnabled()
{
   KConfig cfg("kwalletrc");
   KConfigGroup walletGroup(&cfg, "Wallet");
   return walletGroup.readEntry("Enabled", true);
}

int main(int argc, char **argv)
{
   QCA::Initializer qcaInit;
   QTextStream err(stderr, QIODevice::WriteOnly);

   KAboutData aboutdata("pam_kwallet", 0, ki18n("KDE Wallet PAM Opener"),
                        "0.1", ki18n("Try to open a KDE Wallet on login"),
                        KAboutData::License_LGPL,
                        ki18n("(C) 2009, 2010 Pierre Ducroquet, Michael Leupold"));
   aboutdata.addAuthor(ki18n("Pierre Ducroquet"), ki18n("Original author"),
                       "pinaraf@pinaraf.info");
   aboutdata.addAuthor(ki18n("Michael Leupold"), ki18n("Adaption to new API"),
                       "lemma@confuego.org");
   KCmdLineArgs::init(argc, argv, &aboutdata);
   KCmdLineOptions options;
   options.add("localwallet", ki18n("Try to open the local wallet."));
   options.add("nonetworkwallet", ki18n("Try to open the network wallet."));
   options.add("start-daemon", ki18n("If set, try to start the KWallet daemon."));
   options.add("maxwait <time>",
               ki18n("Maximum number of seconds to wait for the wallet daemon to appear on the D-Bus."),
               "30");
   options.add("session-timeout <time>",
               ki18n("Maximum number of seconds to forcefully keep the session open after unlocking."),
               "120");
   KCmdLineArgs::addCmdLineOptions(options);
   KComponentData cmpData(&aboutdata);
   
   KCmdLineArgs *args = KCmdLineArgs::parsedArgs();
   QCoreApplication app(argc, argv);

   // check command-line arguments
   uint maxwait = args->getOption("maxwait").toUInt() * 1000;
   uint sessionTimeout = args->getOption("session-timeout").toUInt() * 1000;
   if (maxwait <= 0) {
      err << i18n("maxwait has to be a positive value.") << endl;
      return 3;
   }
   if (sessionTimeout <= 0) {
      err << i18n("session-timeout has to be a positive value.") << endl;
      return 3;
   }

   // make sure there is nothing fishy
   if (geteuid() != getuid() || getegid() != getgid() || !getuid()) {
      err << i18n("Priviledged environment!") << endl;
      return 2;
   }

   if (isatty(0)) {
      err << i18n("Need password from <stdin>.") << endl;
      return 1;
   }

   if (!isWalletEnabled()) {
      err << i18n("The KDE Wallet subsystem is disabled.") << endl;
      return 1;
   }

   // read the hash from stdin
   qint64 length;
   QCA::SecureArray hashedPw(56);
   QFile f;
   if (f.open(stdin, QIODevice::ReadOnly)) {
      length = f.readLine(hashedPw.data(), 256);
      hashedPw.resize((int)length);
   }

   daemon(0, 0);

   Opener opener;
   opener.setHashedPassword(hashedPw);
   opener.setOpenLocalWallet(args->isSet("localwallet"));
   opener.setOpenNetworkWallet(args->isSet("networkwallet"));
   opener.setMaxwait(maxwait);
   opener.setSessionTimeout(sessionTimeout);
   opener.setStartDaemon(args->isSet("start-daemon"));
   opener.start();
   return app.exec();
}
