/*
 * Copyright (C) 2007 Dirk Mueller <mueller@kde.org>
 * Copyright (C) 2010 Michael Leupold <lemma@confuego.org>
 *
 * based on pam_kwallet, by
 *
 * Copyright (C) 2004 W. Michael Petullo <mike@flyn.org>
 *                    Jonathan Nettleton <jon.nettleton@gmail.com>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 */

#include "config.h"
#include "hash.h"

#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <syslog.h>
#include <pwd.h>
#include <glib.h>
#include <sys/wait.h>
#include <sys/types.h>

#include <security/pam_appl.h>
#include <security/pam_modules.h>
#include <security/pam_ext.h>

#ifdef HAVE_SELINUX
   #include <selinux/selinux.h>
   #include <selinux/get_context_list.h>
#endif

#define UNUSED_ARGUMENT(x) (void)x

/**
 * Flags that can be passed to the PAM module
 */
enum {
   ARG_NO_NETWORKWALLET    = 1 << 0,    /** Don't open the network wallet */
   ARG_LOCALWALLET         = 1 << 1,    /** Open the local wallet */
   ARG_USE_FIRST_PASS      = 1 << 2,    /** Use the previous password */
   ARG_TRY_FIRST_PASS      = 1 << 3,    /** Try to use the previous password */
   ARG_START_DAEMON        = 1 << 4,    /** Start the KWallet Daemon */
   ARG_DEBUG               = 1 << 5,    /** Print debug messages */
   ARG_NO_WARN             = 1 << 6     /** Don't print warning messages */
};

/**
 * Structure passed to the g_spawn child_setup function.
 */
typedef struct preexec_t {
   const char *user;         /** User to run as */
   const char *evar;         /** Environment variables to set */
   const char *display;      /** X11 DISPLAY to use */
   const char *session_bus;  /** D-Bus session bus to use */
} preexec_t;

#ifdef HAVE_SELINUX
/**
 * Setup the selinux context for the user before launching/killing the daemon.
 *
 * @param ph PAM handle
 * @param user username to setup the context for
 */
static int selinux_setup(pam_handle_t *ph, const char *user)
{
   int retval = -1;
   security_context_t context;
   char *seuser = NULL;
   char *level = NULL;
   /* priority and return values to use in case of errors */
   int error_log_priority = LOG_ERR;
   int error_retval = PAM_SERVICE_ERR;

   /* do nothing if selinux is not enabled */
   if (is_selinux_enabled() != 1) {
      return PAM_SUCCESS;
   }

   if (security_getenforce() == 0) {
      /* permissive mode */
      error_log_priority = LOG_WARNING;
      error_retval = PAM_SUCCESS;
   } else {
      /* enforcing mode */
      error_log_priority = LOG_ERR;
      error_retval = PAM_SERVICE_ERR;
   }

   if (getseuserbyname(user, &seuser, &level) == 0) {
      retval = get_default_context_with_level(seuser, level, 0, &context);
      free(seuser);
      free(level);
   }

   if (retval < 0) {
      /* either seuser or default context couldn't be retrieved */
      pam_syslog(ph, error_log_priority,
                 "pam_dbus_launch: couldn't get default selinux context for %s", user);
      return error_retval;
   }

   /* set the security context for the exec */
   if (setexeccon(context) != 0) {
      pam_syslog(ph, error_log_priority,
                 "pam_dbus_launch: couldn't set executable context %s", (char*)context);
      freecon(context);
      return error_retval;
   }

   freecon(context);
   return PAM_SUCCESS;
}
#endif

/**
 * Parse the arguments passed to the PAM module.
 *
 * @param ph PAM handle
 * @param argc number of arguments
 * @param argv array of arguments
 * @param kwalletopener kwalletopener argument, path to the kwalletopener binary
 * @param maxwait maxwait argument, maximum number of seconds to wait for kwalletd
 * @param session_timeout session_timeout argument, maximum number of seconds to keep
 *                        the auto-unlocked wallet open
 * @return ORed flags that have been parsed
 */
static uint parse_args (pam_handle_t *ph, int argc, const char **argv,
                        const char **kwalletopener, int *maxwait, int *session_timeout)
{
   uint args = 0;
   const void *svc;
   int kwalletopener_len;
   int maxwait_len;
   int session_timeout_len;
   int i;
   const char *maxwait_str;
   const char *session_timeout_str;
   int tmp;

   svc = NULL;
   if (pam_get_item (ph, PAM_SERVICE, &svc) != PAM_SUCCESS) {
      svc = NULL;
   }

   kwalletopener_len = strlen("kwalletopener=");
   maxwait_len = strlen("maxwait=");
   session_timeout_len = strlen("session_timeout=");

   // defaults
   *maxwait = -1;
   *session_timeout = -1;

   /* Parse the arguments */
   for (i = 0; i < argc; i++) {
      if (strcmp(argv[i], "no_networkwallet") == 0) {
         args |= ARG_NO_NETWORKWALLET;
      } else if (strcmp(argv[i], "localwallet") == 0) {
         args |= ARG_LOCALWALLET;
      } else if (strcmp(argv[i], "start_daemon") == 0) {
         args |= ARG_START_DAEMON;
      } else if (strcmp(argv[i], "debug") == 0) {
         args |= ARG_DEBUG;
      } else if (strcmp(argv[i], "no_warn") == 0) {
         args |= ARG_NO_WARN;
      } else if (strncmp(argv[i], "kwalletopener=", kwalletopener_len) == 0) {
         *kwalletopener = (argv[i]) + kwalletopener_len;
         if (**kwalletopener == '\0') {
            *kwalletopener = NULL;
            pam_syslog(ph, LOG_WARNING,
                       "pam_kwallet: kwalletopener= specification missing argument");
         }
      } else if (strncmp(argv[i], "maxwait=", maxwait_len) == 0) {
         maxwait_str = (argv[i]) + maxwait_len;
         if (*maxwait_str != '\0') {
            tmp = atoi(maxwait_str);
            if (tmp <= 0) {
               pam_syslog(ph, LOG_WARNING,
                          "pam_kwallet: invalid value maxwait=%s", maxwait_str);
            } else {
               *maxwait = tmp;
            }
         }
      } else if (strncmp(argv[i], "session_timeout=", session_timeout_len) == 0) {
         session_timeout_str = (argv[i]) + session_timeout_len;
         if (*session_timeout_str != '\0') {
            tmp = atoi(session_timeout_str);
            if (tmp <= 0) {
               pam_syslog(ph, LOG_WARNING,
                          "pam_kwallet: invalid value session_timeout=%s", session_timeout_str);
            } else {
               *session_timeout = tmp;
            }
         }
      } else if (strcmp(argv[i], "use_first_pass") == 0) {
         args |= ARG_USE_FIRST_PASS;
      } else if (strcmp(argv[i], "try_first_pass") == 0) {
         args |= ARG_TRY_FIRST_PASS;
      } else {
         pam_syslog (ph, LOG_ERR, "pam_kwallet: invalid option %s", argv[i]);
      }
   }

   return args;
}

/**
 * Write the contents of a buffer to the given filedescriptor while ignoring
 * the SIGPIPE signal.
 *
 * @param fd open file descriptor to write to
 * @param buf buffer whose contents should be written
 * @param count number of bytes to write
 * @return same return values as the actual write call, -1 if there's a problem
 *         ignoring the SIGPIPE signal
 */
int pipewrite(int fd, const void *buf, size_t count)
{
   int fnval;
   struct sigaction ignoresact, oldsact;
   ignoresact.sa_handler = SIG_IGN;

   assert(fd >= 0);
   assert(buf != NULL);

   /* avoid bomb on command exiting before data written */
   if (sigaction(SIGPIPE, &ignoresact, &oldsact) < 0) {
      fnval = -1;
   } else {
      fnval = write(fd, buf, count);
      /* restore old handler */
      if (sigaction(SIGPIPE, &oldsact, NULL) < 0) {
         fnval = -1;
      }
   }
   return fnval;
}

/**
 * This function is passed to g_spawn as its child_setup function. It is executed
 * in between fork() and execve() and sets up the child's environment.
 *
 * @param data pointer to a preexec_t structure containing the neccessary data
 */
static void preexec(const gpointer data)
{
   struct passwd *passwd_ent;

   assert(data);
   assert(((preexec_t *) data)->user);

   passwd_ent = getpwnam((char *) ((preexec_t *) data)->user);
   if (!passwd_ent) {
      syslog(LOG_ERR, "pam_kwallet: error looking up user information for %s", (char *) ((preexec_t *) data)->user);
      exit(EXIT_FAILURE);
   }
   /* must exit because g_spawn_sync does not check return value before
    * executing argument vector */
   if (setgid(passwd_ent->pw_gid) == -1) {
      syslog(LOG_ERR, "pam_kwallet: error setting gid (%s)",  strerror(errno));
      exit(EXIT_FAILURE);
   }
   if (setuid(passwd_ent->pw_uid) == -1) {
      syslog (LOG_ERR, "pam_kwallet: error setting uid (%s)", strerror(errno));
      exit(EXIT_FAILURE);
   }
   if (seteuid(passwd_ent->pw_uid) == -1) {
      syslog (LOG_ERR, "pam_kwallet: error setting euid: %s", strerror(errno));
      exit(EXIT_FAILURE);
   }
   if (setenv( "HOME", passwd_ent->pw_dir, 1) == -1) {
      syslog (LOG_ERR, "pam_kwallet: error setting home: %s",  passwd_ent->pw_dir);
      exit(EXIT_FAILURE);
   }
   if (setenv("DISPLAY", ((preexec_t*)data)->display, 1) == -1) {
      syslog(LOG_ERR, "pam_kwallet: error setting display: %s", ((preexec_t*)data)->display);
      exit(EXIT_FAILURE);
   }
   if (setenv("DBUS_SESSION_BUS_ADDRESS", ((preexec_t*)data)->session_bus, 1) == -1) {
      syslog(LOG_ERR, "pam_kwallet: error setting session bus address: %s",
             ((preexec_t*)data)->session_bus);
      exit(EXIT_FAILURE);
   }

   if (((preexec_t*)data)->evar) {
      /* putenv does not want const char *, only char * */
      putenv((char*)((preexec_t*)data)->evar);
   }
}

/**
 * Concatenate two strings.
 *
 * @param dest already existing destination string to be appended to
 * @param src source string to be appended to the destination
 */
static void safe_strcat(char **dest, const char *src)
{
   char *tmp_str = g_strconcat(*dest, src, NULL);
   g_free(*dest);
   *dest = tmp_str;
}

/* overwrite the password hash and free the memory.
 * graciously taken from pam_keyring. */
/**
 * This method cleans up a hash by overwriting it.
 *
 * @param hash Hash to be overwritten
 * @remarks Graciously taken from GNOME keyring's gkr-pam-module.c
 */
static void overwrite_hash(char *hash)
{
   volatile char *vh;
   size_t len;

   if (!hash) {
      return;
   }

   /* Defeats some optimizations */
   len = strlen(hash);
   memset(hash, 0xAA, len);
   memset(hash, 0xBB, len);

   /* Defeats others */
   vh = (volatile char*)hash;
   while (*vh) {
      *(vh++) = 0xAA;
   }
}

/**
 * Cleanup function to free a hash stored with the PAM data.
 *
 * @param ph PAM handle
 * @param data data to be freed
 * @param pam_end_status ignored
 */
static void cleanup_free_hash(pam_handle_t *ph, void *data, int pam_end_status)
{
   UNUSED_ARGUMENT(ph);
   UNUSED_ARGUMENT(pam_end_status);
   
   overwrite_hash(data);
   free(data);
}

/**
 * Start the kwalletopener to open the specified wallets.
 *
 * @param ph PAM handle
 * @param data preexec data containing the values neccessary to start the opener
 * @param hash hashed password for opening the wallet
 * @param hashlen length of the hash
 * @param args ORed flags for this module
 * @param kwalletopener path of the kwalletopener binary
 * @param maxwait maximum number of seconds to wait for the KWallet daemon to get ready
 * @param session_timeout number of seconds to keep the wallet open
 */
static int unlock(pam_handle_t *ph, const preexec_t *data, const char *hash, int hashlen,
                  uint args, const char *kwalletopener, int maxwait, int session_timeout)
{
   gchar **opener_argv = NULL;
   char *tmp_str = NULL;
   GError *err = NULL;
   pid_t opener_pid = -1;
   int retval = PAM_SUCCESS;
   char *opener_cmd_line = NULL;
   int cstdin;

   assert(data);
   assert(data->user);
   assert(hash);

   if (kwalletopener != NULL) {
      opener_cmd_line = g_strconcat(kwalletopener, NULL);
   } else {
      opener_cmd_line = g_strconcat(KWALLETOPENER_PATH, NULL);
   }

   if ((args & ARG_LOCALWALLET) == ARG_LOCALWALLET) {
      safe_strcat(&opener_cmd_line, " --localwallet");
   }
   if ((args & ARG_NO_NETWORKWALLET) == ARG_NO_NETWORKWALLET) {
      safe_strcat(&opener_cmd_line, " --no-networkwallet");
   }
   if ((args & ARG_START_DAEMON) == ARG_START_DAEMON) {
      safe_strcat(&opener_cmd_line, " --start-daemon");
   }
   if (maxwait > 0) {
      tmp_str = g_strdup_printf(" --maxwait %d", maxwait);
      safe_strcat(&opener_cmd_line, tmp_str);
      g_free(tmp_str);
   }
   if (session_timeout > 0) {
      tmp_str = g_strdup_printf(" --session-timeout %d", session_timeout);
      safe_strcat(&opener_cmd_line, tmp_str);
      g_free(tmp_str);
   }

   if ((args & ARG_DEBUG) == ARG_DEBUG) {
      pam_syslog(ph, LOG_DEBUG, "pam_kwallet: going to execute %s", opener_cmd_line);
   }
   if (!g_shell_parse_argv(opener_cmd_line, NULL, &opener_argv, &err)) {
      pam_syslog(ph, LOG_ERR, "pam_kwallet: error parsing %s", opener_cmd_line);
      retval = PAM_SERVICE_ERR;
   } else {
#ifdef HAVE_SELINUX
      /* try to setup selinux executable context */
      retval = selinux_setup(ph, data->user);
#endif
      if (retval != PAM_SUCCESS || g_spawn_async_with_pipes(NULL, opener_argv, NULL,
                                                            0, preexec, (void *)data,
                                                            &opener_pid, &cstdin, NULL,
                                                            NULL, &err) == FALSE) {
#ifdef HAVE_SELINUX
         /* in case exec wasn't called, reset the selinux exec context */
         setexeccon(NULL);
#endif
         pam_syslog(ph, LOG_ERR, "pam_kwallet: error executing %s", opener_cmd_line);
         retval = PAM_SERVICE_ERR;
      } else {
         if (pipewrite(cstdin, hash, hashlen) != hashlen) {
            pam_syslog(ph, LOG_ERR, "pam_kwallet: error writing hash to kwalletopener");
            retval = PAM_SERVICE_ERR;
         } else {
            close(cstdin);
         }
      }
   }

   g_strfreev(opener_argv);
   g_free(opener_cmd_line);
   return retval;
}

/**
 * This function uses a PAM conversation function to get the password from the user.
 * It's getting used if use_first_pass is not specified as an argument and no previous
 * password was available.
 *
 * @param ph PAM handle
 * @return PAM_SUCCESS on success, PAM_SERVICE_ERR on error
 */
static int converse(pam_handle_t *ph)
{
   struct pam_conv *conv;
   struct pam_message msg[1];
   struct pam_message *pmsg[1];
   struct pam_response *resp;
   pmsg[0] = &msg[0];
   msg[0].msg_style = PAM_PROMPT_ECHO_OFF;
   msg[0].msg = "Password: ";

   if (pam_get_item(ph, PAM_CONV, (const void**)&conv) == PAM_SUCCESS) {
      if (conv->conv(1, (const struct pam_message**)pmsg, &resp, conv->appdata_ptr) == PAM_SUCCESS) {
         /* got the password */
         if (pam_set_item(ph, PAM_AUTHTOK, resp[0].resp) != PAM_SUCCESS) {
            pam_syslog(ph, LOG_ERR, "pam_kwallet: can't set password item");
            return PAM_SERVICE_ERR;
         }
         return PAM_SUCCESS;
      } else {
         pam_syslog(ph, LOG_ERR, "pam_kwallet: error conversing with application");
         return PAM_SERVICE_ERR;
      }
   } else {
      pam_syslog(ph, LOG_ERR, "pam_kwallet: error getting conversation function");
      return PAM_SERVICE_ERR;
   }
}

/**
 * PAM function called during authentication.
 *
 * This function first tries to get a password from PAM, then hashes it. Afterwards two
 * scenarios are possible:
 *
 * - A D-Bus session is already available which usually means that the user is already
 *   logged on and PAM has been used inside the screensaver. In that case, try to reopen
 *   the wallet by launching the opener directly.
 *
 * - A D-Bus session is not yet available. Store the hashed password inside PAM data so
 *   it can be retrieved during pam_open_session to start the opener.
 *
 * @param ph PAM handle
 * @param unused unused
 * @param argc number of arguments for this PAM module
 * @param argv array of arguments for this PAM module
 * @return any of the PAM return values
 */
PAM_EXTERN int pam_sm_authenticate(pam_handle_t *ph, int unused, int argc, const char **argv)
{
   struct passwd *pwd;
   const char *password;
   const char *kwalletopener = NULL;
   int maxwait, session_timeout;
   char hash[57];
   uint args;
   int ret, hashsz, pwlen;
   preexec_t data = { NULL, NULL, NULL, NULL };

   args = parse_args(ph, argc, argv, &kwalletopener, &maxwait, &session_timeout);
   
   /* Figure out and/or prompt for the user name */
   ret = pam_get_user(ph, &data.user, NULL);
   if (ret != PAM_SUCCESS || !data.user) {
      pam_syslog(ph, LOG_ERR, "pam_kwallet: couldn't get the user name: %s",
                 pam_strerror(ph, ret));
      return PAM_SERVICE_ERR;
   }

   pwd = getpwnam(data.user);
   if (!pwd) {
      pam_syslog(ph, LOG_ERR, "pam_kwallet: error looking up user information for: %s", data.user);
      return PAM_SERVICE_ERR;
   }

   /* Look up the password */
   if ((args & ARG_USE_FIRST_PASS) == 0 && (args & ARG_TRY_FIRST_PASS) == 0) {
      /* not allowed to use existing password. */
      ret = converse(ph);
      if (ret != PAM_SUCCESS) {
         return ret;
      }
   }
   /* now we might have a password set. */
   ret = pam_get_item(ph, PAM_AUTHTOK, (const void**)&password);

   if ((ret != PAM_SUCCESS || password == NULL) &&
       (args & ARG_TRY_FIRST_PASS) == ARG_TRY_FIRST_PASS) {
      /* no password set, but as try_first_pass is set we are allowed to ask. */
      ret = converse(ph);
      if (ret != PAM_SUCCESS) {
         return ret;
      }
      /* now we might have a password set. try regetting it. */
      ret = pam_get_item(ph, PAM_AUTHTOK, (const void**)&password);
   }

   if (ret != PAM_SUCCESS || password == NULL) {
      if (ret == PAM_SUCCESS) {
         if ((args & ARG_NO_WARN) != ARG_NO_WARN) {
            pam_syslog(ph, LOG_WARNING, "pam_kwallet: no password is available for user");
         }
      } else {
         if ((args & ARG_NO_WARN) != ARG_NO_WARN) {
            pam_syslog(ph, LOG_WARNING, "pam_kwallet: no password is available for user: %s",
                     pam_strerror(ph, ret));
         }
      }
      return PAM_SUCCESS;
   }

   pwlen = strlen((const char*)password);
   if (pwlen == 103 || pwlen == 167 || pwlen == 231) {
      pam_syslog(ph, LOG_ERR, "pam_kwallet: passwords with length 103, 167 and 231 are not supported");
      return PAM_SUCCESS;
   }
   
   password_to_hash((unsigned char*)hash, &hashsz, (const unsigned char*)password,
                    strlen(password));
   hash[hashsz] = 0;

   /* check if the bus is already running. this usually means that we are already
    * inside a session (eg. running inside screensaver) and can launch the opener
    * right away. */
   data.session_bus = getenv("DBUS_SESSION_BUS_ADDRESS");
   if (data.session_bus == NULL) {
      data.session_bus = pam_getenv(ph, "DBUS_SESSION_BUS_ADDRESS");
   }
   if (data.session_bus != NULL) {
      if ((args & ARG_DEBUG) == ARG_DEBUG) {
         pam_syslog(ph, LOG_DEBUG, "pam_kwallet: launching opener during pam_authenticate");
      }
      ret = pam_get_item(ph, PAM_XDISPLAY, (const void**)(&data.display));
      if (ret != PAM_SUCCESS || data.display == NULL) {
         data.display = getenv("DISPLAY");
         if (data.display == NULL) {
            pam_syslog(ph, LOG_ERR, "pam_kwallet: DISPLAY not set");
            return PAM_SUCCESS;
         }
      }
      
      return unlock(ph, &data, hash, hashsz, args, kwalletopener, maxwait, session_timeout);
   } else {
      /* new session. set hashed password as pam data and launch during open_session. */
      if (pam_set_data(ph, "kwallet_password_hash", strdup(hash), cleanup_free_hash) != PAM_SUCCESS) {
         pam_syslog(ph, LOG_ERR, "pam_kwallet: error storing password hash");
         return PAM_AUTHTOK_RECOVER_ERR;
      }
      return PAM_SUCCESS;
   }
}

/**
 * PAM function called during opening the session.
 *
 * Retrieves the hashed password stored during authentication from PAM data, then uses
 * it to start the opener.
 *
 * @param ph PAM handle
 * @param flags currently unused, TODO: check for silent flag
 * @param argc number of arguments for this PAM module
 * @param argv array of arguments for this PAM module
 * @return any of the PAM return values
 */
PAM_EXTERN int pam_sm_open_session(pam_handle_t *ph, int flags, int argc, const char **argv)
{
   struct passwd *pwd;
   const char *kwalletopener = NULL;
   int maxwait, session_timeout;
   const char *hash = NULL;
   const char *cleaner = NULL;
   int hashsz;
   uint args;
   int retval;
   preexec_t data = { NULL, NULL, NULL, NULL };

   args = parse_args(ph, argc, argv, &kwalletopener, &maxwait, &session_timeout);

   /* Figure out and/or prompt for the user name */
   retval = pam_get_user(ph, &data.user, NULL);
   if (retval != PAM_SUCCESS || !data.user) {
      pam_syslog(ph, LOG_ERR, "pam_kwallet: couldn't get the user name: %s",
                 pam_strerror(ph, retval));
      return PAM_SERVICE_ERR;
   }

   pwd = getpwnam(data.user);
   if (!pwd) {
      pam_syslog(ph, LOG_ERR, "pam_kwallet: error looking up user information for: %s", data.user);
      return PAM_SERVICE_ERR;
   }

   /* retrieve the stored password hash. */
   if (pam_get_data(ph, "kwallet_password_hash", &hash) != PAM_SUCCESS || hash == NULL) {
      /* no password might mean the application didn't do authentication which might happen.
         don't complain in that case. */
      if ((args & ARG_NO_WARN) != ARG_NO_WARN) {
         pam_syslog(ph, LOG_WARNING, "pam_kwallet: couldn't retrieve stored password hash");
      }
      return PAM_SUCCESS;
   }

   hashsz = strlen(hash);

   /* first, try to get the session bus address from the environment. */
   data.session_bus = getenv("DBUS_SESSION_BUS_ADDRESS");
   if (data.session_bus == NULL || *data.session_bus == '\0') {
      data.session_bus = pam_getenv(ph, "DBUS_SESSION_BUS_ADDRESS");
      if (data.session_bus == NULL || *data.session_bus == '\0') {
         pam_syslog(ph, LOG_ERR, "pam_kwallet: D-Bus session bus not available");
         return PAM_SERVICE_ERR;
      }
   }
   
   retval = pam_get_item(ph, PAM_XDISPLAY, &data.display);
   if (retval != PAM_SUCCESS || data.display == NULL) {
      pam_syslog(ph, LOG_ERR, "pam_kwallet: DISPLAY not set");
      return PAM_SERVICE_ERR;
   }

   retval = unlock(ph, &data, hash, hashsz, args, kwalletopener, maxwait, session_timeout);

   /* overwrite the password hash and re-set it. */
   cleaner = strdup(hash);
   overwrite_hash(cleaner);
   pam_set_data(ph, "kwallet_password_hash", cleaner, cleanup_free_hash);

   return retval;
}

/**
 * This is called when the PAM session is closed. It should check if the KWallet daemon
 * and the kwalletopener launched are still running and close them if they are.
 *
 * Currently it does nothing though.
 *
 * @param ph PAM handle
 * @param flags currently unused, TODO: check for silent flag
 * @param argc number of arguments for this PAM module
 * @param argv array of arguments for this PAM module
 * @return PAM_SUCCESS
 */
PAM_EXTERN int pam_sm_close_session(pam_handle_t *ph, int flags, int argc, const char **argv)
{
   return PAM_SUCCESS;
}

/**
 * This is called when PAM is invoked to change the user's authentication token. It's not
 * needed.
 *
 * @param ph PAM handle
 * @param flags currently unused, TODO: check for silent flag
 * @param argc number of arguments for this PAM module
 * @param argv array of arguments for this PAM module
 * @return PAM_SUCCESS
 * @remarks GNOME keyring uses it to change the keyring's password as well. For this module
 *          this isn't a planned feature though.
 */
 PAM_EXTERN int pam_sm_setcred(pam_handle_t *ph, int flags, int argc, const char **argv)
{
   return PAM_SUCCESS;
}
