/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   Copyright (C) 2008 by Dennis Nienhüser <earthwings@gentoo.org>        *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

#include "contacts.h"

#include <QGraphicsLinearLayout>
#include <QGraphicsLayoutItem>
#include <QGraphicsWidget>
#include <QPainter>
#include <QListWidget>
#include <QModelIndexList>
#include <QAbstractItemModel>
#include <QStandardItemModel>

#include <KDebug>
#include <KConfigDialog>

#include <Plasma/Svg>
#include <Plasma/Applet>
#include <Plasma/Theme>
#include <Plasma/DataEngine>
#include <Plasma/FlashingLabel>

#include "contactwidget.h"
#include "cellfactory.h"

#include "globalwidget.h"
#include "marblecontactswidget.h"
#include <marble/MapThemeManager.h>

PlasmaContacts::PlasmaContacts(QObject *parent, const QVariantList &args) :
	Plasma::Applet(parent, args), m_layout(0)
{
	setHasConfigurationInterface(true);
	setAspectRatioMode(Plasma::IgnoreAspectRatio);
	setBackgroundHints(Applet::DefaultBackground);
	resize(200, 100);
}

PlasmaContacts::~PlasmaContacts()
{
	m_theme->deleteLater();
	m_flash->deleteLater();
}

void PlasmaContacts::init()
{
	// Setup a background theme. Not used atm.
	m_theme = new Plasma::Svg(this);
	m_theme->setImagePath("widgets/contacts");
	m_theme->setContainsMultipleImages(false);

	m_layout = new QGraphicsLinearLayout(Qt::Vertical);
	setLayout(m_layout);

	// Setup the config indicator used when no contacts are added
	m_flash = new Plasma::FlashingLabel(0);
	m_flash->setColor(Qt::gray);
	QFont fnt = qApp->font();
	fnt.setBold(true);
	QFontMetrics fm(fnt);
	m_flash->setFont(fnt);

	m_engine = dataEngine("contacts");
	load();
}

void PlasmaContacts::dataUpdated(const QString &source,
		const Plasma::DataEngine::Data &data)
{
	Plasma::DataEngine::Data::const_iterator iter = data.constBegin();
	QVariantMap properties;
	for (; iter != data.constEnd(); ++iter)
	{
		properties[iter.key()] = iter.value();
	}

	// Update single contact
	if (m_contactWidgets.contains(source))
	{
		// Forward data update to the contact row
		m_contactWidgets.value(source)->propertiesChanged(properties);
	}

	// Update global widgets
	foreach(GlobalWidget *widget, m_globalWidgets)
	{
		widget->propertiesChanged(source, properties);
	}
}

void PlasmaContacts::createConfigurationInterface(KConfigDialog *parent)
{
	QWidget *contactsWidget = new QWidget(0);
	contactsPage.setupUi(contactsWidget);

	QWidget *displayWidget = new QWidget(0);
	displayPage.setupUi(displayWidget);

	/** @todo: code duplication here, split config handling out to its own class */
	KConfigGroup cg = config();
	QStringList contacts = cg.readEntry("contacts", QStringList());

	if (m_engine)
	{
		QStringList sources = m_engine->sources();
		foreach(const QString &contactId, sources)
		{
			QListWidgetItem *item = new QListWidgetItem(contactId);
			item->setData(Qt::UserRole, contactId);
			LabelUpdater *updater = new LabelUpdater(item, contactId, m_engine);
			m_labelUpdaters.append(updater);
			m_engine->connectSource(contactId, updater);
			if (contacts.contains(contactId))
			{
				contactsPage.contacts->selectedListWidget()->addItem(item);
			}
			else
			{
				contactsPage.contacts->availableListWidget()->addItem(item);
			}
		}

		QStringList cells = cg.readEntry("cells", CellFactory::defaultCells());
		QStringList knownCells = CellFactory::knownCells();
		foreach (const QString &cell, knownCells)
		{
			// TODO: use some i18n()ed readable value instead of the raw factory identifier
			QListWidgetItem *item = new QListWidgetItem(cell);
			if (cells.contains(cell))
			{
				displayPage.contactsSelection->selectedListWidget()->addItem(item);
			}
			else
			{
				displayPage.contactsSelection->availableListWidget()->addItem(item);
			}
		}

		Marble::MapThemeManager *marbleThemeManager = new Marble::MapThemeManager;
		marbleThemeManager->updateMapThemeModel();
		/** @todo: restore marble theme selection */
		displayPage.contactsShowWidget->setChecked(cg.readEntry("contacts_enabled", true));
		displayPage.marbleTheme->setModel(marbleThemeManager->mapThemeModel());
		displayPage.marbleShowWidget->setChecked(cg.readEntry("marble_enabled", false));
		displayPage.marbleRecenterOnLocationChange->setChecked(cg.readEntry("marble_recenter", true));
	}

	connect(parent, SIGNAL(applyClicked()), this, SLOT(configAccepted()));
	connect(parent, SIGNAL(okClicked()), this, SLOT(configAccepted()));

	parent->addPage(contactsWidget, i18n("Contact Selection"), "system-users");
	parent->addPage(displayWidget, i18n("Display Properties"), "preferences-contact-list");
}

void PlasmaContacts::configAccepted()
{
	// Stop updating labels
	foreach(LabelUpdater* updater, m_labelUpdaters)
	{
		updater->deleteLater();
	}
	m_labelUpdaters.clear();

	// Store contact ids
	QStringList contacts;
	for (int i = 0; i < contactsPage.contacts->selectedListWidget()->count(); ++i)
	{
		QListWidgetItem *item =
				contactsPage.contacts->selectedListWidget()->item(i);
		Q_ASSERT(item);
		contacts << item->data(Qt::UserRole).toString();
	}

	// Store display configuration
	QStringList cells;
	for (int i = 0; i < displayPage.contactsSelection->selectedListWidget()->count(); ++i)
	{
		cells << displayPage.contactsSelection->selectedListWidget()->item(i)->text();
	}

	// Store general layout configuration
	QModelIndexList selectedThemes = displayPage.marbleTheme->selectionModel()->selectedIndexes();
	QString marbleTheme = "earth/openstreetmap/openstreetmap.dgml";
	if (selectedThemes.size() == 1)
	{
		QAbstractItemModel *model = displayPage.marbleTheme->model();
		QModelIndex colindex = model->index( selectedThemes.at(0).row(), 1, QModelIndex() );
		marbleTheme = (model->data( colindex )).toString();
	}

	KConfigGroup cg = config();
	cg.writeEntry("marble_enabled", displayPage.marbleShowWidget->isChecked());
	cg.writeEntry("marble_recenter", displayPage.marbleRecenterOnLocationChange->isChecked());
	cg.writeEntry("marble_theme", marbleTheme);
	cg.writeEntry("contacts_enabled", displayPage.contactsShowWidget->isChecked());
	cg.writeEntry("contacts", contacts);
	cg.writeEntry("cells", cells);
	load();
}

void PlasmaContacts::load()
{
	for (int i = m_layout->count() - 1; i >= 0; --i)
	{
		m_layout->removeAt(i);
	}

	QHash<QString, ContactWidget*>::const_iterator itr =
			m_contactWidgets.constBegin();
	while (itr != m_contactWidgets.constEnd())
	{
	    m_engine->disconnectSource(itr.key(), this);
		itr.value()->deleteLater();
		++itr;
	}
	m_contactWidgets.clear();

	foreach(GlobalWidget *widget, m_globalWidgets)
	{
		widget->deleteLater();
	}
	m_globalWidgets.clear();

	KConfigGroup cg = config();
	QStringList contacts = cg.readEntry("contacts", QStringList());
	bool contactsWidgetEnabled = cg.readEntry("contacts_enabled", true);

	if (contactsWidgetEnabled)
	{
		QStringList cells = cg.readEntry("cells", CellFactory::defaultCells());
		QStringList widgets = cg.readEntry("widgets", QStringList());
		foreach(const QString &contact, contacts)
		{
			m_engine->connectSource(contact, this);
			ContactWidget *widget = new ContactWidget(cells, QVariantMap());
			m_contactWidgets[contact] = widget;
			dataUpdated(contact, m_engine->query(contact));
			m_layout->addItem(widget);
		}
	}

	if (contacts.size())
	{
		m_flash->kill();

		/** @todo: Move to some factory class if the list grows */
		bool marbleWidgetEnabled = cg.readEntry("marble_enabled", false);
		bool marbleRecenter = cg.readEntry("marble_recenter", false);
		QString marbleTheme = cg.readEntry("marble_theme", "earth/openstreetmap/openstreetmap.dgml");

		if (marbleWidgetEnabled)
		{
			MarbleContactsWidget *marbleWidget = new MarbleContactsWidget(marbleTheme);
			marbleWidget->setRecenterOnLocationChange(marbleRecenter);
			m_layout->addItem(marbleWidget);
			m_globalWidgets.append(marbleWidget);
		}
	}
	else
	{
		m_layout->addItem(m_flash);
		m_flash->flash(i18n("You can add contacts in the settings dialog."), 10000);
	}
}

void PlasmaContacts::paintInterface(QPainter *p,
		const QStyleOptionGraphicsItem *option, const QRect &contentsRect)
{
	// TODO: rendering the background svg looks strange for some reason
	//    p->setRenderHint(QPainter::SmoothPixmapTransform);
	//	m_theme->paint(p, contentsRect.topLeft().x(), contentsRect.topLeft().y());
	Plasma::Applet::paintInterface(p, option, contentsRect);
}

#include "contacts.moc"

K_EXPORT_PLASMA_APPLET(contacts, PlasmaContacts)
