/***************************************************************************
 *   Copyright (C) 2009 by Matthieu Gallien <matthieu_gallien@yahoo.fr>    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

#include "tasks_view.h"

#include "tasks_roles.h"
#include "tasks_view_item.h"

#include <QGraphicsLinearLayout>
#include <QAbstractItemModel>
#include <KDebug>

TasksView::TasksView(DataEnginesTasks *applet,
                     Plasma::DataEngine *tasksEngine,
                     Plasma::DataEngine *statusNotifierItemEngine,
                     QGraphicsItem *parent,
                     Qt::WindowFlags wFlags)
    : QGraphicsWidget(parent, wFlags),
      m_applet(applet),
      m_tasksEngine(tasksEngine),
      m_statusNotifierItemEngine(statusNotifierItemEngine),
      m_layout(NULL),
      m_model(NULL),
      m_items()
{
    setSizePolicy(QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding));
    m_layout = new QGraphicsLinearLayout(this);
    m_layout->setOrientation(Qt::Vertical);
    m_layout->setSpacing(1);
    m_layout->setContentsMargins(0, 0, 0, 0);
    setContentsMargins(0, 0, 0, 0);
}

QAbstractItemModel* TasksView::model() const
{
    return m_model;
}

void TasksView::setOrientation(Qt::Orientation orientation)
{
    m_layout->setOrientation(orientation);
}

void TasksView::setModel(QAbstractItemModel *model)
{
    if (m_model) {
        disconnect(m_model, SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)),
                   this, SLOT(sourceDataChanged(const QModelIndex&, const QModelIndex&)));
        disconnect(m_model, SIGNAL(modelReset()),
                   this, SLOT(sourceModelReset()));
        disconnect(m_model, SIGNAL(rowsInserted(const QModelIndex&, int, int)),
                   this, SLOT(sourceRowsInserted(const QModelIndex&, int, int)));
        disconnect(m_model, SIGNAL(rowsAboutToBeRemoved(const QModelIndex&, int, int)),
                   this, SLOT(sourceRowsAboutToBeRemoved(const QModelIndex&, int, int)));
    }
    m_model = model;
    connect(m_model, SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)),
            this, SLOT(sourceDataChanged(const QModelIndex&, const QModelIndex&)));
    connect(m_model, SIGNAL(modelReset()),
            this, SLOT(sourceModelReset()));
    connect(m_model, SIGNAL(rowsInserted(const QModelIndex&, int, int)),
            this, SLOT(sourceRowsInserted(const QModelIndex&, int, int)));
    connect(m_model, SIGNAL(rowsAboutToBeRemoved(const QModelIndex&, int, int)),
            this, SLOT(sourceRowsAboutToBeRemoved(const QModelIndex&, int, int)));
    resetView();
}

void TasksView::resetView()
{
    for (int taskIndice = 0; taskIndice < m_items.size(); ++taskIndice) {
        m_layout->removeAt(taskIndice);
        m_items[taskIndice]->deleteLater();
        m_items.removeAt(taskIndice);
    }

    sourceRowsInserted(QModelIndex(), 0, m_model->rowCount() - 1);
}

void TasksView::sourceDataChanged(const QModelIndex &topLeft, const QModelIndex &bottomRight)
{
    if (topLeft.isValid()) {
        for (int i = topLeft.row(); i <= bottomRight.row(); ++i) {
            if (i < m_items.size()) {
                m_items[i]->dataChanged();
            }
        }
    }
}

void TasksView::sourceModelReset()
{
    resetView();
}

void TasksView::sourceRowsInserted(const QModelIndex &parent, int start, int end)
{
    kDebug() << "parent:" << parent << "start:" << start << "end:" << end;
    for (int i = start; i <= end; ++i) {
        const QModelIndex &insertedItem(m_model->index(i, 0, parent));
        sourceRowInserted(insertedItem, parent);
    }
}

void TasksView::sourceRowInserted(const QModelIndex &newItem,
                                  const QModelIndex &parent)
{
    kDebug() << "newItem: " << newItem << "parent:" << parent;
    //insert the child items inline in the layout
    const int newPosition = newItem.row();
    insertNewItemInView(newItem, newPosition);
}

void TasksView::insertNewItemInView(const QModelIndex &modelIndex, int viewPosition)
{
    kDebug() << "modelIndex: " << modelIndex << "viewPosition:" << viewPosition;
    TasksViewItem *newTaskItem = new TasksViewItem(modelIndex, m_tasksEngine,
                                                   m_statusNotifierItemEngine,
                                                   m_applet, this);
    m_items.insert(viewPosition, newTaskItem);
    m_layout->insertItem(viewPosition, newTaskItem);
}

void TasksView::sourceRowsAboutToBeRemoved(const QModelIndex &parent, int start, int end)
{
    for (int i = start; i <= end; ++i) {
        sourceRowAboutToBeRemoved(m_model->index(i, 0, parent), parent);
    }
}

void TasksView::sourceRowAboutToBeRemoved(const QModelIndex &newItem, const QModelIndex &parent)
{
    Q_UNUSED(parent);

    //insert the child items inline in the layout
    int removedPosition = newItem.row();
    m_layout->removeAt(removedPosition);
    m_items[removedPosition]->deleteLater();
    m_items.removeAt(removedPosition);
}

void TasksView::debugPrint() const
{
    for (int i = 0; i < m_items.size(); ++i) {
        if (m_items[i]) {
            kDebug() << "item at position" << i << m_items[i]->getIndex() << m_items[i]->getIndex().data(Qt::DisplayRole);
        } else {
            kDebug() << "item at position" << i << "is null";
        }
    }
}
