/*
 *   Copyright 2008 Vladimir Kuznetsov <ks.vladimir@gmail.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2,
 *   or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "embed-win.h"

#include <QGraphicsView>
#include <QGraphicsLinearLayout>
#include <QGraphicsProxyWidget>
#include <QLabel>
#include <QAction>

#include <QApplication>

#include <Plasma/Theme>

#include <QX11Info>
#include <X11/Xlib.h>

EmbedWin::EmbedWin(QObject *parent, const QVariantList &args)
    : Plasma::Applet(parent, args),
      m_layout(0), m_proxy(0), m_label(0), m_grabber(0),
      m_window(0), m_reparenting(false), m_container(0), m_dashboardView(0)
{
    setAspectRatioMode(Plasma::IgnoreAspectRatio);
    resize(500, 500);
    setMinimumSize(50, 50);
}

EmbedWin::~EmbedWin()
{
    if(m_container) m_container->discardClient();
    delete m_container;
    delete m_grabber;
}

void EmbedWin::init()
{
    m_grabber = new QWidget( 0,  Qt::X11BypassWindowManagerHint );
    m_grabber->move(-1000, -1000);
    m_grabber->installEventFilter(this);
    m_grabber->hide();

    m_label = new QLabel(i18n(WELCOME_MSG));
    m_label->setAlignment(Qt::AlignHCenter | Qt::AlignVCenter);
    m_label->setWordWrap(true);
    
    QColor color = Plasma::Theme::defaultTheme()->color(Plasma::Theme::TextColor);
    QPalette palette = m_label->palette();
    palette.setColor(QPalette::Background, Qt::transparent);
    palette.setColor(QPalette::Normal, QPalette::WindowText, color);
    palette.setColor(QPalette::Inactive, QPalette::WindowText, color);
    m_label->setPalette(palette);

    connect(m_label, SIGNAL(linkActivated(const QString&)),
            this, SLOT(linkActivated(const QString&)));

    m_layout = new QGraphicsLinearLayout;
    m_proxy = new QGraphicsProxyWidget(this);
    m_proxy->setWidget(m_label);
    m_layout->addItem(m_proxy);
    setLayout(m_layout);

    m_container = new QX11EmbedContainer();
    m_container->hide();
    connect(m_container, SIGNAL(clientIsEmbedded()),
            this, SLOT(containerClientIsEmbedded()));
    connect(m_container, SIGNAL(clientClosed()),
            this, SLOT(containerClientClosed()));
    connect(m_container, SIGNAL(error(QX11EmbedContainer::Error)),
            this, SLOT(containerError(QX11EmbedContainer::Error)));

    m_actionEmbed = new QAction(i18n("Embed new application"), this);
    m_actionDiscard = new QAction(i18n("Release embedded application"), this);
    connect(m_actionEmbed, SIGNAL(triggered(bool)), this, SLOT(doEmbed()));
    connect(m_actionDiscard, SIGNAL(triggered(bool)), this, SLOT(doDiscard()));

    m_label->installEventFilter(this);
}

QList<QAction*> EmbedWin::contextualActions()
{
    QList<QAction*> actions;
    if(m_window != 0) actions << m_actionDiscard;
    else actions << m_actionEmbed;
    return actions;
}

void EmbedWin::updateWidgetGeometry()
{
    if(m_reparenting || !scene() || !view() || !m_container) return;

    QGraphicsView *activeView = view();
    QList<QGraphicsView *> views = scene()->views();
    foreach(QGraphicsView *view, views) {
        if(view->metaObject()->className() == QString("DashboardView")) {
            if(view != m_dashboardView) {
                m_dashboardView = view;
                m_dashboardView->installEventFilter(this);
            }
            if(view->isVisible()) {
                activeView = view;
                break;
            }
        }
    }

    if(m_container->parentWidget() != activeView) {
        if(m_container->clientWinId() != 0) {
            kDebug() << "beginSwitch";
            // We can't reparent container while client is embedded
            // containerClientClosed function will call us again
            // and then reembed client. We will also temporary unmap
            // the window in order to avoid flicker
            m_reparenting = true;
            XUnmapWindow(QX11Info::display(), m_window);
            m_container->discardClient();
        } else {
            kDebug() << "endSwitch";
            m_container->setParent(activeView);
        }
    }

    qreal leftMargin, topMargin, rightMargin, bottomMargin;
    getContentsMargins(&leftMargin, &topMargin, &rightMargin, &bottomMargin);

    QRectF rf = rect().adjusted(leftMargin, topMargin, -rightMargin, -bottomMargin);
    QRect ri = mapToView(activeView, rf);
    m_container->setGeometry(ri);
}

void EmbedWin::linkActivated(const QString& link)
{
    if(link == "embed://embed")
        doEmbed();
}

void EmbedWin::doEmbed()
{
    Q_ASSERT(m_window == 0);
    m_label->setText(i18n("Click with the left mouse button on the window you want to embed.\n"
                          "Click anywhere with the right mouse button to cancel."));
    m_grabber->show();
    m_grabber->grabMouse(Qt::CrossCursor);
}

void EmbedWin::doDiscard()
{
    m_window = 0;
    m_container->discardClient();
}

bool EmbedWin::eventFilter(QObject *obj, QEvent *event)
{
    if((obj == m_label && event->type() == QEvent::Paint) ||
            (obj == m_dashboardView && event->type() == QEvent::Hide)) {
        updateWidgetGeometry();
    } else if(obj == m_grabber && QWidget::mouseGrabber() == m_grabber &&
            event->type() == QEvent::MouseButtonPress) {
        QMouseEvent* mouseEvent = static_cast<QMouseEvent*>(event);
        m_grabber->releaseMouse();
        m_grabber->hide();
        if(mouseEvent->button() == Qt::LeftButton) {
            embedCurrentWindow();
        } else {
            m_label->setText(i18n(WELCOME_MSG));
        }
    }
    return Plasma::Applet::eventFilter(obj, event);
}

static
Window findRealWindow( Window w, int depth = 0 )
{
    if( depth > 5 )
        return None;
    static Atom wm_state = XInternAtom( QX11Info::display(), "WM_STATE", False );
    Atom type;
    int format;
    unsigned long nitems, after;
    unsigned char* prop;
    if( XGetWindowProperty( QX11Info::display(), w, wm_state, 0, 0, False, AnyPropertyType,
                &type, &format, &nitems, &after, &prop ) == Success ) {
        if( prop != NULL )
            XFree( prop );
        if( type != None )
            return w;
    }
    Window root, parent;
    Window* children;
    unsigned int nchildren;
    Window ret = None;
    if( XQueryTree( QX11Info::display(), w, &root, &parent, &children, &nchildren ) != 0 ) {
    for( unsigned int i = 0;
         i < nchildren && ret == None;
         ++i )
        ret = findRealWindow( children[ i ], depth + 1 );
    if( children != NULL )
        XFree( children );
    }
    return ret;
}

void EmbedWin::embedCurrentWindow()
{
    Q_ASSERT(m_container->parentWidget() != 0);

    Window root;
    Window child;
    uint mask;
    int rootX, rootY, winX, winY;

    XGrabServer( QX11Info::display() );
    XQueryPointer( QX11Info::display(), QX11Info::appRootWindow(), &root, &child,
           &rootX, &rootY, &winX, &winY, &mask );

    if( child == None )
        child = QX11Info::appRootWindow();

    Window real_child = findRealWindow( child );
    if( real_child != None ) // test just in case
        child = real_child;

    XUngrabServer( QX11Info::display() );

    if(child == None || child == QX11Info::appRootWindow()
            || (view() && child == view()->effectiveWinId())) {
        m_label->setText(i18n("Cannot embed this window.")
                        + "<br /><br />" + i18n(WELCOME_MSG));
        return;
    }

    m_label->setText(i18n("Embedding client..."));
    m_window = child;
    m_container->show();
    m_container->embedClient(m_window);
}

void EmbedWin::containerError(QX11EmbedContainer::Error)
{
    m_window = 0;
    m_container->hide();
    m_reparenting = false;
    updateWidgetGeometry();
    m_label->setText(i18n("Error while embedding client")
                     + "<br /><br />" + i18n(WELCOME_MSG));
}

void EmbedWin::containerClientIsEmbedded()
{
    m_container->show();

    // Force resizeing of client
    QResizeEvent* event = new QResizeEvent(m_container->size(), m_container->size());
    QApplication::postEvent(m_container, event);

    m_label->setText(
        i18n("Client is embedded. If you see this text please report a bug."));
}

void EmbedWin::containerClientClosed()
{
    if(m_reparenting) {
        // window was discarded temporary to reparent the container
        m_reparenting = false;
        updateWidgetGeometry();
        m_container->embedClient(m_window);
    } else {
        if(m_window != 0)
            m_label->setText(i18n("Application have been closed")
                            + "<br /><br />" + i18n(WELCOME_MSG));
        else
            m_label->setText(i18n("Application have been detached")
                            + "<br /><br />" + i18n(WELCOME_MSG));
        m_window = 0;
        m_container->hide();
    }
}

const char* EmbedWin::WELCOME_MSG =
        "Please <a href=\"embed://embed\">select</a> a window to embed";

#include "embed-win.moc"

