/***************************************************************************
 *   Copyright (C) 2009 George Kiagiadakis <kiagiadakis.george@gmail.com>  *
 *   Copyright © 2009 Reza Fatahilah Shah <rshah0385@kireihana.com>        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/
#include "videoinputdevice.h"
#include <KDebug>
#include <gst/gst.h>
#include "libqtgstreamer/qgstdeclarations.h"
#include "libqtgstreamer/qgstpad.h"
#include "libqtgstreamer/qgstcaps.h"
#include "libqtgstreamer/qgststructure.h"
#include "libqtgstreamer/qgstelementfactory.h"
#include "libqtgstreamer/interfaces/qgstcolorbalance.h"
using namespace QtGstreamer;

class VideoInputDevicePrivate
{
    public:
        enum ColorProperty { Brightness, Contrast, Saturation, Hue, AutoWhiteBalance };
        
        QGstColorBalance *colorBalance;
        QGstElementPtr videoDevice;
        QList<QSize> resolutions;
        QHash< QString,QList<QSize> > frameRates;
        
        void setBalanceValue(ColorProperty colorProperty,int value);
        int getBalanceValue(ColorProperty colorProperty);
        QGstColorBalanceChannelPtr getBalanceChannel(ColorProperty colorProperty);
};

void VideoInputDevicePrivate::setBalanceValue(ColorProperty colorProperty,int value)
{
    int targetLevel=0;

    QGstColorBalanceChannelPtr balanceChannel = getBalanceChannel(colorProperty);
    if (balanceChannel) {
        targetLevel = (0.5+value * ((double) balanceChannel->max_value() -
        balanceChannel->min_value()) / 65535 + balanceChannel->min_value());

        colorBalance->setValue(balanceChannel,targetLevel);
        kDebug()<< balanceChannel->label() << ", scaled value=" << targetLevel << ", min=" << balanceChannel->min_value() << ",max=" << balanceChannel->max_value();
    }
    else {
        kDebug() << "Cannot set color balance channel";
    }
}

int VideoInputDevicePrivate::getBalanceValue(ColorProperty colorProperty)
{
    int currentLevel=65535/2; //default value

    QGstColorBalanceChannelPtr balanceChannel = getBalanceChannel(colorProperty);
    if (balanceChannel) {
        currentLevel = colorBalance->getValue(balanceChannel);
        currentLevel = 0.5 +
        ((double) currentLevel - balanceChannel->min_value()) * 65535 /
        ((double) balanceChannel->max_value() - balanceChannel->min_value());
        kDebug()<< balanceChannel->label() << ", value=" << currentLevel << ", min=" << balanceChannel->min_value() << ",max=" << balanceChannel->max_value();
    }
    else {
        kDebug() << "Cannot get color balance channel";
    } 

    return currentLevel;
}

QGstColorBalanceChannelPtr VideoInputDevicePrivate::getBalanceChannel(ColorProperty colorProperty)
{
    QList<QGstColorBalanceChannelPtr> channels;
    channels = colorBalance->listChannels();
    QList<QGstColorBalanceChannelPtr>::iterator channel = channels.begin();
    QString label;

    //white balance value only 0 and 1
    for(; channel != channels.end(); ++channel) {
        label = (*channel)->label().toLower();
        if ((colorProperty == Brightness && label == "brightness")
            || (colorProperty == Contrast && label == "contrast")
            || (colorProperty == Saturation && label == "saturation")
            || (colorProperty == Hue && label == "hue")
            || (colorProperty == AutoWhiteBalance && label == "white balance temperature, auto") ) {
            return QGstColorBalanceChannel::fromGstColorBalanceChannel((*channel)->peekNativeObject());
        }
    }

    return QGstColorBalanceChannelPtr();
}

////////////////
////////////////
VideoInputDevice::VideoInputDevice(GstElement *element)
    : QGstElement(element), d(new VideoInputDevicePrivate)
{
    d->colorBalance = NULL;
    d->videoDevice = QGstElement::fromGstElement(element);
}

VideoInputDevice::~VideoInputDevice()
{
    delete d->colorBalance;
    delete d;
}

VideoInputDevicePtr VideoInputDevice::newDevice(const QString &driver, const QString &devicePath)
{
    QGstElementPtr element = QGstElementFactory::make(driver.toLatin1());
    if ( element.isNull() ) {
        kWarning() << "Could not construct" << driver;
        return VideoInputDevicePtr();
    }

    if ( driver == "v4l2src" ) {
        element->setProperty("device", devicePath.toLatin1());
    }

    //HACK to detect if the device has a v4l1 or v4l2 driver
    if ( driver == "v4l2src" && !element->setState(QGstElement::Paused) ) {
        QGstElementPtr element_v4l1 = QGstElementFactory::make("v4lsrc");
        if ( element_v4l1.isNull() ) {
            kWarning() << "Could not construct v4lsrc";
            //fall back and return the v4l2src, although it is not going to work...
            return VideoInputDevicePtr(new VideoInputDevice(GST_ELEMENT(element->getNativeObject())));
        }
        element_v4l1->setProperty("device", devicePath.toLatin1());
        return VideoInputDevicePtr(new VideoInputDevice(GST_ELEMENT(element_v4l1->getNativeObject())));
    }
    return VideoInputDevicePtr(new VideoInputDevice(GST_ELEMENT(element->getNativeObject())));
}

bool VideoInputDevice::hasColorBalance()
{
    if (d->colorBalance) {
        return true;
    }
    else {
        if (QGstColorBalance::isColorBalance(d->videoDevice)) {
            d->colorBalance = new QGstColorBalance(QGstElement::fromGstElement(GST_ELEMENT(d->videoDevice->peekNativeObject())));
            return true;
        }
        kDebug() << "Element does not have color balance interface";
        return false;
    }
}

void VideoInputDevice::setBrightness(int level)
{
    d->setBalanceValue(VideoInputDevicePrivate::Brightness,level);
}

int VideoInputDevice::brightness()
{
    return d->getBalanceValue(VideoInputDevicePrivate::Brightness);
}

void VideoInputDevice::setContrast(int level)
{
    d->setBalanceValue(VideoInputDevicePrivate::Contrast,level);
}

int VideoInputDevice::contrast()
{
    return d->getBalanceValue(VideoInputDevicePrivate::Contrast);
}

void VideoInputDevice::setSaturation(int level)
{
    d->setBalanceValue(VideoInputDevicePrivate::Saturation,level);
}

int VideoInputDevice::saturation()
{
    return d->getBalanceValue(VideoInputDevicePrivate::Saturation);
}

void VideoInputDevice::setHue(int level)
{
    d->setBalanceValue(VideoInputDevicePrivate::Hue,level);
}

int VideoInputDevice::hue()
{
    return d->getBalanceValue(VideoInputDevicePrivate::Hue);
}

void VideoInputDevice::queryDeviceData()
{
    this->setState(QtGstreamer::QGstElement::Paused);
    QGstPadPtr pad = this->getStaticPad("src");
    QGstCapsPtr caps = pad->getCaps();
    QGstStructure *capsStructure;
    
    int totalCaps = caps->getSize();
    
    kDebug() << "Total capabilities size=" << totalCaps;

    int width=-1,height=-1;
    int width2=0,height2=0;
    int wMark=-1,hMark=-1;
    bool okToAdd = true;
    bool isRange = false;

    QtGstreamer::QGstValue resolutionValue;

    for (int ii= 0; ii < totalCaps; ii++) {
        okToAdd = true;
        isRange = false;
        //TODO: find best way to solve below
        capsStructure = caps->getStructure(ii);
        if (!capsStructure->hasName("video/x-raw-yuv") && !capsStructure->hasName("video/x-raw-rgb")) {
            continue;
        }
        //check if value is range or not
        resolutionValue = capsStructure->getValue("width");
        if (resolutionValue.holdsIntRange()) {
            width = resolutionValue.getIntRangeMin();
            width2 = resolutionValue.getIntRangeMax();

            resolutionValue = capsStructure->getValue("height");
            height = resolutionValue.getIntRangeMin();
            height2 = resolutionValue.getIntRangeMax();
            isRange = true;                
        }
        else if (resolutionValue.holdsInt()) {
            width = resolutionValue.getInt();

            resolutionValue = capsStructure->getValue("height");
            height = resolutionValue.getInt();
        }
        else {
            okToAdd = false;
            kDebug() << "Unsupported video resolution value";
        }

        if (wMark == width && hMark == height) {
            break;
        }
        if (wMark == -1) {
            wMark = width;
            hMark = height;
        }
        if (okToAdd) {
            kDebug() << "w=" << width << " h=" <<height;
            d->resolutions.append(QSize(width,height));
            QList<QSize> frameRates = queryFramerates(*capsStructure);
            if (frameRates.size()>0)
            {
                QString key = QString("%1%2").arg(width).arg(height);
                d->frameRates.insert(key,frameRates);
            }
            if (isRange) {
                kDebug() << "w=" << width2 << " h=" <<height2;
                d->resolutions.append(QSize(width2,height2));
            }
        }
    }
 
    this->setState(QtGstreamer::QGstElement::Null);
}

QList<QSize> VideoInputDevice::queryFramerates(const QtGstreamer::QGstStructure &capsStructure)
{
    QList<QSize> frameRates;
    int numerator,denominator;
    QGstValue frameRateValue;

    frameRateValue = capsStructure.getValue("framerate");;

    if (frameRateValue.holdsList()) {
        int totalItems = frameRateValue.listGetSize();
        QGstValue value;
        for(int ii=0;ii<totalItems;++ii) {
            value = frameRateValue.listGetValue(ii);
            numerator = value.getFractionNumerator();
            denominator = value.getFractionDenominator();
            frameRates.append(QSize(numerator,denominator));
        }
    }
    else if (frameRateValue.holdsFraction()) {
        numerator = frameRateValue.getFractionNumerator();
        denominator = frameRateValue.getFractionDenominator();
        kDebug() << "Num/denom =" << numerator<<"/"<<denominator;

        frameRates.append(QSize(numerator,denominator));
    }
    else if (frameRateValue.holdsFractionRange()) {
        QGstValue rangeMin,rangeMax;
        int numeratorMin,denominatorMin;
        int numeratorMax,denominatorMax;

        rangeMin = frameRateValue.getFractionRangeMin();
        rangeMax = frameRateValue.getFractionRangeMax();
        numeratorMin = rangeMin.getFractionNumerator();
        denominatorMin = rangeMin.getFractionDenominator();
        numeratorMax = rangeMax.getFractionNumerator();
        denominatorMax = rangeMax.getFractionDenominator();
        kDebug() << "Num/denom Min=" << numeratorMin<<"/"<<denominatorMin;
        kDebug() << "Num/denom Max=" << numeratorMax<<"/"<<denominatorMax;

        frameRates.append(QSize(numeratorMin,denominatorMin));
        frameRates.append(QSize(numeratorMax,denominatorMax));
    }
    else {
        kDebug() << "unsupported FPS value";
    }
    return frameRates;
}

QList<QSize> VideoInputDevice::getSupportedResolution() const
{
    return d->resolutions;
}

QList<QSize> VideoInputDevice::getSupportedFrameRate(const QSize &resolution) const
{
    QString key = QString("%1%2").arg(resolution.width()).arg(resolution.height());
    return d->frameRates[key];    
}

#include "videoinputdevice.moc"