/*
 * This file is part of the KDE project
 * Copyright (C) 2009 Eva Brucherseifer <eva.brucherseifer@basyskom.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */


#include "configview.h"
#include "killswitchmanager.h"
#include "killswitchdevice.h"

#include <KAboutData>
#include <KConfig>
#include <KDebug>
#include <KDialog>
#include <KIconLoader>
#include <KLocale>
#include <kdemacros.h>

#include <QLabel>
#include <QVBoxLayout>
#include <QListWidgetItem>
#include <QToolButton>


class SwitchItem : public QListWidgetItem
{
  public:
    SwitchItem( const QString &name, QListWidget *parent )
      : QListWidgetItem( parent ), mInfo( name )
    {
      //setIcon( 0, KIcon( mInfo.icon() ) );
      setText( mInfo );
      //setToolTip( 0, mInfo.comment() );
      setFlags( Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsUserCheckable );
    }

    QString name() const
    {
      return mInfo;
    }

    virtual QString text( int column ) const
    {
/*
      if ( column == 0 ) {
        return mInfo.name();
      } else if ( column == 1 ) {
        return mInfo.comment();
      } else {
        return QString();
      }
*/
      return mInfo;
    }

  private:
    QString mInfo;
};

SwitchSelector::SwitchSelector( QWidget *parent )
  : QListWidget( parent )
{
  setSelectionMode( QAbstractItemView::SingleSelection );
  setSelectionBehavior( QAbstractItemView::SelectRows );
  //setColumnCount( 1 );
  //setHeaderLabel( i18nc( "@title:column killswitch name", "Radio Frequency Switches" ) );
}

SwitchSelector::~SwitchSelector()
{
}


ConfigView::ConfigView( QWidget *parent, KillSwitchManager *mgr )
  : QWidget( parent ), p_mgr(mgr)
{
  QVBoxLayout *vlayout = new QVBoxLayout( this );
  vlayout->setSpacing( KDialog::spacingHint() );
  //layout->setMargin( 0 );

  QLabel *label =
    new QLabel( i18n( "Select the kill switches to show in the applet." ), this );
  vlayout->addWidget( label );

  QHBoxLayout *layout = new QHBoxLayout( this );
  layout->setSpacing( KDialog::spacingHint() );
  vlayout->addLayout( layout );

  m_switchSelector = new SwitchSelector( this );
  layout->addWidget( m_switchSelector );

  layout->setStretchFactor( m_switchSelector, 1 );

  QVBoxLayout *orderLayout = new QVBoxLayout( this );
  layout->addLayout(orderLayout);
  //orderLayout->insertStretch(-1);

  m_upButton = new QToolButton ( this );
  m_upButton->setIcon(KIcon("go-up"));
  m_upButton->setEnabled( false ); // b/c no item is selected yet
  orderLayout->addWidget(m_upButton);

  m_downButton = new QToolButton ( this );
  m_downButton->setIcon(KIcon("go-down"));
  m_downButton->setEnabled( false ); // b/c no item is selected yet
  orderLayout->addWidget(m_downButton);   
  orderLayout->insertStretch(-1);

  load();
//  connect( m_switchSelector, SIGNAL(itemChanged(QListWidgetItem*,int)),
//           this, SLOT(changed()) );

 connect( m_switchSelector, SIGNAL( currentItemChanged( QListWidgetItem*, QListWidgetItem* ) ),
           this, SLOT( slotSelectionChanged( QListWidgetItem* ) ) );
//  connect( m_switchSelector, SIGNAL( itemClicked( QListWidgetItem* ) ),
//           this, SLOT( slotItemClicked( QListWidgetItem* ) ) );

  connect( m_upButton, SIGNAL( clicked() ), this, SLOT( slotMoveUp() ) );
  connect( m_downButton, SIGNAL( clicked() ), this, SLOT( slotMoveDown() ) );
}



void ConfigView::slotSelectionChanged( QListWidgetItem *item )
{
//  bool state = ( item != 0 );
  m_downButton->setEnabled( item && (m_switchSelector->row( item ) != (m_switchSelector->count() - 1)) );
  m_upButton->setEnabled( item && (m_switchSelector->row( item ) != 0) );
}


static void swapItems( SwitchItem *item, SwitchItem *other )
{
  SwitchItem tmp = *item;
  *item = *other;
  *other = tmp;
}


void ConfigView::slotMoveUp()
{
  const QList<QListWidgetItem*> selectedItems = m_switchSelector->selectedItems();
  if ( selectedItems.count() == 0 )
    return;

  SwitchItem *item = static_cast<SwitchItem *>( m_switchSelector->selectedItems().first() );
  if ( !item )
    return;

  SwitchItem *above = static_cast<SwitchItem *>( m_switchSelector->item( m_switchSelector->row( item )- 1 ) );
  if ( !above )
    return;

  swapItems( item, above );

  m_switchSelector->setCurrentItem( above );
  above->setSelected( true );
}

void ConfigView::slotMoveDown()
{
  const QList<QListWidgetItem*> selectedItems = m_switchSelector->selectedItems();
  if ( selectedItems.count() == 0 )
    return;

  SwitchItem *item = static_cast<SwitchItem *>( m_switchSelector->selectedItems().first() );
  if ( !item )
    return;

  SwitchItem *below = static_cast<SwitchItem *>( m_switchSelector->item( m_switchSelector->row( item )+ 1 ) );
  if ( !below )
    return;

  swapItems( item, below );

  m_switchSelector->setCurrentItem( below );
  below->setSelected( true );
}



void ConfigView::configAccepted()
{
  save();
}


void ConfigView::load()
{
  QStringList inactiveSwitches;
  QStringList switchOrder;

  KSharedConfigPtr configPtr = KSharedConfig::openConfig( "plasma_applet_killswitchrc" );
  KConfigGroup grp( configPtr, QString() );
  if ( grp.hasKey( "InactiveSwitches" ) ) 
  {
    inactiveSwitches = grp.readEntry( "InactiveSwitches", QStringList() );
  }
  if ( grp.hasKey( "SwitchOrder" ) ) 
  {
    switchOrder = grp.readEntry( "SwitchOrder", QStringList() );
  }

  m_switchSelector->clear();

  QStringList allSwitches;
  QStringList rfkill_uids;
  rfkill_uids = p_mgr->devicesFromQuery("");    
  KillSwitchDevice* device;

  foreach (const QString &uid, rfkill_uids) {
        device = dynamic_cast<KillSwitchDevice*>(p_mgr->createDevice(uid));
        if (device->type() != "unknown")
           allSwitches << device->name();
  }

  foreach (const QString &uid, allSwitches) {
        if (!switchOrder.contains(uid))
             switchOrder.append(uid);
  }

  foreach (const QString &uid, switchOrder) {
        if (!allSwitches.contains(uid))
             switchOrder.removeAll(uid);
  }


  QStringList::Iterator it;
  for ( it = switchOrder.begin(); it != switchOrder.end(); ++it ) {
      SwitchItem* item = new SwitchItem(*it, m_switchSelector);
      if ( inactiveSwitches.contains( *it ) ) {
        item->setCheckState( Qt::Unchecked );
      } else {
        item->setCheckState( Qt::Checked );
      }
  }
}

void ConfigView::save()
{
  QStringList inactiveSwitches;
  QStringList switchOrder;

  for ( int i = 0; i < m_switchSelector->count(); ++i ) 
  {
    SwitchItem *item = static_cast<SwitchItem *>( m_switchSelector->item(i) );
    switchOrder.append( item->name() );
    if ( item->checkState() == Qt::Unchecked ) 
    {
      inactiveSwitches.append( item->name() );
    }
  }

  KSharedConfigPtr configPtr = KSharedConfig::openConfig( "plasma_applet_killswitchrc" );
  KConfigGroup grp( configPtr, QString() );
  grp.writeEntry( "InactiveSwitches", inactiveSwitches );
  grp.writeEntry( "SwitchOrder", switchOrder );
}

#include "configview.moc"
