/*
    This file is part of the Program Menu plasmoid, part of the KDE project.

    Copyright (c) 2002 Siegfried Nijssen <snijssen@liacs.nl>
    Copyright (c) 2003 Lubos Lunak <l.lunak@suse.cz>
    Copyright 2008 Friedrich W. H. Kossebau <kossebau@kde.org>

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
    AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
    AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "menuplasmoid.h"

// Plasma
// KDE
#include <KGlobalSettings>
#include <KWindowSystem>
#include <kmanagerselection.h>
#include <netwm.h>
#include <KDebug>
// Qt
#include <QtGui/QX11Info>
#include <QtGui/QGraphicsProxyWidget>
#include <QtGui/QLabel>
// plasmoid
#include "menubarembedcontainer.h" // last due to X11 header in there, sigh...


static const long SUPPORTED_WINDOW_TYPES =
    NET::NormalMask | NET::DesktopMask | NET::DockMask
    | NET::ToolbarMask | NET::MenuMask | NET::DialogMask | NET::OverrideMask
    | NET::TopMenuMask | NET::UtilityMask | NET::SplashMask;


// If there's no menu for the window, try finding menu for its mainwindow
// (where the window's WM_TRANSIENT_FOR property points).
// If the window is modal (_NET_WM_STATE_MODAL), stop.
static inline WId tryTransientForNonModal( WId windowId )
{
    const KWindowInfo info = KWindowSystem::windowInfo( windowId, NET::WMState );
    if( info.state() & NET::Modal )
        return None;

    WId transientFor = KWindowSystem::transientFor( windowId );
    if( transientFor == QX11Info::appRootWindow() )
        transientFor = None;

    return transientFor;
}


MenuPlasmoid::MenuPlasmoid( QObject* parent, const QVariantList &arguments )
 : Plasma::Applet( parent, arguments ),
   mActiveMenuContainer( 0 ),
   mDesktopMenuContainer( 0 ),
   mSelectionOwner( 0 ),
   mSelectionWatcher( 0 ),
   mUseDesktopMenu( false ),
   mTopEdgeOffset( 0 )
{
    mMenuContainerProxyWidget = new QGraphicsProxyWidget( this );
    setHasConfigurationInterface( false );
    setAspectRatioMode(Plasma::IgnoreAspectRatio);

mLabel = new QLabel("Hello!");
mLabel->resize( 100, 50 );
mMenuContainerProxyWidget->setWidget( mLabel );

    // toolbarAppearanceChanged(int) is sent when changing macstyle
    connect( KGlobalSettings::self(), SIGNAL(toolbarAppearanceChanged( int )), SLOT(readSettings()) );
    claimSelection();
    readSettings();
    updateTopEdgeOffset();
}


bool MenuPlasmoid::isDisabled() const
{
    Q_ASSERT( ( mSelectionOwner == 0 && mSelectionWatcher != 0 )
            || ( mSelectionOwner != 0 && mSelectionWatcher == 0 ));
    return ( mSelectionOwner == 0 );
}


void MenuPlasmoid::activateMenu( MenuBarEmbedContainer *menuContainer )
{
    if( menuContainer == mActiveMenuContainer )
        return;

kDebug() << "activate:" << menuContainer;
    if( mActiveMenuContainer != 0 )
        mActiveMenuContainer->hide();

    mActiveMenuContainer = menuContainer;
    mMenuContainerProxyWidget->setWidget( menuContainer );

    if( mActiveMenuContainer != 0 )
    {
//TODO        mActiveMenuContainer->setMinimumSize( width(), height() + mTopEdgeOffset );
        setPreferredSize( mActiveMenuContainer->size() );
//         mMenuContainerProxyWidget->resize( effectiveSizeHint(Qt::PreferredSize) );

        mActiveMenuContainer->show();
    }
//     emit updateLayout();
}


void MenuPlasmoid::registerPotentialMenu( WId windowId )
{
    const NETWinInfo info( QX11Info::display(), windowId, QX11Info::appRootWindow(), NET::WMWindowType );
kDebug()<<info.windowType(SUPPORTED_WINDOW_TYPES);
    if( info.windowType(SUPPORTED_WINDOW_TYPES) != NET::TopMenu )
{
kDebug() << "not a topmenu:" << windowId;
        return;
}
kDebug() << "a topmenu:" << windowId;
    WId mainWindowId = KWindowSystem::transientFor( windowId );
    if( mainWindowId == None )
{
        return;
kDebug() << "without a mainwindow";
}
kDebug() << "with a mainwindow"<<mainWindowId;
    MenuBarEmbedContainer *menuContainer = new MenuBarEmbedContainer( mainWindowId );
    menuContainer->hide();
    menuContainer->move( 0, -mTopEdgeOffset );
    menuContainer->resize( menuContainer->width(), size().height() + mTopEdgeOffset );
    connect( menuContainer, SIGNAL(menuLost( MenuBarEmbedContainer* )), SLOT(onMenuLost( MenuBarEmbedContainer* )) );
    connect( menuContainer, SIGNAL(sizeChanged( MenuBarEmbedContainer* )), SLOT(onSizeChanged( MenuBarEmbedContainer* )) );

    menuContainer->embedClient( windowId );

    mMenuContainers.append( menuContainer );

    bool mainWindowIsDesktop;
    if( mainWindowId == QX11Info::appRootWindow() )
    {
        mainWindowIsDesktop = true;
    }
    else
    {
        const KWindowInfo info2 = KWindowSystem::windowInfo( mainWindowId, NET::WMWindowType );
        mainWindowIsDesktop = ( info2.windowType(SUPPORTED_WINDOW_TYPES) == NET::Desktop );
    }
    if( mainWindowIsDesktop )
        mDesktopMenuContainer = menuContainer;

    // in case the app mapped its menu after its mainwindow, check which menu should be shown
    onActiveWindowChanged( KWindowSystem::activeWindow() );
}


// - if the active window has its topmenu -> show the menu
// - if desktop menu is enabled (i.e. explicitly in kdesktop) :
//   - show it
//   - otherwise show nothing
void MenuPlasmoid::onActiveWindowChanged( WId newActiveWindowId )
{
mLabel->setText( QString::number(newActiveWindowId) );
kDebug() << "active:" << newActiveWindowId;

    // if the active window has a topmenu -> show the menu
    for( WId windowId = newActiveWindowId;
         windowId != None;
         windowId = ::tryTransientForNonModal(windowId) )
    {
        foreach( MenuBarEmbedContainer *menuContainer, mMenuContainers )
        {
            if( windowId == menuContainer->mainWindowId() )
            {
                activateMenu( menuContainer );
                return;
            }
        }
    }

kDebug() << "no menu for active";
    // No menu for active window found
    // if desktop menu (in kdesktoprc) is enabled,
    // use kdesktop's menu instead of none.
    bool tryDesktopMenu = mUseDesktopMenu;
    // perhaps the active is still the desktop?
    if( !tryDesktopMenu && newActiveWindowId != None )
    {
        const KWindowInfo info = KWindowSystem::windowInfo( newActiveWindowId, NET::WMWindowType );
        const bool activeWindowIsDesktop = ( info.windowType(SUPPORTED_WINDOW_TYPES) == NET::Desktop );
        tryDesktopMenu = activeWindowIsDesktop;
    }

    MenuBarEmbedContainer *desktopMenuContainer = ( tryDesktopMenu ) ? mDesktopMenuContainer : 0;

    activateMenu( desktopMenuContainer );
}


void MenuPlasmoid::onSizeChanged( MenuBarEmbedContainer *menuContainer )
{
Q_UNUSED( menuContainer )
//     if( menuContainer == mActiveMenuContainer )
//         emit updateLayout();
}


void MenuPlasmoid::onMenuLost( MenuBarEmbedContainer *lostMenuContainer )
{
    for( QList< MenuBarEmbedContainer* >::Iterator it = mMenuContainers.begin();
         it != mMenuContainers.end();
         ++it )
    {
        if( *it == lostMenuContainer )
        {
            mMenuContainers.erase( it );
            lostMenuContainer->deleteLater();

kDebug() << "deleting:" << lostMenuContainer->mainWindowId();
            if( lostMenuContainer == mDesktopMenuContainer )
                mDesktopMenuContainer = 0;

            if( lostMenuContainer == mActiveMenuContainer )
            {
                mActiveMenuContainer = 0;
                // trigger selecting new active menu
                onActiveWindowChanged( KWindowSystem::activeWindow() );
            }
            break;
        }
    }
}

#if 0
void MenuPlasmoid::positionChange( Position )
{
    updateTopEdgeOffset();
}
#endif
// Detect mouse movement at the top screen edge also if the menubar
// has a popup open - in such case, Qt has a grab, and this avoids
// Kicker's FittsLawFrame. Therefore move the menubar a bit higher,
// so that it actually is positioned exactly at the screen edge
// (i.e. at a negative y coordinate within this applet, due to
// Kicker's frame).
void MenuPlasmoid::updateTopEdgeOffset()
{
#if 0
    QPoint p = topLevelWidget()->mapToGlobal( QPoint( 0, 0 ));
    if( p.y() <= 2 ) // 2 = work also when running in appletproxy
        mTopEdgeOffset = mapToGlobal( QPoint( 0, 0 )).y() - p.y();
    else
        mTopEdgeOffset = 0;
    if( mActiveMenuContainer != 0 )
        mActiveMenuContainer->move( mActiveMenuContainer->x(), -mTopEdgeOffset );
}

void MenuPlasmoid::paletteChange(const QPalette & /* oldPalette */)
{
    if( mActiveMenuContainer != 0 )
    {
        mActiveMenuContainer->hide();
        mActiveMenuContainer->show();
    }
#endif
}

void MenuPlasmoid::claimSelection()
{
kDebug() << "claiming...";
    Q_ASSERT( mSelectionOwner == 0 );
    mSelectionOwner = new KSelectionOwner( MenuBarEmbedContainer::makeSelectionAtom(), DefaultScreen(QX11Info::display()) );
    // force taking the selection, but don't kill previous owner
    if( mSelectionOwner->claim(true,false) )
    {
        delete mSelectionWatcher;
        mSelectionWatcher = 0;
        connect( mSelectionOwner, SIGNAL(lostOwnership()), SLOT(onLostSelection()) );

        KWindowSystem *windowSystem = KWindowSystem::self();
        connect( windowSystem, SIGNAL(windowAdded( WId )), SLOT(registerPotentialMenu( WId )) );
        connect( windowSystem, SIGNAL(activeWindowChanged( WId )), SLOT(onActiveWindowChanged( WId )) );

// TODO: if claimSelection() is called from the constructor, calling windows() on KWindowSystem inits it
// and enforces the build up of the window list which will emit a windowAdded()
// so we try to register the menus twice in that case
// especially unneeded if a toplevel menu has not yet been embedded:
// there will be a second try which will fail, well, works, but calls for design improvements
        const QList< WId > &windowIds = windowSystem->windows();
kDebug() << "successfully claimed, adding windows:"<<windowIds.size();
        foreach( WId windowId, windowIds )
            registerPotentialMenu( windowId );

        onActiveWindowChanged( windowSystem->activeWindow());
    }
    else
    {
        onLostSelection();
    }
}

void MenuPlasmoid::onLostSelection()
{
    if( mSelectionOwner == 0 )
        return;

kDebug() << "oh dear, lost selection";
    // delete all MenuBarEmbedContainer's = release all mMenuContainers
    qDeleteAll( mMenuContainers );
    mMenuContainers.clear();

    mActiveMenuContainer = 0;
    mDesktopMenuContainer = 0;
    if( mSelectionWatcher == 0 )
    {
        mSelectionWatcher = new KSelectionWatcher( MenuBarEmbedContainer::makeSelectionAtom(), DefaultScreen(QX11Info::display()) );
        connect( mSelectionWatcher, SIGNAL(lostOwner()), SLOT(claimSelection()) );
    }

    mSelectionOwner->deleteLater();
    mSelectionOwner = 0;
    // mSelectionWatcher stays
}

void MenuPlasmoid::readSettings()
{
    const KConfig desktopConfig( "kdesktoprc" );
    const KConfigGroup menuBarConfigGroup( KGlobal::config(), "Menubar" );
    mUseDesktopMenu = menuBarConfigGroup.readEntry( "ShowMenubar", false );

    const KConfigGroup kdeConfigGroup( KGlobal::config(), "KDE" );
    const bool useCentralMenuBar = kdeConfigGroup.readEntry( "macStyle", false );

    if( useCentralMenuBar || mUseDesktopMenu )
        mMenuContainerProxyWidget->setToolTip( QString() );
    else
        mMenuContainerProxyWidget->setToolTip( i18n(
            "You do not appear to have enabled the standalone menubar; "
            "enable it in the Behavior control module for desktop." ));
    if( !isDisabled() && mActiveMenuContainer == 0 )
        onActiveWindowChanged( KWindowSystem::activeWindow() ); //enforce mUseDesktopMenu
}

#if 0
int MenuPlasmoid::widthForHeight( int ) const
    {
    if (mActiveMenuContainer)
        return mActiveMenuContainer->width();
    return 0; // we're stretch applet
    }

int MenuPlasmoid::heightForWidth( int ) const
{
    // *shrug* running this applet in vertical mode is a bad idea anyway
    return 50;
}
#endif

MenuPlasmoid::~MenuPlasmoid()
{
    onLostSelection(); // release all mMenuContainers before really loosing the selection

    delete mSelectionOwner;
    delete mSelectionWatcher;
}

K_EXPORT_PLASMA_APPLET( menubar, MenuPlasmoid )
