/***************************************************************************
 *   Copyright (C) 2009 by Kai Dombrowe <just89@gmx.de>                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/


// own
#include "game.h"
#include "gamelist.h"

// Qt
#include <QFile>
#include <QDir>
#include <QTextStream>

// KDE
#include <KDesktopFile>
#include <KStandardDirs>
#include <KRun>
#include <KMessageBox>
#include <klocalizedstring.h>
#include <KUrl>
#include <KDialog>


Game::Game(QObject *parent, const QVariantMap &data, GameList *list)
    : QObject(parent),
    m_data(data),
    m_list(list)
{

    connect(this, SIGNAL(changed()), this, SLOT(save()));

    const bool newGame = m_data.contains("NewGame");
    const bool createFile = m_data["DesktopFile"].toString().isEmpty();
    const bool copy = m_data.contains("Copy");
    
    if (newGame) {
        return;
    }
    
    if (createFile) {
        createDesktopFile();
    }
    
    init();
    
    if (copy) {
        copyData();
    }
        
}


Game::~Game()
{


}

QString Game::title() const
{

    return m_data["Title"].toString();

}


QString Game::description() const
{

    return m_data["Description"].toString();

}


QString Game::icon() const
{

    return m_data["Icon"].toString();

}


int Game::rating() const
{

    return m_data["Rating"].toInt();

}


QString Game::desktopFile() const
{

    return m_data["DesktopFile"].toString();

}


QString Game::exec() const
{

    return m_data["Exec"].toString();

}


bool Game::startX() const
{
    
#ifdef Q_WS_X11 
    return m_data["StartX"].toBool();
#else
    return false;
#endif
    
}


void Game::moved()
{

    emit changed();

}


bool Game::operator<(const Game &other) const
{

    switch (m_list->sort()) {
        case GameList::SortByTitle: return title() < other.title(); 
        case GameList::SortByDescription:  return description() < other.description();
#ifdef NEPOMUK
        case GameList::SortByRating: return rating() > other.rating();
#endif
        case GameList::NoSort:
        default: return true;
     }
    
}


bool Game::lessThan(const Game* g1, const Game* g2)
{

    return *g1 < *g2;

}


void Game::setTitle(const QString &title)
{

    if (m_data["Title"].toString() != title) {
        m_data["Title"] = title;
        emit changed();
    }
    
}


void Game::setDescription(const QString &description)
{

    if (m_data["Description"].toString() != description) {
        m_data["Description"] = description;
        emit changed();
    }

}


void Game::setIcon(const QString &icon)
{

    if (m_data["Icon"].toString() != icon) {
        m_data["Icon"] = icon;
        emit changed();
    }
    
}


void Game::setRating(const int &rating)
{

    if (m_data["Rating"].toInt() != rating) {
        m_data["Rating"] = rating;
        emit changed();
    }

}


void Game::setDesktopFile(const QString &file)
{

    if (m_data["DesktopFile"].toString() != file) {
        m_data["DesktopFile"] = file;
        emit changed();
    }
    
}
    

void Game::setExec(const QString &exec)
{

    if (m_data["Exec"].toString() != exec) {
        m_data["Exec"] = exec;
        emit changed();
    }
    
}


void Game::setStartX(const bool &start)
{
 
    if (m_data["StartX"].toBool() != start) {
        m_data["StartX"] = start;
        emit changed();
    }

}


void Game::start()
{
    
    if (!startX()) {
        new KRun(desktopFile(), 0);       
        return;
    }
    
#ifdef Q_WS_X11 
    const QString home = QDir::homePath();

    //xbin + xinitrc
    const KUrl xbin = KStandardDirs::findExe("X");

    if (!QFile(xbin.toLocalFile()).exists()) {                                          
        KMessageBox::sorry(0, i18n("Cannot find executable file: X"), QString(), 0);                                                                       
        return;
    }                                                                                                     

    QFile rc((home + "/.xinitrc_pgame"));
    if (rc.exists()) {
        rc.remove();                               
    }
        
    if (!rc.open(QIODevice::WriteOnly|QIODevice::Text)) { 
        KMessageBox::sorry( 0, i18n("Cannot open \"%1\".", (home+"/.xinitrc_pgame")), 
                            QString(), 0);                     
        return;
    }                                                                                               

    QTextStream out(&rc);
    out << QString("exec %1").arg(exec()) << "\n";

    
    //xnumber
    KUrl tmp  = QString(getenv("TMPDIR"));
    if (!QDir(tmp.toLocalFile()).exists() || tmp.isEmpty()) {
        tmp = QString("/tmp");
    }                                 
                                

    if (!QDir(tmp.toLocalFile()).exists()) {                                        
        KMessageBox::sorry(0, i18n("Cannot find tmp."), QString(), 0);
        return;
    }                                                                                                     


    int xnumber = 0;
    while (true || xnumber < 100) {
        if (QFile((tmp.toLocalFile()+QString("/.X%1-lock").arg(xnumber))).exists()) {
            xnumber++;
        } else {
            break;                                                                      
        }
    }
        
    //start x
    const QString start = QString("xinit %1/.xinitrc_pgame -- %3 :%2 -br").arg(home).arg(
        xnumber).arg(xbin.toLocalFile());                                                                         

    KRun::runCommand(start, 0);   
#endif

}


void Game::edit()
{

    KDialog *dialog = new KDialog;
    dialog->setAttribute(Qt::WA_DeleteOnClose);
    
    QWidget *widget = new QWidget(dialog);
    ui_gameDialog.setupUi(widget);

    dialog->setMainWidget(widget);
        
    connect(dialog, SIGNAL(okClicked()), this, SLOT(editFinished()));
    connect(dialog, SIGNAL(okClicked()), dialog, SLOT(close()));
    
    ui_gameDialog.titleEdit->setText(title());
    ui_gameDialog.descriptionEdit->setText(description());
    ui_gameDialog.iconButton->setIcon(icon());
    ui_gameDialog.ratingSpinBox->setValue(rating());
    ui_gameDialog.commandEdit->setText(exec());
#ifdef Q_WS_X11 
    ui_gameDialog.startXCheck->setChecked(startX());
#else
    ui_gameDialog.startXCheck->hide();
#endif

    dialog->resize(400, 400);
    dialog->show();

}


void Game::remove()
{

    QFile desktop(desktopFile());
    QString path = gameDir();
    path += KUrl(desktop.fileName()).fileName().toLower();
    
    if (path != QFileInfo(desktop).absoluteFilePath()) {
        kDebug() << "wrong path...";
        return;
    }
    
    if (!desktop.remove()) {
        kDebug() << "remove failed:" << path << desktop.error() << desktop.errorString();
    }
    
    disconnect(this);
    
    emit changed(); // save gamelist
    
}


void Game::init()
{

    kDebug() << "init" << desktopFile();
    
    KConfig cfg(desktopFile());
    KConfigGroup desktopCfg(&cfg, "Desktop Entry");
    
    KDesktopFile desktop(desktopFile());
    m_data["Title"] = desktop.readName();
    m_data["Description"] = desktop.readComment();
    m_data["Icon"] = desktop.readIcon();
#ifdef NEPOMUK
    m_data["Rating"] = desktopCfg.readEntry("Rating", 0);
#else
    m_data["Rating"] = 0;
#endif
    m_data["Exec"] = desktopCfg.readEntry("Exec", QString());
    m_data["StartX"] = desktopCfg.readEntry("StartX", false);
    
    emit changed();
    
}


void Game::createDesktopFile()
{

    QString path = gameDir();
    path += title().toLower()+".desktop";
    
    while (QFile(path).exists()) {
        path.insert(path.length()-8, "_");
    }
    
    m_data["DesktopFile"] = path;
    
    kDebug() << "create DesktopFile" << path;
    
    KConfig cfg(path);
    KConfigGroup desktop(&cfg, "Desktop Entry");
    desktop.writeEntry("Name", title());
    desktop.writeEntry("Comment", description());
    desktop.writeEntry("Icon", icon());
    desktop.writeEntry("Rating", rating());
    desktop.writeEntry("Exec", exec());
    desktop.writeEntry("StartX", startX());
    desktop.writeEntry("Categories", "Game;");
    desktop.writeEntry("Type", "Application");
    
}


void Game::editFinished()
{

    if (m_data.contains("NewGame")) {
        m_data.remove("NewGame");
        m_data["Title"] = ui_gameDialog.titleEdit->text();
        createDesktopFile();
        init();
    }
    
    setTitle(ui_gameDialog.titleEdit->text());
    setDescription(ui_gameDialog.descriptionEdit->text());
    setIcon(ui_gameDialog.iconButton->icon());
    setRating(ui_gameDialog.ratingSpinBox->value());
    setExec(ui_gameDialog.commandEdit->text());
    setStartX(ui_gameDialog.startXCheck->isChecked());
    
}


QString Game::gameDir() const
{

    const QString path = "pgame/"+m_data["Id"].toString()+"/games/";
    return KGlobal::dirs()->locateLocal("appdata", path, true);

}


void Game::copyData()
{

    init();
    createDesktopFile();

}


void Game::save()
{

    KConfig cfg(desktopFile());
    KConfigGroup desktop(&cfg, "Desktop Entry");
    desktop.writeEntry("Name", title());
    desktop.writeEntry("Comment", description());
    desktop.writeEntry("Icon", icon());
    desktop.writeEntry("Rating", rating());
    desktop.writeEntry("Exec", exec());
    desktop.writeEntry("StartX", startX());
    cfg.sync();
    
}



#include "game.moc"
