/***************************************************************************
 *   Copyright (C) 2009 by Björn Ruberg <bjoern@ruberg-wegener.de>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

#include "ScreenControl.h"
#include "outputframe.h"
#include "screenbutton.h"
#include <QGraphicsLinearLayout>
#include <QProcess>
#include <plasma/widgets/tabbar.h>
#include <KIcon>

ScreenControl::ScreenControl(QGraphicsWidget *parent) :
	QGraphicsWidget(parent)
{
    setMinimumSize(450,230);

    m_rows = new QGraphicsLinearLayout(this);
    m_rows->setOrientation( Qt::Vertical );

    setLayout(m_rows);

    m_tabbar = new Plasma::TabBar(this);
    m_tabbar->nativeWidget()->setIconSize(QSize(32,32));
    m_tabbar->nativeWidget()->setExpanding(true);
    m_tabbar->nativeWidget()->setMovable(true);
    m_tabbar->nativeWidget()->setSelectionBehaviorOnRemove(QTabBar::SelectPreviousTab);
    m_rows->addItem(m_tabbar);

    m_screen_row = new QGraphicsLinearLayout(m_rows);
    m_screen_row->setOrientation( Qt::Horizontal );
    m_rows->addItem( m_screen_row );

    m_screen_row->addItem(new ScreenButton(i18n("Clone"), this));
    m_screen_row->addItem(new ScreenButton(i18n("Left-to-Right"), this));
    m_screen_row->addItem(new ScreenButton(i18n("Right-to-Left"), this));
    m_screen_row->setContentsMargins(6,0,6,0);

    m_acceptIcon = new Plasma::IconWidget(KIcon("dialog-ok"),i18n("Accept"));
    m_rows->addItem(m_acceptIcon);

    init();

    connect( Kephal::Outputs::self(), SIGNAL( outputConnected( Kephal::Output*) ), this, SLOT( init() ));
    connect( Kephal::Outputs::self(), SIGNAL( outputDisconnected( Kephal::Output*) ), this, SLOT( init() ));

    connect( m_acceptIcon, SIGNAL( clicked() ), this, SLOT( acceptConfiguration() ));

}

ScreenControl::~ScreenControl()
{
}

void ScreenControl::init(){

    while( m_tabbar->count() > 0 ){
	m_tabbar->removeTab( 0 );
    }
    qDeleteAll(m_outputs);
    m_outputs.clear();

    QList<Kephal::Output *> outputs = Kephal::Outputs::self()->outputs();

    foreach(Kephal::Output* output, outputs){
	if(!output->isConnected()){
	    continue;
	}
	OutputFrame* frame = new OutputFrame ( output, this );
	m_tabbar->addTab(KIcon("preferences-desktop-display-randr"), output->id(), frame );
	m_outputs << frame;
    }

    if(m_outputs.count() > 1)
    {
	int i = m_screen_row->count();
	while( i > 0 )
	{
	    i--;
	    ((ScreenButton*)m_screen_row->itemAt(i))->setEnabled(true);
	}
	selectScreenMode();
    }
    else
    {
	int i = m_screen_row->count();
	while( i > 0 )
	{
	    i--;
	    ((ScreenButton*)m_screen_row->itemAt(i))->setEnabled(false);
	}
    }
}

void ScreenControl::enableScreenButtons()
{
    int i = m_screen_row->count();
    while( i > 0 )
    {
	i--;
	((ScreenButton*)m_screen_row->itemAt(i))->enable();
    }
}

void ScreenControl::disableScreenButtons()
{
    int i = m_screen_row->count();
    while( i > 0 )
    {
	i--;
	((ScreenButton*)m_screen_row->itemAt(i))->disable();
    }
}

void ScreenControl::acceptConfiguration()
{
    // adapt resolution, refresh rates and orientation
    foreach(OutputFrame* output, m_outputs)
    {
	output->acceptConfiguration();
    }

    // configure screen positions

    if(m_outputs.count()>1)
    {
	QString command;

	if(((ScreenButton*)(m_screen_row->itemAt(0)))->isEnabled())
	{
	    command = " --same-as ";
	}
	else if(((ScreenButton*)(m_screen_row->itemAt(1)))->isEnabled())
	{
	    command = " --left-of ";
	}
	else if(((ScreenButton*)(m_screen_row->itemAt(1)))->isEnabled())
	{
	    command = " --right-of ";
	}

	int i=1;

	while (i < m_outputs.count())
	{
	    QProcess::execute ( "xrandr --output "+ m_outputs[i-1]->output()->id() + command + m_outputs[i]->output()->id());
	    i++;
	}
    }

}

ScreenControl::Mode ScreenControl::detectScreenMode()
{
    if(m_outputs.count() > 1)
    {
	if(m_outputs[0]->output()->position() == m_outputs[1]->output()->position())
	{
	    return ScreenControl::clone;
	}
	else
	{
	    int x_1 = m_outputs[0]->output()->position().x();
	    int x_2 = m_outputs[1]->output()->position().x();
	    if(x_1 < x_2)
	    {
		return ScreenControl::leftOf;
	    }
	    else if(x_1 > x_2)
	    {
		return ScreenControl::rightOf;
	    }
	    else
	    {
		return ScreenControl::stacked;
	    }

	}
    }
    return ScreenControl::single;
}

void ScreenControl::selectScreenMode()
{
    disableScreenButtons();

    Mode mode = detectScreenMode();

    if(mode == ScreenControl::clone)
    {
	((ScreenButton*)m_screen_row->itemAt(0))->enable();
    }
    else if(mode == ScreenControl::leftOf)
    {
	((ScreenButton*)m_screen_row->itemAt(1))->enable();
    }
    else if(mode == ScreenControl::rightOf)
    {
	((ScreenButton*)m_screen_row->itemAt(2))->enable();
    }
}
