/*
    This file is part of the KDE project.

    Copyright 2005,2009 Friedrich W. H. Kossebau <kossebau@kde.org>
    Based upon kdmlib files.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License version 2 as published by the Free Software Foundation.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301  USA
*/

#include "klocalsessionmanager.h"

// applet
#include "kworkspace.h"
// KDE
#include <KLocale>
#include <KMessageBox>
// Qt
#include <QtDBus/QtDBus>
// X11
#include <X11/Xauth.h>
#include <X11/Xlib.h>
// C
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

// Windows
#ifdef Q_OS_WIN
#define _WIN32_WINNT 0x0500 // require NT 5.0 (win 2k pro)
#include <windows.h>
#endif // Q_OS_WIN

#include <KDebug>

enum KDMType { DMUnknown, NoDM, NewKDM, OldKDM, GDM };

class KDMData
{
  public:
    KDMData();

  public:
    KDMType type()                       const { return mDMType; }
    const char* displayName()            const { return mDisplayName; }
    const char* displayManagerSettings() const { return mDisplayManagerSettings; }

  protected:
    KDMType mDMType;
    const char* mDisplayName;
    const char* mDisplayManagerSettings;
};


inline KDMData::KDMData()
{
    if( (mDisplayName = ::getenv("DISPLAY")) == 0 )
        mDMType = NoDM;
    else if( (mDisplayManagerSettings = ::getenv( "DM_CONTROL" )) != 0 )
        mDMType = NewKDM;
    else if( (mDisplayManagerSettings = ::getenv( "XDM_MANAGED")) != 0  && mDisplayManagerSettings[0] == '/' )
        mDMType = OldKDM;
    else if( ::getenv("GDMSESSION") != 0 )
        mDMType = GDM;
    else
        mDMType = NoDM;
}


static KDMData DM;


static void doGdmAuthenticate();

class KDisplayManagerPipe
{
  public:
    KDisplayManagerPipe();
    ~KDisplayManagerPipe();

  public:
    /**
    * Send a command to the display manager
    * @param Command the command to execute.
    * @param Buffer the result buffer.
    * @return result:
    *  @li If true, the command was successfully executed.
    *   @p Buffer might contain addional results.
    *  @li If false and @p Buffer is empty, a communication error occurred
    *   (most probably KDM is not running).
    *  @li If false and @p Buffer is non-empty, it contains the error message
    *   from KDM.
    */
    bool send( const char* command, QByteArray& buffer );
    /**
     * Send a command to the display manager
     */
    bool send( const char* command );

  protected:
    int mPipe;
};

KDisplayManagerPipe::KDisplayManagerPipe()
  : mPipe( -1 )
{
    // connect to display manager
    switch( DM.type() )
    {
    case NewKDM:
    case GDM:
        mPipe = ::socket( PF_UNIX, SOCK_STREAM, 0 );
        // could create socket for connection?
        if( mPipe >= 0 )
        {
            struct sockaddr_un pipeSocket;
            pipeSocket.sun_family = AF_UNIX;

            if( DM.type() == GDM )
            {
                strcpy( pipeSocket.sun_path, "/var/run/gdm_socket" );
                if( ::connect(mPipe,(struct sockaddr*)&pipeSocket,sizeof( pipeSocket )) )
                {
                    strcpy( pipeSocket.sun_path, "/tmp/.gdm_socket" );
                    if( ::connect(mPipe,(struct sockaddr*)&pipeSocket,sizeof( pipeSocket )) )
                    {
                        ::close( mPipe );
                        mPipe = -1;
                        break;
                    }
                }
                doGdmAuthenticate();
            }
            else
            {
                // get end of display name
                const char* const displayName = DM.displayName();
                const char* n = ::strchr( displayName, ':' );
                if( n != 0 )
                    n = ::strchr( n, '.' );
                ::snprintf( pipeSocket.sun_path, sizeof(pipeSocket.sun_path),
                        "%s/dmctl-%.*s/socket", DM.displayManagerSettings(), n!=0 ? (int)(n-displayName) : 512, displayName );
                // cannot create connection?
                if( ::connect(mPipe,(struct sockaddr*)&pipeSocket,sizeof( pipeSocket )) != 0 )
                {
                    ::close( mPipe );
                    mPipe = -1;
                }
            }
        }
        break;
    case OldKDM:
        {
            QString pipeName( DM.displayManagerSettings() );
            pipeName.truncate( pipeName.indexOf( ',' ) );
            mPipe = ::open( pipeName.toLatin1(), O_WRONLY );
        }
        break;
    default:
        ;
    }
}


bool KDisplayManagerPipe::send( const char* command )
{
    QByteArray buffer;
    return send( command, buffer );
}

bool KDisplayManagerPipe::send( const char* command, QByteArray& buffer )
{
    bool result = false;

    int tl;
    if( mPipe < 0 )
        goto busted;

    tl = ::strlen( command );
    if( ::write(mPipe,command,tl) != tl )
    {
bust:
        ::close( mPipe );
        mPipe = -1;
busted:
        buffer.resize( 0 );
        return false;
    }
    if( DM.type() == OldKDM )
    {
        buffer.resize( 0 );
        return true;
    }

    int length = 0;
    for( ;; )
    {
        if( buffer.size() < 128 )
            buffer.resize( 128 );
        else if( buffer.size() < length * 2 )
            buffer.resize( length * 2 );

        tl = ::read( mPipe, buffer.data()+length, buffer.size()-length );

        if( tl <= 0 )
        {
            if( tl < 0 && errno == EINTR )
                continue;
            goto bust;
        }

        length += tl;
        if( buffer[length-1] == '\n' )
        {
            buffer[length-1] = 0;
            if( length > 2 && buffer[0] == 'o' && buffer[1] == 'k' && buffer[2] < (char)(32) )
                result = true;
            break;
        }
    }

    return result;
}


KDisplayManagerPipe::~KDisplayManagerPipe()
{
    if( mPipe >= 0 )
        close( mPipe );
}

static void doGdmAuthenticate()
{
    const char* displayName = DisplayString( QX11Info::display() );
    if( ! displayName )
    {
        displayName = ::getenv( "DISPLAY" );
        if( ! displayName )
            return;
    }

    const char* const displayNumber = strchr( displayName, ':' ) + 1;
    const char* const displaySubNumber = strchr( displayName, '.' );
    const int displayNumberLength = displaySubNumber ? displaySubNumber - displayNumber : strlen( displayNumber );

    /* XXX should do locking */
    FILE* fp;
    if( ! (fp = fopen( XauFileName(), "r" )) )
        return;

    Xauth* xau;
    while( (xau = XauReadAuth( fp )) )
    {
        if( xau->family == FamilyLocal &&
            xau->number_length == displayNumberLength &&
            ! memcmp( xau->number, displayNumber, displayNumberLength ) &&
            xau->data_length == 16 &&
            xau->name_length == 18 && !memcmp( xau->name, "MIT-MAGIC-COOKIE-1", 18 ) )
        {
            QByteArray command( "AUTH_LOCAL " );
            for( int i = 0; i < 16; i++ )
                command += QByteArray::number( (uchar)xau->data[i], 16 ).rightJustified( 2, '0');
            command += '\n';
            if( KDisplayManagerPipe().send(command) )
            {
                XauDisposeAuth( xau );
                break;
            }
        }
        XauDisposeAuth( xau );
    }

    fclose (fp);
}


KLocalSessionManager::KLocalSessionManager( QWidget* screen, int screenNumber )
  : mScreen( screen ),
    mScreenNumber( screenNumber ),
    mLockOnSwitch( false )
{
}


void KLocalSessionManager::setLockOnSwitch( bool lock )
{
    mLockOnSwitch = lock;
}


bool KLocalSessionManager::doesLockOnSwitch() const { return mLockOnSwitch; }

bool KLocalSessionManager::canCreateSession() const { return true; }//kapp->authorize( "start_new_session" ); }
bool KLocalSessionManager::canLogoutSession() const { return true; }//kapp->authorize( "logout" ); }
bool KLocalSessionManager::canLockSession()   const { return true; }//kapp->authorize( "lock_screen" ); }

bool KLocalSessionManager::canSaveSession()   const
{
    return false;
//   KConfig SMServerConfig( "ksmserverrc", false, false );
//   SMServerConfig.setGroup( "General" );
//   return SMServerConfig.readEntry("loginMode") == "restoreSavedSession";
}

bool KLocalSessionManager::canShutdown() const
{
    if( DM.type() == OldKDM )
        return strstr( DM.displayManagerSettings(), ",maysd" ) != 0;

    QByteArray response;

    if( DM.type() == GDM )
        return KDisplayManagerPipe().send( "QUERY_LOGOUT_ACTION\n", response ) && ( response.indexOf("HALT") >= 0 );

    return KDisplayManagerPipe().send( "caps\n", response ) && ( response.indexOf("\tshutdown") >= 0 );
}

bool KLocalSessionManager::canSwitchSession() const
{
    if( DM.type() == OldKDM )
        return DM.displayName()[0] == ':';

    if( DM.type() == GDM )
        return KDisplayManagerPipe().send( "QUERY_VT\n" );

    QByteArray response;
    return KDisplayManagerPipe().send( "caps\n", response ) && ( response.indexOf("\tlocal") >= 0 );
}


void KLocalSessionManager::createSession() const
{
    if( ! canCreateSession() )
        return;
#if 0
    const int answer = KMessageBox::warningContinueCancel(
      mScreen,
      i18n("<p>You have chosen to open another desktop session.<br>"
              "The Current session will be hidden "
              "and a new login screen will be displayed.<br>"
              "An F-key is assigned to each session; "
              "F%1 is usually assigned to the first session, "
              "F%2 to the second session and so on. "
              "You can switch between sessions by pressing "
              "Ctrl, Alt and the appropriate F-key at the same time. "
              "Additionally, the KDE Panel and Desktop menus have "
              "actions for switching between sessions.</p>")
                          .arg(7).arg(8),
      i18n("Warning - New Session"),
      KGuiItem(i18n("&Start New Session"), "fork"),
      ":confirmNewSession",
      KMessageBox::PlainCaption|KMessageBox::Notify );

    if( answer == KMessageBox::Cancel )
        return;
#endif
    if( mLockOnSwitch )
        lockSession();

    const char* const command = ( DM.type() == GDM ) ? "FLEXI_XSERVER\n" : "reserve\n";
    KDisplayManagerPipe().send( command );
}


void KLocalSessionManager::switchToTerminal( int terminalId ) const
{
    const char* const command = ( DM.type() == GDM ) ? "SET_VT %1\n" : "activate\tvt%1\n";

    const bool switched = KDisplayManagerPipe().send( QString::fromLatin1(command).arg(terminalId).toLatin1() );

    if( switched && mLockOnSwitch )
        lockSession();
}


void KLocalSessionManager::saveSession() const
{
//   if( canSaveSession() )
//     kapp->dcopClient()->send( "SMServerConfig", "default",
//                               "saveCurrentSession()", QByteArray() );
}

void KLocalSessionManager::lockSession() const
{
    if( !canLockSession() )
        return;

#ifndef Q_OS_WIN
    QDBusInterface screensaver( "org.freedesktop.ScreenSaver", "/ScreenSaver", "org.freedesktop.ScreenSaver" );
    screensaver.call( "Lock" );
#else
    LockWorkStation();
#endif // !Q_OS_WIN

//TODO: what is the difference to:
//          if (DMType != GDM)
//  KDisplayManagerPipe().send( "lock\n" );

}


void KLocalSessionManager::logoutSession() const
{
    if( canLogoutSession() )
        KWorkSpace::requestShutDown( KWorkSpace::ShutdownConfirmYes,
                                     KWorkSpace::ShutdownTypeDefault,
                                     KWorkSpace::ShutdownModeDefault );
}


void KLocalSessionManager::shutdown( ShutdownType shutdownType, ShutdownMode shutdownMode, const QString& bootOption )
{
    if( shutdownType == ShutdownTypeNone || shutdownType == ShutdownTypeLogout )
        return;

    bool canAsk;
    if( DM.type() == NewKDM )
    {
        QByteArray response;
        canAsk = KDisplayManagerPipe().send( "caps\n", response ) && ( response.indexOf("\tshutdown ask") >= 0 );
    }
    else
    {
        if( ! bootOption.isEmpty() )
            return;
        canAsk = false;
    }

    if( ! canAsk && shutdownMode == ShutdownModeInteractive )
        shutdownMode = ShutdownModeForceNow;

    QByteArray command;
    if( DM.type() == GDM )
    {
        command.append( shutdownMode == ShutdownModeForceNow ? "SET_LOGOUT_ACTION " : "SET_SAFE_LOGOUT_ACTION " );
        command.append( shutdownType == ShutdownTypeReboot ? "REBOOT\n" : "HALT\n" );
    }
    else
    {
        command.append( "shutdown\t" );
        command.append( (shutdownType == ShutdownTypeReboot) ? "reboot\t" : "halt\t" );
        if( ! bootOption.isEmpty() )
            command.append( "=" ).append( bootOption.toLocal8Bit() ).append( '\t' );
        command.append(
            shutdownMode == ShutdownModeInteractive ? "ask\n" :
            shutdownMode == ShutdownModeForceNow ?    "forcenow\n" :
            shutdownMode == ShutdownModeTryNow ?      "trynow\n" :
            /*else*/                                  "schedule\n" );
    }

    KDisplayManagerPipe().send( command.data() );
}

bool KLocalSessionManager::getBootOptions( QStringList& options, int& defaultOption, int& currentOption ) const
{
    if( DM.type() != NewKDM )
        return false;

    QByteArray response;
    bool success = KDisplayManagerPipe().send( "listbootoptions\n", response );
    if( success )
    {
        options = QString::fromLocal8Bit( response.constData() ).split( '\t', QString::SkipEmptyParts );
        if( options.size() >= 4 )
        {
            defaultOption = options[2].toInt( &success );
            if( success )
            {
                currentOption = options[3].toInt( &success );
                if( success )
                {
                    options = options[1].split( ' ', QString::SkipEmptyParts );
                    for( QStringList::Iterator it = options.begin(); it != options.end(); ++it )
                        (*it).replace( "\\s", " " );
                }
            }
        }
    }

    return success;
}


int KLocalSessionManager::numberOfReservedSessions() const
{
    if( DM.type() == GDM )
        return 1; /* Bleh */

    if( DM.type() == OldKDM )
        return strstr( DM.displayManagerSettings(), ",rsvd" ) ? 1 : -1;

    QByteArray response;
    int p;
    if( !(KDisplayManagerPipe().send( "caps\n", response ) && (p = response.indexOf( "\treserve " )) >= 0) )
          return -1;
    return atoi( response.constData() + p + 9 );
}

KSessionList KLocalSessionManager::localSessions()
{
    KSessionList sessionList;
    getLocalSessions( sessionList );
    return sessionList;
}

bool KLocalSessionManager::getLocalSessions( KSessionList& sessionList )
{
    if( DM.type() == OldKDM )
        return false;

    if( DM.type() == GDM )
    {
        QByteArray response;
        if( ! KDisplayManagerPipe().send("CONSOLE_SERVERS\n",response) )
            return false;

        const QStringList sessions = QString( response.constData() +3 ).split( QLatin1Char(';'), QString::SkipEmptyParts );
        foreach( const QString& s, sessions )
        {
            const QStringList sessionDetails = s.split( QLatin1Char(',') );

            KSession session;
            session.DisplayName = sessionDetails[0];
            session.LoginName = sessionDetails[1];
            session.VTId = sessionDetails[2].toInt();
            session.Session = "<unknown>";
            session.Current = ( sessionDetails[0] == ::getenv("DISPLAY") ); /* Bleh */
            session.IsTty = false;

            sessionList.append( session );
        }
    }
    else
    {
        QByteArray response;
        if( ! KDisplayManagerPipe().send("list\talllocal\n",response) )
            return false;

        const QStringList sessions = QString( response.constData() +3 ).split( QLatin1Char('\t'), QString::SkipEmptyParts );
        foreach( const QString& s, sessions )
        {
            const QStringList sessionDetails = s.split( QLatin1Char(',') );

            KSession session;
            session.DisplayName = sessionDetails[0];
            // remote?
            if( sessionDetails[1][0] == '@' )
            {
                session.Origin = sessionDetails[1].mid( 1 );
                session.VTId = 0;
            }
            else
                session.VTId = sessionDetails[1].mid( 2 ).toInt();
            session.LoginName = sessionDetails[2];
            session.Session = sessionDetails[3];
            session.Current = ( sessionDetails[4].indexOf('*') >= 0 );
            session.IsTty = ( sessionDetails[4].indexOf('t') >= 0 );

            sessionList.append( session );
        }
    }

    return true;
}
