/*
 *   Copyright (C) 2009 Jesper Thomschütz <jesperht@yahoo.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <KIcon>

#include <QLineEdit>
#include <QDebug>
#include <QPainter>

#include "webapp.h"

#define WIDTH 320
#define HEIGHT 480

Webapp::Webapp(QObject *parent, const QVariantList &args)
              : Plasma::Applet(parent, args)
{
    resize(260, 500);
    setHasConfigurationInterface(true);
    setAspectRatioMode(Plasma::KeepAspectRatio);
}

Webapp::~Webapp()
{
    delete m_background;
}

void Webapp::init()
{
    m_background = new QSvgRenderer(this);

    m_page = new Page(this);

    QObject::connect(m_page, SIGNAL(loadStarted()), this, SLOT(loadStarted()));
    QObject::connect(m_page, SIGNAL(loadFinished(bool)), this, SLOT(loadFinished()));

    m_webview = new Plasma::WebView(this);
    m_webview->setPage(m_page);
    m_webview->setUrl(KUrl("http://www.facebook.com"));

    m_button = new Plasma::ToolButton(this);

    QObject::connect(m_button, SIGNAL(clicked()), this, SLOT(showConfigurationInterface()));

    setBackground(":phone.svg");
}

void Webapp::setBackground(QString path)
{
    m_background->load(path);

    if (m_background->elementExists("minimalId"))
        m_backgroundHint = Plasma::Applet::DefaultBackground;
    else
        m_backgroundHint = Plasma::Applet::NoBackground;

    constraintsEvent(Plasma::SizeConstraint);
    update(boundingRect());
}

QString Webapp::validUrl(QString url)
{
    if (!url.startsWith("http"))
        url = "http://" + url;

    return url;
}

void Webapp::viewPressed()
{
    QString url = validUrl(ui.urlCombo->lineEdit()->text());
    ui.urlCombo->lineEdit()->setText(url);
    m_webview->setUrl(url);
}

void Webapp::favoritePressed()
{
    QString url = validUrl(ui.urlCombo->lineEdit()->text());
    ui.urlCombo->lineEdit()->setText(url);
    ui.urlCombo->addItem(url);
}

void Webapp::createConfigurationInterface(KConfigDialog *dialog)
{
    QWidget *widget = new QWidget();
    ui.setupUi(widget);

    QStringList urls;
    urls    << "http://www.facebook.com"
            << "http://talkgadget.google.com/talkgadget/m"
            << "http://www.google.com"
            << "http://www.digg.com/iphone/"
            << "http://planet.kde.org"
            << "http://www.tadalist.com";


    ui.urlCombo->addItems(urls);
    ui.urlCombo->setDuplicatesEnabled(false);

    QStringList themes;
    themes  << ":phone.svg"
            << ":minimal.svg";

    ui.themeCombo->addItems(themes);

    QObject::connect(ui.viewButton, SIGNAL(clicked()), this, SLOT(viewPressed()));
    QObject::connect(ui.favoriteButton, SIGNAL(clicked()), this, SLOT(favoritePressed()));

    QObject::connect(ui.viewButton, SIGNAL(clicked()), this, SLOT(viewPressed()));
    QObject::connect(ui.themeCombo, SIGNAL(activated(QString)), this, SLOT(setBackground(QString)));

    dialog->addPage(widget, i18n("Webapp Settings"), icon());
}

void Webapp::loadStarted()
{
    setBusy(true);
}

void Webapp::loadFinished()
{
    setBusy(false);
}

void Webapp::constraintsEvent(Plasma::Constraints constraints)
{
    Q_UNUSED(constraints);
    setBackgroundHints(m_backgroundHint);

    // Thanks to Capisce for helping me figure out how to get the "real"
    // geometry of SVG elements.
    QMatrix elementScale;
    QRectF viewBox = m_background->viewBoxF();
    elementScale.scale(boundingRect().width() / viewBox.width(),
                       boundingRect().height() / viewBox.height());
    QRectF r = (m_background->matrixForElement("screenId") * elementScale)
                  .mapRect(m_background->boundsOnElement("screenId"));

    //Put the browser in "native" phone resolution
    m_webview->setGeometry(QRectF(r.x(), r.y(), WIDTH, HEIGHT));
    m_webview->resetTransform();
    //Scale it down so that it fits in the screen area of the SVG.
    m_webview->scale(r.width()/WIDTH, r.height()/HEIGHT);

    r = (m_background->matrixForElement("buttonId") * elementScale)
                  .mapRect(m_background->boundsOnElement("buttonId"));

    m_button->setGeometry(r);
}

void Webapp::paintInterface(QPainter *painter, const QStyleOptionGraphicsItem *option,
                            const QRect &rect)
{
    Q_UNUSED(option);

    m_background->render(painter, rect);
}

#include "webapp.moc"
