/*
 *   Copyright (C) 2009 Ivan Cukic <ivan.cukic@gmail.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "gmailengine.h"
#include "gmailengine_p.h"
#include <KMessageBox>
#include <KRun>
#include <QXmlStreamReader>

#define FEED_URL KUrl("https://mail.google.com/mail/feed/atom")
#define MAX_ITEMS 10

// GmailJob
GmailJob::GmailJob(const QString & operation, const QMap < QString, QVariant > & parameters, QObject * parent)
    : ServiceJob("none", operation, parameters, parent)
{
    qDebug() << "GmailJob::GmailJob:" << operation << parameters;
}

void GmailJob::start()
{
    if (operationName() == "activate") {
        QByteArray data = parameters()["data"].toByteArray();
        qDebug() << "GmailJob::GmailJob: " << data;
        if (!data.isEmpty()) {
            new KRun(QUrl::fromEncoded(data), NULL);
        }
    }
    setResult(true);
}

// GmailService
GmailService::GmailService()
{
    qDebug() << "GmailService::GmailService";
    setName("lancelot-gmail");
}

GmailService::~GmailService()
{
    qDebug() << "GmailService::~GmailService";
}

Plasma::ServiceJob * GmailService::createJob(const QString & operation,
        QMap < QString, QVariant > & parameters)
{
    qDebug() << "GmailEngine::createJob:" << operation << parameters;
    return new GmailJob(operation, parameters, this);
}

// GmailEngine
GmailEngine::Private::Private(GmailEngine * parent)
    : job(NULL), q(parent)
{
}

GmailEngine::GmailEngine(QObject * parent, const QVariantList & args)
    : Plasma::DataEngine(parent, args), d(new Private(this))
{
    setMinimumPollingInterval(2 * 60 * 1000); // 2 minutes minimum
    setPollingInterval(5 * 60 * 1000); // 5 minutes default
    sourceRequestEvent(".metadata");
}

GmailEngine::~GmailEngine()
{
}

bool GmailEngine::sourceRequestEvent(const QString & name)
{
    qDebug() << "GmailEngine::sourceRequestEvent:" << name;

    if (name == ".metadata") {
        d->loadMetadata();
    } else if (name == "data") {
        d->loadData();
    }

    return true;
}

void GmailEngine::Private::loadMetadata()
{
    QMap < QString, QVariant > metadata;
    metadata["version"] = "1.0";
    metadata["modelIcon"] = "mail-receive";
    metadata["modelTitle"] = i18n("GMail");
    q->setData(".metadata", "lancelot", metadata);
}

void GmailEngine::Private::loadData()
{
    if (job) {
        job->kill();
    }

    job = KIO::get(FEED_URL, KIO::Reload, KIO::HideProgressInfo);
    connect(job,  SIGNAL (data(KIO::Job *, const QByteArray &)),
            this, SLOT   (recv(KIO::Job *, const QByteArray &)));
    connect(job,  SIGNAL (result(KJob *)),
            this, SLOT   (result(KJob *)));

    q->setData("data", "title", QStringList(i18n("Loading...")));
    q->setData("data", "description", QStringList(QString()));
    q->setData("data", "icon", QStringList("chronometer"));
    q->setData("data", "data", QStringList(QString()));
}

Plasma::Service * GmailEngine::serviceForSource(const QString & name)
{
    GmailService * service = new GmailService();

    qDebug() << "GmailEngine::serviceForSource:" << name;
    return service;
}

bool GmailEngine::updateSourceEvent(const QString & source)
{
    qDebug() << "GmailEngine::updateSourceEvent:" << source;
    if (source == "data") {
        d->loadData();
    }

    return true;
}

void GmailEngine::Private::recv(KIO::Job * j, const QByteArray & data)
{
    if (job != j) {
        return;
    }

    xml += data;
}

void GmailEngine::Private::result(KJob * j)
{
    if (job != j) {
        return;
    }

    job = NULL;
    qDebug() << xml;

    QStringList title;
    QStringList description;
    QStringList icon;
    QStringList data;
    int index = -1;

    QXmlStreamReader reader(xml);
    while (!reader.atEnd()) {
         reader.readNext();

         QString tag = reader.name().toString();
         if (reader.isStartElement()) {
             if (tag == "entry") {
                 index++;
                 if (index >= MAX_ITEMS) {
                     break;
                 }
                 icon << "mail-mark-unread";
             }

             if (index < 0) continue;

             if (tag == "title") {
                 reader.readNext();
                 title << reader.text().toString().trimmed();
                 qDebug() << "We have a title" << title.last();
             } else if (tag == "link") {
                 data << reader.attributes().value("href").toString();
                 qDebug() << "We have the link" << data.last();
             } else if (tag == "author") {
                 QString name, email;

                 while (!reader.isEndElement() || reader.name().toString() != "author") {
                     reader.readNext();
                     QString intag = reader.name().toString();
                     if (reader.isStartElement()) {
                         if (intag == "name") {
                             reader.readNext();
                             name = reader.text().toString().trimmed();
                         } else if (intag == "email") {
                             reader.readNext();
                             email = reader.text().toString().trimmed();
                         }
                     }
                 }

                 description << name;
                 qDebug() << "We have the author" << description.last();
             }
         }
    }

    if (index >= MAX_ITEMS) {
        title << i18n("More...");
        description << QString();
        icon << "mail-mark-unread";
        data << "http://mail.google.com/mail";
    } else if (index < 0) {
        title << i18n("No unread mail");
        description << QString();
        icon << "mail-mark-unread";
        data << "http://mail.google.com/mail";
    }

    q->setData("data", "title", title);
    q->setData("data", "description", description);
    q->setData("data", "icon", icon);
    q->setData("data", "data", data);
}

#include "gmailengine.moc"
#include "gmailengine_p.moc"
