/***************************************************************************
 *   Portions of Kweather's METAR parser                                   *
 *   Copyright (C) 2004 by John Ratke <jratke@comcast.net>                 *
 *   Copyright (C) 2002-2004 Nadeem Hasan <nhasan@kde.org>                 *
 *   Copyright (C) 2002-2004 Ian Geiser <geiseri@kde.org>                  *
 *                                                                         *
 *   Conversion to ion framework                                           *
 *   Copyright (C) 2009 by Shawn Starr <shawn.starr@rogers.com>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Ion for NOAA's National Weather Service METAR data */

#include "ion_noaametar.h"

class NOAAMETARIon::Private : public QObject
{
public:
    struct METARMapInfo {
        QString stationID;
        QString placeName;
        QString stateName;
        QString countryName;
        QString stationLat;
        QString stationLon;
    };

    // Key dicts
    QHash<QString, NOAAMETARIon::Private::METARMapInfo> m_places;

    // Weather information
    QHash<QString, WeatherData> m_weatherData;

    // Store KIO jobs
    QMap<KJob *, QByteArray> m_jobMetar;
    QMap<KJob *, QString> m_jobList;
    QByteArray m_metarSetup;

    QDateTime m_dateFormat;
    bool emitWhenSetup;
    int m_conditionTypeFlag;
    bool m_mainCondition;
    Private() : emitWhenSetup(false), m_conditionTypeFlag(0), m_mainCondition(false) {
    }

    QRegExp m_coverRegExp;
    QRegExp m_currentRegExp;
    QRegExp m_windRegExp;
    QRegExp m_visRegExp;
    QRegExp m_visFracRegExp;
    QRegExp m_tempRegExp;
    QRegExp m_timeRegExp;
    QRegExp m_dateRegExp;
    QRegExp m_pressRegExp;
    QRegExp m_tempTenRegExp;
};

QMap<QString, IonInterface::ConditionIcons> NOAAMETARIon::setupConditionIconMappings(void)
{

    QMap<QString, ConditionIcons> conditionList;
     
    conditionList["clear"] = ClearDay;
    conditionList["a few clouds"] = FewCloudsDay;
    conditionList["partly cloudy"] = PartlyCloudyDay;
    conditionList["mainly cloudy"] = Overcast;
    conditionList["overcast"] = Overcast;
    conditionList["undetermined precipitation"] = NotAvailable;
    conditionList["shallow fog"] = Mist;
    conditionList["patches of fog"] = Mist;
    conditionList["partial fog"] = Mist;
    conditionList["drifting snow"] = Snow;
    conditionList["drifting dust"] = NotAvailable;
    conditionList["drifting sand"] = NotAvailable;
    return conditionList;
}

QMap<QString, IonInterface::ConditionIcons> const& NOAAMETARIon::conditionIcons(void)
{
    static QMap<QString, ConditionIcons> const condval = setupConditionIconMappings();
    return condval;
}

// ctor, dtor
NOAAMETARIon::NOAAMETARIon(QObject *parent, const QVariantList &args)
        : IonInterface(parent, args), d(new Private())
{
    Q_UNUSED(args)
}

void NOAAMETARIon::reset()
{
    delete d;
    d = new Private();
    d->emitWhenSetup = true;
    setInitialized(false);
    getMetarSetup();
}

NOAAMETARIon::~NOAAMETARIon()
{
    // Destroy dptr
    delete d;
}

// Get the master list of locations to be parsed
void NOAAMETARIon::init()
{
    d->m_coverRegExp    = QRegExp("^(AUTO|NSC|FEW|SCT|BKN|OVC|SKC|CLR|CAVOK)([0-9]{3})?(?:TCU|CB)?$");
    d->m_currentRegExp  = QRegExp("^(\\+|-|VC)?([A-Z]{2,4})$");
    d->m_windRegExp     = QRegExp("^([0-9]{3}|VRB)([0-9]{2,3})(?:G([0-9]{2,3}))?(KT|KMH|MPS)$");
    d->m_visRegExp      = QRegExp("^([0-9]{1,2})SM$");
    d->m_visFracRegExp  = QRegExp("^1/(2|4)SM$");
    d->m_tempRegExp     = QRegExp("^(M)?([0-9]{2})/(?:(M)?([0-9]{2}))?$");
    d->m_timeRegExp     = QRegExp("^([0-9]{2}:[0-9]{2})$");
    d->m_dateRegExp     = QRegExp("^([0-9]{4}/[0-9]{2}/[0-9]{2})$");
    d->m_pressRegExp    = QRegExp("^([AQ])([0-9]{4})$");
    d->m_tempTenRegExp  = QRegExp("^T([01][0-9]{3})([01][0-9]{3})$");

    // Get the real city XML URL so we can parse this
    getMetarSetup();
}

QStringList NOAAMETARIon::validate(const QString& source) const
{
    QStringList placeList;
    QString station;
    QString sourceNormalized = source.toUpper();

    QHash<QString, NOAAMETARIon::Private::METARMapInfo>::const_iterator it = d->m_places.constBegin();
    // If the source name might look like a station ID, check these too and return the name
    bool checkState = source.count() == 2;

    while (it != d->m_places.constEnd()) {
        if (checkState) {
            if (it.value().stateName == source) {
                placeList.append(QString("place|").append(it.key()));
            }
        } else if (it.key().toUpper().contains(sourceNormalized)) {
            placeList.append(QString("place|").append(it.key()));
        } else if (it.value().stationID == sourceNormalized) {
            station = QString("place|").append(it.key());
        }

        ++it;
    }

    placeList.sort();
    if (!station.isEmpty()) {
        placeList.prepend(station);
    }

    return placeList;
}

bool NOAAMETARIon::updateIonSource(const QString& source)
{
    // We expect the applet to send the source in the following tokenization:
    // ionname:validate:place_name - Triggers validation of place
    // ionname:weather:place_name - Triggers receiving weather of place

    QStringList sourceAction = source.split('|');

    // Guard: if the size of array is not 2 then we have bad data, return an error
    if (sourceAction.size() < 2) {
        setData(source, "validate", "noaametar|malformed");
        return true;
    }

    if (sourceAction[1] == "validate" && sourceAction.size() > 2) {
        QStringList result = validate(sourceAction[2]);

        if (result.size() == 1) {
            setData(source, "validate", QString("noaametar|valid|single|").append(result.join("|")));
            return true;
        } else if (result.size() > 1) {
            setData(source, "validate", QString("noaametar|valid|multiple|").append(result.join("|")));
            return true;
        } else if (result.size() == 0) {
            setData(source, "validate", QString("noaametar|invalid|single|").append(sourceAction[2]));
            return true;
        }

    } else if (sourceAction[1] == "weather" && sourceAction.size() > 2) {
        getMetarData(source);
        return true;
    } else {
        setData(source, "validate", "noaametar|malformed");
        return true;
    }

    return false;
}

// Parses city list and gets the correct city based on ID number
void NOAAMETARIon::getMetarSetup()
{
    KIO::TransferJob *job = KIO::get(KUrl("http://weather.noaa.gov/data/nsd_cccc.txt"), KIO::NoReload, KIO::HideProgressInfo);

    if (job) {
        connect(job, SIGNAL(data(KIO::Job *, const QByteArray &)), this,
                SLOT(setup_slotDataArrived(KIO::Job *, const QByteArray &)));
        connect(job, SIGNAL(result(KJob *)), this, SLOT(setup_slotJobFinished(KJob *)));
    } else {
        kDebug() << "Could not create place name list transfer job";
    }
}

// Gets specific city XML data
void NOAAMETARIon::getMetarData(const QString& source)
{
    QString dataKey = source;
    dataKey.remove("noaametar|weather|");
    KUrl url(QString("ftp://anonymous:kdeusers@tgftp.nws.noaa.gov/data/observations/metar/stations/%1.TXT").arg(d->m_places[dataKey].stationID));

    kDebug() << "Will Try URL: " << url.url();
    // If this is empty we have no valid data, send out an error and abort.
    if (url.url().isEmpty()) {
        setData(source, "validate", QString("noaametar|malformed"));
        return;
    }

    KIO::TransferJob * const m_job = KIO::get(url.url(), KIO::Reload, KIO::HideProgressInfo);
    d->m_jobMetar.insert(m_job, QByteArray());
    d->m_jobList.insert(m_job, source);

    if (m_job) {
        connect(m_job, SIGNAL(data(KIO::Job *, const QByteArray &)), this,
             SLOT(slotDataArrived(KIO::Job *, const QByteArray &)));
        connect(m_job, SIGNAL(result(KJob *)), this, SLOT(slotJobFinished(KJob *)));
    }
}

void NOAAMETARIon::setup_slotDataArrived(KIO::Job *job, const QByteArray &data)
{
    Q_UNUSED(job)

    if (data.isEmpty()) {
        return;
    }

    // Keep track of individual data
    d->m_metarSetup.append(data);
}

void NOAAMETARIon::slotDataArrived(KIO::Job *job, const QByteArray &data)
{
    if (data.isEmpty() || !d->m_jobMetar.contains(job)) {
        return;
    }
    d->m_jobMetar[job].append(data);
}

void NOAAMETARIon::slotJobFinished(KJob *job)
{
    removeAllData(d->m_jobList[job]);
    parseMetarData(d->m_jobList[job], d->m_jobMetar[job]);
    d->m_jobList.remove(job);
    d->m_jobMetar.remove(job);
}

void NOAAMETARIon::setup_slotJobFinished(KJob *job)
{
    Q_UNUSED(job)
    const bool success = readMetarSetup();
    setInitialized(success);
    if (d->emitWhenSetup) {
        d->emitWhenSetup = false;
        emit(resetCompleted(this, success));
    }
}

// Parse the city list and store into a QMap
bool NOAAMETARIon::readMetarSetup()
{
    QTextStream stream(d->m_metarSetup.data());
    QString line;
    QString buildKey;
    while (!stream.atEnd()) {
       line = stream.readLine();
       QStringList data = line.split(';', QString::KeepEmptyParts);
       NOAAMETARIon::Private::METARMapInfo info;
       info.stationID = data[0].trimmed();
       info.placeName = data[3].trimmed();
       info.countryName = data[5].trimmed();
       info.stationLat = data[7].trimmed();
       info.stationLon = data[8].trimmed();

       // USA specific, drop country name
       if (info.countryName == "United States") {
           info.stateName = data[4].trimmed(); // NOAA only provides this for USA.
           buildKey = info.placeName + ", " + info.stateName; // Build the key name.
       } else {
           info.stateName = "N/A";
           buildKey = info.placeName + ", " + info.countryName; 
       }
       d->m_places[buildKey] = info;
    }
    return true;
}

void NOAAMETARIon::parseMetarData(const QString& source, const QByteArray &metar)
{
    QString dataKey = source;
    dataKey.remove("noaametar|weather|");

    WeatherData data;

    data.stationID = d->m_places[dataKey].stationID;
    data.locationName = d->m_places[dataKey].placeName;
    data.stateName = d->m_places[dataKey].stateName;
    data.countryName = d->m_places[dataKey].countryName;
    data.stationLat = d->m_places[dataKey].stationLat;
    data.stationLon = d->m_places[dataKey].stationLon;
    data.windKMH = 0.0;
    data.tempC = 0.0;
    data.dewC = 0.0;
    data.weather = "";
    //data.coverList = "N/A";
    //data.currentList = "N/A";
    data.obsDate = QDate();
    data.obsTime = QTime();
    data.pressure = "N/A";
    data.temperature = "N/A";
    data.dewPoint = "N/A";
    data.visibility = "N/A";
    data.humidity = 0.0;
    data.place = "N/A";
    data.stationOOS = false;

    
    QString weatherInfo = metar.data();
    QStringList dataList = weatherInfo.split(QRegExp("\\s+"));
    
    
    bool found = false;
    bool beforeRemark = true;
 
    // Reset flags
    d->m_mainCondition = false; 
    for (QStringList::ConstIterator it = dataList.constBegin();
         it != dataList.constEnd(); ++it)
    {
         // Don't try to parse the ICAO location code
         if ((!found) && (*it == data.stationID.toUpper().trimmed())) {
             found = true;
             continue;
         }
 
         //kDebug() << "Processing Token: " << *it;
 
         if (*it == "RMK") {  // Remarks
             beforeRemark = false;
             continue;
         }

         if (beforeRemark) {
             if (parseDate(data, *it)) 
                 continue;
             if (parseTime(data, *it)) 
                 continue;
             if (parseWindSpeed(data, *it))
                 continue;
             if (parseVisibility(data, it)) // note, pass in iterator.
                 continue;
             if (parseTemperature(data, *it))
                 continue;
             if (parseCloudCover(data, *it))
                 continue;
             if (!d->m_mainCondition) { 
                 if (parseCurrentCondition(data, *it)) {
                     d->m_mainCondition = true;
                     continue;
                 }
             }
             if (parsePressure(data, *it)) {
                 beforeRemark = false;
                 continue;
             }
                   
         } else {
             //if (parseTemperatureTenths(data, *it))
             //    continue;
             //if (parseStationNeedsMaintenance(data, *it))
                 continue;
         }
     }

     // Pass this to weather next!
     calculateHumidity(data);
     calculateHumidex(data);
     calculateWindchill(data);
}

void NOAAMETARIon::calculateHumidity(WeatherData& weatherInfo)
{
    float humidity = vaporPressure(weatherInfo.dewC) / vaporPressure(weatherInfo.tempC) * 100;
    weatherInfo.humidity = humidity;
    kDebug() << "WEATHER METAR: Humidity: " << weatherInfo.humidity;
}
   
void NOAAMETARIon::calculateHumidex(WeatherData& weatherInfo)
{

    // We only care if temperature is 20C degrees, humidity can make this quite uncomfortable
    if (weatherInfo.tempC >= 20 && weatherInfo.dewC > 0) {
        float t = 7.5 * weatherInfo.tempC / (237.7 + weatherInfo.tempC);
        float e = (6.112 * std::pow(10.0, t) * (weatherInfo.humidity / 100));
        weatherInfo.humidexC.setNum(weatherInfo.tempC + 0.5555 * (e-10));
    } else {
        weatherInfo.humidexC = "N/A";
    }
    kDebug() << "WEATHER METAR: Humidex: " << weatherInfo.humidexC;
}

void NOAAMETARIon::calculateWindchill(WeatherData& weatherInfo)
{
    // We only care if temperature is 5C or less and wind speed is 5km/h or more
    if (weatherInfo.tempC <= 5 && weatherInfo.windKMH >= 5) {
        float windChill = 13.12 + 0.6215 * weatherInfo.tempC - 11.37 * std::pow(weatherInfo.windKMH, 0.16)
                          + 0.3965 * weatherInfo.tempC * std::pow(weatherInfo.windKMH, 0.16);
        weatherInfo.windchillC.setNum(windChill);
    } else {
        weatherInfo.windchillC = "N/A";
    }
    kDebug() << "WEATHER METAR: Windchill: " << weatherInfo.windchillC;
}

bool NOAAMETARIon::parseDate(WeatherData& weatherInfo, const QString& token)
{
    if (d->m_dateRegExp.indexIn(token) > -1) {
        QString dateString = d->m_dateRegExp.cap(1);
        QString day, month, year;

        day = dateString.mid(8,2);
        month = dateString.mid(5,2);
        year = dateString.mid(0,4);

        QDate theDate(year.toInt(), month.toInt(), day.toInt());

        weatherInfo.obsDate = theDate;
        kDebug() << "METAR DATA: Date: " << weatherInfo.obsDate;
        return true;
    }
 
    return false;
}

bool NOAAMETARIon::parseTime(WeatherData& weatherInfo, const QString& token)
{
    if (d->m_timeRegExp.indexIn(token) > -1) {
        QString hour, minute, dateString;

        dateString = d->m_timeRegExp.cap(1);
        hour = dateString.mid(0,2);
        minute = dateString.mid(3,2);
        QTime theTime(hour.toInt(), minute.toInt());

        weatherInfo.obsTime = theTime;
	kDebug() << "METAR DATA: Time: " << weatherInfo.obsTime;
        return true;
    }
 
    return false;
}

/** Parse out the current visibility */
bool NOAAMETARIon::parseVisibility(WeatherData& weatherInfo, QStringList::ConstIterator it)
{
	float fVisibility = 0;
	
	if (d->m_visRegExp.indexIn(*it) > -1)
	{
		fVisibility = d->m_visRegExp.cap(1).toFloat();

	}
	else if (d->m_visFracRegExp.indexIn(*it) > -1)
	{
		// got a fractional visibility, go back to previous string in the list
		// and get the whole part.
		fVisibility = (*(it--)).toFloat();
		// shouldn't be necessary?  
		//it++;
		fVisibility += ( 1 / d->m_visFracRegExp.cap(1).toFloat() );
	}
	
	if (fVisibility > 0) {
	    weatherInfo.visibility.setNum(WeatherUtils::convertDistance(fVisibility, WeatherUtils::Miles, WeatherUtils::Kilometers));
            kDebug() << "METAR DATA: Visibility: " << weatherInfo.visibility;
	    return true;
	} else {
	    return false;
	}
}

/** Parse out the current pressure. */
bool NOAAMETARIon::parsePressure(WeatherData& weatherInfo, const QString& token)
{
	if (d->m_pressRegExp.indexIn(token) > -1)
	{
		QString type = d->m_pressRegExp.cap(1);
		float fPressure = d->m_pressRegExp.cap(2).toFloat();

                kDebug() << "METAR DATA: Pressure Type: " << type;
               
		if (type == "A") { // this is in inhg
		    weatherInfo.pressure.setNum(WeatherUtils::convertPressure(fPressure / 100, WeatherUtils::InchesHG, WeatherUtils::Kilopascals), 'f', 2);
                }
                if (type == "Q") { // this is in hPa
                    weatherInfo.pressure.setNum(WeatherUtils::convertPressure(fPressure, WeatherUtils::Hectopascals, WeatherUtils::Kilopascals), 'f', 2);
                }
                kDebug() << "METAR DATA: Pressure in kPa" << weatherInfo.pressure;
		return true;
	}
	return false;
}

struct wind_info
{
	unsigned int   number;
	QString        name;
};

static const struct wind_info wind_direction[] = 
{
	{ 0,   i18n("N")   },   // North is 0 to 11, and so on
	{ 12,  i18n("NNE") },
	{ 33,  i18n("NE")  },
	{ 57,  i18n("ENE") },
	{ 79,  i18n("E")   }, 
	{ 102, i18n("ESE") },
	{ 124, i18n("SE")  },
	{ 147, i18n("SSE") },
	{ 169, i18n("S")   },
	{ 192, i18n("SSW") },
	{ 214, i18n("SW")  },
	{ 237, i18n("WSW") },
	{ 259, i18n("W")   },
	{ 282, i18n("WNW") },
	{ 304, i18n("NW")  },
	{ 327, i18n("NNW") },
	{ 349, i18n("N")   },
	{ 360, i18n("N")   }
};


QString NOAAMETARIon::parseWindDirection(const unsigned int direction)
{
	unsigned int i = 0;
	
	for (i = 0; i < (sizeof(wind_direction) / sizeof(wind_info)) - 1; i++)
	{
		if (direction >= wind_direction[i].number &&
		    direction < wind_direction[i + 1].number)
		{
			break;
		}
	}
	
	return wind_direction[i].name;
}

/** Parse out the current temperature */
bool NOAAMETARIon::parseTemperature(WeatherData& weatherInfo, const QString &token)
{
	if (d->m_tempRegExp.indexIn(token) > -1)
	{
		float temp = d->m_tempRegExp.cap(2).toFloat();
		if (d->m_tempRegExp.cap(1) == "M" && temp != 0 )
			temp *= -1;
		float dewPoint = d->m_tempRegExp.cap(4).toFloat();
		if (d->m_tempRegExp.cap(3) == "M" && dewPoint != 0 )
			dewPoint *= -1;
		
		weatherInfo.tempC = temp;
		weatherInfo.dewC = dewPoint;
                kDebug() << "METAR DATA: Temperature in C: " << weatherInfo.tempC;
                kDebug() << "METAR DATA: Dewpoint in C: " << weatherInfo.dewC;
		return true;
	}
	return false;
}

/** Parse the current cover type */
bool NOAAMETARIon::parseCloudCover(WeatherData& weatherInfo, const QString &token)
{
	if (d->m_coverRegExp.indexIn(token) > -1)
	{
		QString sCode = d->m_coverRegExp.cap(1);
		float height = d->m_coverRegExp.cap(2).toFloat();  // initially in 100's of feet
		QString sClouds;

		height *= 100;
		// using plural i18n form for proper translations
		sClouds = i18np("1 foot", "%1 feet", (int)height);

		if (sCode == "FEW" || sCode == "NSC")
		{
			weatherInfo.clouds += 2;
		}
		else if (sCode == "SCT")
		{
			weatherInfo.clouds += 4;
		}
		else if (sCode == "BKN")
		{
			weatherInfo.clouds += 8;
		}
		else if (sCode == "OVC")
		{
			weatherInfo.clouds += 64;
		}
		else if ((sCode == "CLR") || (sCode == "SKC") || (sCode == "CAVOK"))
		{
                        weatherInfo.weather = "clear";
		        kDebug() << "METAR DATA: Cloud/Sky Condition: " << weatherInfo.weather;
		}
                else if (sCode == "AUTO") {
                        weatherInfo.weather = "N/A";
                        kDebug() << "METAR DATA: Cloud/Sky Condition: N/A";
                }

                // if we have a weather condition set, we dont care about the cloud types
                if (weatherInfo.weather.isEmpty() == false && weatherInfo.cloudsCondition == false) {
                    return true;
                }
                    
                // We gestimate the kind of cloud cover based on the different observation levels
                if ( weatherInfo.weather.isEmpty() || (weatherInfo.weather.isEmpty() == false && weatherInfo.cloudsCondition)) {
                    if (weatherInfo.clouds >= 2 && weatherInfo.clouds <= 3) {
                        weatherInfo.weather = "a few clouds";
                        weatherInfo.cloudsCondition = true;
                    }
                    if (weatherInfo.clouds >= 4 && weatherInfo.clouds <= 7) {
                        weatherInfo.weather = "partly cloudy";
                        weatherInfo.cloudsCondition = true;
                    }
                    if (weatherInfo.clouds >= 8 && weatherInfo.clouds < 64) {
                        weatherInfo.weather = "mainly cloudy";
                        weatherInfo.cloudsCondition = true;
                    }
                    if (weatherInfo.clouds >= 64) {
                        weatherInfo.weather = "overcast";
                        weatherInfo.cloudsCondition = true;
                    }
                }
		kDebug() << "METAR DATA: Cloud/Sky Condition: " << weatherInfo.weather;

		return true;
	}
	return false;
}

/** Parse the wind speed */
bool NOAAMETARIon::parseWindSpeed(WeatherData& weatherInfo, const QString& token)
{
	if (d->m_windRegExp.indexIn(token) > -1)
	{
		unsigned int direction = d->m_windRegExp.cap(1).toInt();
		float windSpeed = d->m_windRegExp.cap(2).toFloat();
		float gustSpeed = d->m_windRegExp.cap(3).toFloat();
		QString sWindUnit = d->m_windRegExp.cap(4);

		weatherInfo.windKMH = windSpeed;

		if (gustSpeed >= 1)
		{
		     weatherInfo.windKMHGust = gustSpeed;
		}

		if ((d->m_windRegExp.cap(1) != "VRB") && (windSpeed >= 1))
		{
			weatherInfo.windDirection = parseWindDirection(direction);
		}

                kDebug() << "METAR DATA: Wind Speed: " << weatherInfo.windKMH;
                kDebug() << "METAR DATA: Wind Gust: " << weatherInfo.windKMHGust;
                kDebug() << "METAR DATA: Wind Direction: " << weatherInfo.windDirection;

		return true;
	}
	return false;
}

bool NOAAMETARIon::parseStationNeedsMaintenance(WeatherData& weatherInfo, const QString &token)
{
	if (token == "$")
	{
		weatherInfo.stationOOS = true;
		kDebug(12006) << "Station Needs Maintenance";
		return true;
	}
	
	return false;
}

/** Parse the current weather conditions */
bool NOAAMETARIon::parseCurrentCondition(WeatherData& weatherInfo, const QString &token)
{
        
	if (d->m_currentRegExp.indexIn(token) > -1)
	{
		QString sIntensity = d->m_currentRegExp.cap(1);
		QString sCode = d->m_currentRegExp.cap(2);
		QString descriptor, phenomena, currentWeather;

                kDebug() << "INTENSITY: " << sIntensity;
                kDebug() << "CONDITION CODE: " << sCode;
                
		// Decode the intensity
		if (sIntensity == "+")
		{
			weatherInfo.weather = "heavy ";
		}
		else if (sIntensity == "-")
		{
			weatherInfo.weather = "light ";
		} 

		// Decode the descriptor
		if (sCode.contains("MI")) {
                    weatherInfo.weather += " shallow ";
                    d->m_conditionTypeFlag = FOG;
                } else if (sCode.contains("BC")) {
		    weatherInfo.weather += " partial ";
                    d->m_conditionTypeFlag = FOG;
		} else if (sCode.contains("PR")) {
		    weatherInfo.weather += " patches ";
                    d->m_conditionTypeFlag = FOG;
		} else if (sCode.contains("DR")) {
	            weatherInfo.weather += " drifting ";
                    d->m_conditionTypeFlag = SNOWSAND;
		} else if (sCode.contains("BL")) {
                    weatherInfo.weather += " blowing ";
                    d->m_conditionTypeFlag = SNOWSAND;
		} else if (sCode.contains("SH")) {
                    weatherInfo.weather += " shower ";
                    d->m_conditionTypeFlag = SHOWER;
		} else if (sCode.contains("TS")) {
                    weatherInfo.weather += " thunderstorm ";
                    d->m_conditionTypeFlag = THUNDERSTORM;
                } else if (sCode.contains("FZ")) {
                    weatherInfo.weather += " freezing ";
                    d->m_conditionTypeFlag = ICE;
                }

                // Decode the precipitation
                if (sCode.contains("DZ") || (sCode.contains("DZ") && (d->m_conditionTypeFlag == ICE || d->m_conditionTypeFlag == THUNDERSTORM))) { 
		    weatherInfo.weather += " drizzle ";
                }

		if (sCode.contains("RA") && (d->m_conditionTypeFlag == SHOWER || d->m_conditionTypeFlag == ICE || d->m_conditionTypeFlag == THUNDERSTORM)) {
		    weatherInfo.weather += " rain ";
		} else if (sCode.contains("SN") && (d->m_conditionTypeFlag == SHOWER || d->m_conditionTypeFlag == SNOWSAND || d->m_conditionTypeFlag == THUNDERSTORM)) {
                    weatherInfo.weather += " snow ";
		} else if (sCode.contains("SG") || (sCode.contains("SG") && (d->m_conditionTypeFlag == SHOWER || d->m_conditionTypeFlag == SNOWSAND || d->m_conditionTypeFlag == THUNDERSTORM))) {
		    weatherInfo.weather += " snow grains ";
		} else if (sCode.contains("GS") || (sCode.contains("GR") && (d->m_conditionTypeFlag == SHOWER || d->m_conditionTypeFlag == THUNDERSTORM))) {
		    weatherInfo.weather += " hail ";
		} else if (sCode.contains("GS") || (sCode.contains("GS") && (d->m_conditionTypeFlag == SHOWER || d->m_conditionTypeFlag == THUNDERSTORM))) {
		    weatherInfo.weather += " snow pellets ";
		} else if (sCode.contains("IC") || (sCode.contains("IC") && d->m_conditionTypeFlag == THUNDERSTORM)) {
	            weatherInfo.weather += " ice crystals ";
		} else if (sCode.contains("PL") || (sCode.contains("PL") && d->m_conditionTypeFlag == THUNDERSTORM)) {
		    weatherInfo.weather += " ice pellets ";
		} else if (sCode.contains("UP")) {
		    weatherInfo.weather += " unknown precipitation ";
		}
 
                // Decode obscuration
		if (sCode.contains("BR") && (d->m_conditionTypeFlag == FOG || d->m_conditionTypeFlag == THUNDERSTORM)) {
	            weatherInfo.weather += " mist ";
		} else if (sCode.contains("FG") && (d->m_conditionTypeFlag == FOG || d->m_conditionTypeFlag == ICE || d->m_conditionTypeFlag == THUNDERSTORM)) {
		    weatherInfo.weather += " fog ";
		} else if (sCode.contains("FU"))
			phenomena = i18n("Smoke");
		else if (sCode.contains("VA"))
			phenomena = i18n("Volcanic Ash");
		else if (sCode.contains("DU"))
			phenomena = i18n("Widespread Dust");
		else if (sCode.contains("SA") || (sCode.contains("SA") && (d->m_conditionTypeFlag == SNOWSAND || d->m_conditionTypeFlag == THUNDERSTORM))) {
			weatherInfo.weather += " sand"; 
                }
		else if (sCode.contains("HZ") || (sCode.contains("HZ") && (d->m_conditionTypeFlag == THUNDERSTORM))) {
			weatherInfo.weather += " haze ";
                }
		else if (sCode.contains("PY"))
			phenomena = i18n("Spray");
		else if (sCode.contains("PO"))
			phenomena = i18n("Dust devil");
		else if (sCode.contains("SQ"))
			phenomena = i18n("Squall");
		else if (sCode.contains("FC"))
		{
			if (sIntensity == "+")
				currentWeather = i18n("Tornado");
			else
				phenomena = i18n("Funnel Cloud");
		}
		else if (sCode.contains("SS"))
			phenomena = i18n("Sand Storm");
		else if (sCode.contains("DS"))
			phenomena = i18n("Dust Storm");
		
		//if (!currentWeather.isEmpty())
	//		weatherInfo.qsCurrentList << currentWeather;
                kDebug() << "METAR DATA: Condition: " << weatherInfo.weather.simplified();
		return true;
	}
	return false;
}

void NOAAMETARIon::updateWeather(const QString& source)
{
    QMap<QString, QString> dataFields;
    Plasma::DataEngine::Data data;

    data.insert("Country", country(source));
    data.insert("Place", place(source));
    data.insert("Station", station(source));

    data.insert("Latitude", latitude(source));
    data.insert("Longitude", longitude(source));

    // Real weather - Current conditions
    data.insert("Observation Period", observationTime(source));
    data.insert("Current Conditions", condition(source));

// FIXME: We'll need major fuzzy logic, this isn't pretty: http://www.weather.gov/xml/current_obs/weather.php
    //QMap<QString, ConditionIcons> conditionList;
    //conditionList = conditionIcons();

    /*
        if ((periodHour(source) >= 0 && periodHour(source) < 6) || (periodHour(source) >= 18)) {
            // Night
            // - Fill in condition fuzzy logic
        } else {
            // Day
            // - Fill in condition fuzzy logic
        }
    */
    data.insert("Condition Icon", "weather-none-available");

    dataFields = temperature(source);
    data.insert("Temperature", dataFields["temperature"]);

    if (dataFields["temperature"] != "N/A") {
        data.insert("Temperature Unit", dataFields["temperatureUnit"]);
    }

/*
    // Do we have a comfort temperature? if so display it
    if (dataFields["comfortTemperature"] != "N/A") {
        if (d->m_weatherData[source].windchill_F != "NA") {
            data.insert("Windchill", QString("%1").arg(dataFields["comfortTemperature"]));
            data.insert("Humidex", "N/A");
        }
        if (d->m_weatherData[source].heatindex_F != "NA" && d->m_weatherData[source].temperature_F.toInt() != d->m_weatherData[source].heatindex_F.toInt()) {
            data.insert("Humidex", QString("%1").arg(dataFields["comfortTemperature"]));
            data.insert("Windchill", "N/A");
        }
    } else {
        data.insert("Windchill", "N/A");
        data.insert("Humidex", "N/A");
    }
*/

    data.insert("Dewpoint", dewpoint(source));
    if (dewpoint(source) != 0.0) {
        data.insert("Dewpoint Unit", dataFields["temperatureUnit"]);
    }

    dataFields = pressure(source);
    data.insert("Pressure", dataFields["pressure"]);

    if (dataFields["pressure"] != "N/A") {
        data.insert("Pressure Unit", dataFields["pressureUnit"]);
    }

    dataFields = visibility(source);
    data.insert("Visibility", dataFields["visibility"]);

    if (dataFields["visibility"] != "N/A") {
        data.insert("Visibility Unit", dataFields["visibilityUnit"]);
    }

    data.insert("Humidity", humidity(source));

    // Set number of forecasts per day/night supported, none for this ion right now
    data.insert(QString("Total Weather Days"), 0);

    dataFields = wind(source);
    data.insert("Wind Speed", dataFields["windSpeed"]);

    if (dataFields["windSpeed"] != "Calm") {
        data.insert("Wind Speed Unit", dataFields["windUnit"]);
    }

    data.insert("Wind Gust", dataFields["windGust"]);
    data.insert("Wind Gust Unit", dataFields["windGustUnit"]);
    data.insert("Wind Direction", dataFields["windDirection"]);
    data.insert("Credit", i18n("Data provided by NOAA National Weather Service"));

    setData(source, data);
}

QString NOAAMETARIon::country(const QString& source)
{
    Q_UNUSED(source);
    return QString("USA");
}

QString NOAAMETARIon::place(const QString& source)
{
    return d->m_weatherData[source].locationName;
}

QString NOAAMETARIon::station(const QString& source)
{
    return d->m_weatherData[source].stationID;
}

QString NOAAMETARIon::latitude(const QString& source)
{
    return d->m_weatherData[source].stationLat;
}

QString NOAAMETARIon::longitude(const QString& source)
{
    return d->m_weatherData[source].stationLon;
}

QTime NOAAMETARIon::observationTime(const QString& source)
{
    return d->m_weatherData[source].obsTime;
}

QString NOAAMETARIon::condition(const QString& source)
{
    if (d->m_weatherData[source].weather.isEmpty() || d->m_weatherData[source].weather == "NA") {
        d->m_weatherData[source].weather = "N/A";
    }
    return d->m_weatherData[source].weather;
}

float NOAAMETARIon::dewpoint(const QString& source)
{
    return d->m_weatherData[source].dewC;
}

QString NOAAMETARIon::humidity(const QString& source)
{
    if (d->m_weatherData[source].humidity == 0.0) {
        return QString("N/A");
    } else {
        return QString("%1%").arg(::round(d->m_weatherData[source].humidity));
    }
}

QMap<QString, QString> NOAAMETARIon::visibility(const QString& source)
{
    QMap<QString, QString> visibilityInfo;
    if (d->m_weatherData[source].visibility.isEmpty()) {
        visibilityInfo.insert("visibility", QString("N/A"));
        return visibilityInfo;
    }
    visibilityInfo.insert("visibility", d->m_weatherData[source].visibility);
    visibilityInfo.insert("visibilityUnit", QString::number(WeatherUtils::Kilometers));
    return visibilityInfo;
}

QMap<QString, QString> NOAAMETARIon::temperature(const QString& source)
{
    QMap<QString, QString> temperatureInfo;
    temperatureInfo.insert("temperature", QString::number(d->m_weatherData[source].tempC));
    temperatureInfo.insert("temperatureUnit", QString::number(WeatherUtils::Celsius));
    temperatureInfo.insert("comfortTemperature", "N/A");

    if (d->m_weatherData[source].humidexC != "N/A" && d->m_weatherData[source].windchillC == "N/A") {
        temperatureInfo.insert("comfortTemperature", d->m_weatherData[source].humidexC);
    }

    if (d->m_weatherData[source].windchillC != "N/A" && d->m_weatherData[source].humidexC == "N/A") {
        temperatureInfo.insert("comfortTemperature", d->m_weatherData[source].windchillC);
    }

    return temperatureInfo;
}

QMap<QString, QString> NOAAMETARIon::pressure(const QString& source)
{
    QMap<QString, QString> pressureInfo;
    if (d->m_weatherData[source].pressure.isEmpty()) {
        pressureInfo.insert("pressure", "N/A");
        return pressureInfo;
    }

    pressureInfo.insert("pressure", d->m_weatherData[source].pressure);
    pressureInfo.insert("pressureUnit", QString::number(WeatherUtils::Kilopascals));
    return pressureInfo;
}

QMap<QString, QString> NOAAMETARIon::wind(const QString& source)
{
    QMap<QString, QString> windInfo;

    // May not have any winds
    if (d->m_weatherData[source].windKMH == 0) {
        windInfo.insert("windSpeed", "Calm");
        windInfo.insert("windUnit", QString::number(WeatherUtils::NoUnit));
    } else {
        windInfo.insert("windSpeed", QString::number(d->m_weatherData[source].windKMH, 'f', 1));
        windInfo.insert("windUnit", QString::number(WeatherUtils::KilometersPerHour));
    }

    // May not always have gusty winds
    if (d->m_weatherData[source].windKMHGust == 0) {
        windInfo.insert("windGust", "N/A");
        windInfo.insert("windGustUnit", QString::number(WeatherUtils::NoUnit));
    } else {
        windInfo.insert("windGust", QString::number(d->m_weatherData[source].windKMHGust, 'f', 1));
        windInfo.insert("windGustUnit", QString::number(WeatherUtils::KilometersPerHour));
    }

    if (d->m_weatherData[source].windDirection.isEmpty()) {
        windInfo.insert("windDirection", "N/A");
    } else {
        windInfo.insert("windDirection", d->m_weatherData[source].windDirection.toLower());
    }
    return windInfo;
}

#include "ion_noaametar.moc"
