/***************************************************************************
 *   Copyright (C) 2009 by Shawn Starr <shawn.starr@rogers.com>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Ion for NOAA's National Weather Service METAR data */

#ifndef ION_NOAAMETAR_H
#define ION_NOAAMETAR_H

#include <QtCore/QStringList>
#include <QDebug>
#include <QDateTime>
#include <kurl.h>
#include <kio/job.h>
#include <kio/scheduler.h>
#include <kdemacros.h>
#include <plasma/dataengine.h>
#include <Plasma/Weather/Ion>
#include <Plasma/Weather/WeatherUtils>

#include <cmath>

class WeatherData
{

public:
    QString stationID;
    QString locationName;
    QString stateName;
    QString countryName;
    QString stationLat;
    QString stationLon;
    float windKMH;
    float windKMHGust;
    QString windDirection;
    float tempC;
    float dewC;
    int clouds;
    QString weather;
    QStringList coverList;
    QStringList currentList;
    QDate obsDate;
    QTime obsTime;
    QString pressure;
    QString temperature;
    QString dewPoint;
    QString visibility;
    float humidity;
    QString humidexC;
    QString windchillC;
    QString place;
    bool stationOOS;
    bool cloudsCondition;
};

class KDE_EXPORT NOAAMETARIon : public IonInterface
{
    Q_OBJECT

public:
    NOAAMETARIon(QObject *parent, const QVariantList &args);
    ~NOAAMETARIon();
    void init(void);  // Setup the city location, fetching the correct URL name.
    bool updateIonSource(const QString& source); // Sync data source with Applet
    void updateWeather(const QString& source);

    enum { FOG, SNOWSAND, SHOWER, THUNDERSTORM, ICE };

public slots:
    virtual void reset();

protected slots:
    void setup_slotDataArrived(KIO::Job *, const QByteArray &);
    void setup_slotJobFinished(KJob *);

    void slotDataArrived(KIO::Job *, const QByteArray &);
    void slotJobFinished(KJob *);

private:
    /* NOAA Methods - Internal for Ion */
    QMap<QString, ConditionIcons> setupConditionIconMappings(void);
    QMap<QString, ConditionIcons> const & conditionIcons(void);

    // Place information
    QString country(const QString& source);
    QString place(const QString& source);
    QString station(const QString& source);
    QString latitude(const QString& source);
    QString longitude(const QString& source);

    // Current Conditions Weather info
    QTime observationTime(const QString& source);
    QString condition(const QString& source);
    QMap<QString, QString> temperature(const QString& source);
    float dewpoint(const QString& source);
    QString humidity(const QString& source);
    QMap<QString, QString> visibility(const QString& source);
    QMap<QString, QString> pressure(const QString& source);
    QMap<QString, QString> wind(const QString& source);

    // Load and Parse the place METAR listing
    void getMetarSetup(void);
    bool readMetarSetup(void);
    void getMetarData(const QString& source);
    void parseMetarData(const QString& source, const QByteArray& metar);
    bool parseDate(WeatherData &data, const QString& token);
    bool parseTime(WeatherData &data, const QString& token);
    bool parseWindSpeed(WeatherData &data, const QString& token);
    QString parseWindDirection(const unsigned int direction);
    bool parseVisibility(WeatherData &data, QStringList::ConstIterator it);
    bool parseTemperature(WeatherData &data, const QString& token);
    bool parsePressure(WeatherData &data, const QString& token);
    bool parseCloudCover(WeatherData &data, const QString& token);
    bool parseCurrentCondition(WeatherData &data, const QString& token);
    bool parseTemperatureTenths(WeatherData &data, const QString& token);
    bool parseStationNeedsMaintenance(WeatherData &data, const QString& token);
    void calculateHumidity(WeatherData& weatherInfo);
    void calculateHumidex(WeatherData& weatherInfo);
    void calculateWindchill(WeatherData& weatherInfo);
  
    // Check if place specified is valid or not
    QStringList validate(const QString& source) const;

    inline float vaporPressure(float temp)
    {
         return (std::pow(10.0, 7.5 * temp / (237.7 + temp)));
    }

private:
    class Private;
    Private * d;
};

K_EXPORT_PLASMA_DATAENGINE(noaametar, NOAAMETARIon)

#endif
