/***************************************************************************
 *   Copyright (C) 2007 by Teemu Rytilahti <tpr@d5k.net>                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

#include <KUrl>
#include <KDebug>

#include <kio/job.h>
#include <plasma/weather/weatherutils.h>

#include "ion_willab.h"

WillabIon::WillabIon(QObject *parent, const QVariantList &args) : IonInterface(parent)
{
    Q_UNUSED(args);
}

WillabIon::~WillabIon()
{
    // delete m_transferjob and qxmlstreamreader?
}

void WillabIon::init()
{
    m_url = "http://weather.willab.fi/weather.xml";
    m_reader.clear();
    m_isFetching = false;
    setInitialized(true);
}

void WillabIon::reset()
{
    m_reader.clear();
    emit resetCompleted(this,true); 
}

void WillabIon::fetch()
{
    if(!m_isFetching) {
        m_isFetching = true;
        m_reader.clear();
        m_transferJob = KIO::get(m_url, KIO::Reload, KIO::HideProgressInfo);
        connect(m_transferJob, SIGNAL(data(KIO::Job *, const QByteArray &)), this, SLOT(dataFetched(KIO::Job *, const QByteArray &)));
        connect(m_transferJob, SIGNAL(result(KJob *)), this, SLOT(fetchingDone(KJob *)));
    }
}

bool WillabIon::updateIonSource(const QString& source)
{
    QStringList src = source.split("|");
    if(src[0] == "willab") {
        if(src[1] == "weather") {
            m_source = source;
            fetch();
            return true;
        }
        else if(src[1] == "validate") {
            /*if(src[2] == "Linnanmaa, Oulu, Finland") {
                setData(source, "validate", "willab|valid|single|place|Linnanmaa, Oulu, Finland");
                return true;
            }
            else
                setData(source, "validate", QString("willab|invalid|single|%1").arg(src[2]));*/
            setData(source, "validate", "willab|valid|single|place|Linnanmaa, Oulu, Finland");
        }
        else if(src[1] == "list") {
            setData(source, "validate", "willab|valid|single|place|Linnanmaa, Oulu, Finland");    
            return true;
        }
    }

    return false;
}

void WillabIon::updateData()
{
    if(m_isFetching == true)
        return;
    
    kDebug() << "updating data!";

    m_data["Place"]  = "Linnanmaa, Oulu";
    m_data["Credit"] = "Willab weather station";
    m_data["Credit Url"] = "http://weather.willab.fi";
    m_data["Condition Icon"] = "N/A";

    while (!m_reader.atEnd()) { 
        m_reader.name().toString();

        m_reader.readNext();
        if(m_reader.hasError()) {
            kDebug() << "error: " << m_reader.errorString() << "on " << QString("%1:%2").arg(m_reader.lineNumber(), m_reader.columnNumber());
            break;
        }

        if (m_reader.isEndElement()) {
            continue;
        }

        if (m_reader.isStartElement()) {
            if(m_reader.name() == "time")
                m_data["Observation Period"] = m_reader.readElementText();
            else if(m_reader.name() == "tempnow") {
                m_data["Temperature"] = m_reader.readElementText();
                m_data["Temperature Unit"] = WeatherUtils::Celsius;
            }
            else if(m_reader.name() == "temphi")
                m_data["Record High Temperature"] = m_reader.readElementText();
            else if(m_reader.name() == "templo")
                m_data["Record Low Temperature"] = m_reader.readElementText();
            else if(m_reader.name() == "dewpoint") {
                m_data["Dewpoint"] = m_reader.readElementText();
                m_data["Dewpoint Unit"] = WeatherUtils::Celsius; // TODO get the attribute
            }
            else if(m_reader.name() == "humidity")
                m_data["Humidity"] = m_reader.readElementText();
            else if(m_reader.name() == "airpressure") {
                m_data["Pressure Unit"] = WeatherUtils::Hectopascals; // TODO get the attribute
                m_data["Pressure"] = m_reader.readElementText();
            }
            else if(m_reader.name() == "windspeed") {
                m_data["Wind Speed Unit"] = WeatherUtils::MetersPerSecond; // TODO get the attribute
                m_data["Wind Speed"] = m_reader.readElementText();
            }
            else if(m_reader.name() == "windspeedmax") {
                m_data["Wind Gust Unit"] = WeatherUtils::MetersPerSecond;
                m_data["Wind Gust"] = m_reader.readElementText();
            }
            else if(m_reader.name() == "winddir")
                m_data["Wind Direction"] = WeatherUtils::degreesToCardinal(m_reader.readElementText().toFloat());
            else if(m_reader.name() == "solarrad")
                m_data["solarrad"] = m_reader.readElementText();
        }
    }
    //kDebug() << m_data;
    setData(m_source, m_data);
}

void WillabIon::dataFetched(KIO::Job *job, const QByteArray &data)
{
    Q_UNUSED(job);
    //kDebug() << "got data:" << data;
    m_reader.addData(data);
    //updateData();
}

void WillabIon::fetchingDone(KJob *job)
{
    if (job && job->error()) {
        kDebug() << "error in fetching: " << job->error() << job->errorString(); 
    }
    else
        kDebug() << "fetching done!";
    
    m_isFetching = false;
    updateData();
}

#include "ion_willab.moc"
