/////////////////////////////////////////////////////////////////////////
// storage.cpp                                                         //
//                                                                     //
// Copyright (C)  2010  Brian Pritchett <batenkaitos@gmail.com>        //
//                                                                     //
// This library is free software; you can redistribute it and/or       //
// modify it under the terms of the GNU Lesser General Public          //
// License as published by the Free Software Foundation; either        //
// version 2.1 of the License, or (at your option) any later version.  //
//                                                                     //
// This library is distributed in the hope that it will be useful,     //
// but WITHOUT ANY WARRANTY; without even the implied warranty of      //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU   //
// Lesser General Public License for more details.                     //
//                                                                     //
// You should have received a copy of the GNU Lesser General Public    //
// License along with this library; if not, write to the Free Software //
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA       //
// 02110-1301  USA                                                     //
/////////////////////////////////////////////////////////////////////////
#include "storage.h"
#include <kdebug.h>
#include "akonadi/item.h"
#include "akonadi/itemcreatejob.h"
#include "akonadi/itemfetchjob.h"
#include "akonadi/itemfetchscope.h"
#include "akonadi/collectionfetchjob.h"
#include "akonadi/collectioncreatejob.h"
#include "akonadi/collectionfetchscope.h"

//Storage Job implentation
StorageAkonadiJob::StorageAkonadiJob(const QString& destination,
                       const QString& operation,
                       const QMap<QString, QVariant>& parameters,
                       const Akonadi::Collection& dataEngineCollection,
                       QObject *parent)
            : ServiceJob(destination, operation, parameters, parent)
{
    m_dataEngineCollection = dataEngineCollection;
    Akonadi::CollectionFetchJob *findSourceCollection =
        new Akonadi::CollectionFetchJob(m_dataEngineCollection,
                Akonadi::CollectionFetchJob::FirstLevel);
    connect(findSourceCollection, SIGNAL( result(KJob*) ),
            this, SLOT(foundSourceCollection(KJob*) ) );
 }

void StorageAkonadiJob::store()
{
    Akonadi::Item item;
    item.setMimeType( "application/x-vnd.kde.storage" );
    item.setParentCollection(*m_sourceCollection);
    QVariant v(parameters()["data"]);
    QByteArray b;
    QDataStream ds(&b, QIODevice::WriteOnly);
    ds << v;
    QVariantHash vh;
    vh[QString("key")] = parameters()["key"];
    vh[QString("data")] = b.toBase64();
    item.setPayload<QVariantHash>(vh);
    Akonadi::ItemCreateJob* job = new Akonadi::ItemCreateJob(item, *m_sourceCollection);
    connect(job, SIGNAL(result(KJob*)), this, SLOT(jobFinished(KJob*)));
}

QVariantHash StorageAkonadiJob::retrieve()
{
    kWarning() << "retrieve";
    QVariantHash d;
    Akonadi::ItemFetchJob *job = new Akonadi::ItemFetchJob(*m_sourceCollection);

    job->fetchScope().fetchFullPayload(true);
    job->exec();
    Akonadi::Item::List itemList = job->items();

    foreach (Akonadi::Item i, itemList) {
        kWarning() << d;
        QVariantHash vh = i.payload<QVariantHash>();
        QByteArray b = QByteArray::fromBase64(vh["data"].toString().toAscii());
        QDataStream ds(&b, QIODevice::ReadOnly);
        QVariant v(ds);
        d[i.remoteId()] = v;
    }

    return d;
}


void StorageAkonadiJob::start()
{
    kWarning() << "start";
    QMap<QString, QVariant> params = parameters();
    if (operationName() == "save") {
      store();
      setResult(true);
      return;
    } else if (operationName() == "retrieve") {
        setResult(retrieve());
        return;
    }
    setError(1);
}

void StorageAkonadi::foundStorageCollection(KJob *job)
{
    Akonadi::CollectionFetchJob * cjob = static_cast < Akonadi::CollectionFetchJob * > ( job );
    foreach (const Akonadi::Collection & collection, cjob->collections()) {
        if (collection.name() == "Storage") {
            m_storageCollection = collection;
            break;
        }
    }

    Akonadi::CollectionFetchJob *deJob = 
        new Akonadi::CollectionFetchJob(m_storageCollection, Akonadi::CollectionFetchJob::FirstLevel);
    connect(deJob, SIGNAL( result( KJob* ) ), this, SLOT( foundDataEngineCollection ( KJob* ) ) );
}

void StorageAkonadi::foundDataEngineCollection(KJob* job)
{
    bool found = false;
    Akonadi::CollectionFetchJob * cjob = static_cast < Akonadi::CollectionFetchJob * > ( job );
    foreach (const Akonadi::Collection & collection, cjob->collections()) {
        if (collection.name() == destination()) {
            m_dataEngineCollection = collection;
            m_dataEngineCollection.setParentCollection(m_storageCollection);
            found = true;
            break;
        }
    }

    if (!found) { //DataEngineCollection needs to be created
        m_dataEngineCollection = Akonadi::Collection();
        m_dataEngineCollection.setParentCollection(m_storageCollection);
        m_dataEngineCollection.setName( destination() );
        QStringList mimeTypes;
        mimeTypes << Akonadi::Collection::mimeType()
                  << "application/x-vnd.kde.storage"
                  << "application/octet-stream";
        m_dataEngineCollection.setContentMimeTypes(mimeTypes);
        Akonadi::CollectionCreateJob* newDataEngine =
            new Akonadi::CollectionCreateJob(m_dataEngineCollection);
        connect(newDataEngine, SIGNAL( result(KJob*) ),
                this, SLOT(createDataEngineCollectionDone(KJob*) ) );
    }
}

void StorageAkonadiJob::foundSourceCollection(KJob* job)
{
    Akonadi::CollectionFetchJob * cjob = static_cast < Akonadi::CollectionFetchJob * > ( job );
    bool found = false;
    foreach (const Akonadi::Collection & collection, cjob->collections()) {
        if (collection.name() == parameters()["source"].toString()) {
            m_sourceCollection = new Akonadi::Collection(collection);
            m_sourceCollection->setParentCollection(m_dataEngineCollection);
            found = true;
            break;
        }
    }

    if (!found) { //Source Collection was not found, needs to be created
        Akonadi::Collection createSource;
        createSource.setName(parameters()["source"].toString());
        createSource.setParentCollection(m_dataEngineCollection);
        Akonadi::CollectionCreateJob *createSourceJob =
            new Akonadi::CollectionCreateJob(createSource);
        connect(createSourceJob, SIGNAL( result(KJob*) ),
            this, SLOT(createSourceCollectionDone(KJob*)));
    }
}

void StorageAkonadi::createDataEngineCollectionDone(KJob* job)
{
    Akonadi::CollectionCreateJob * cjob = static_cast < Akonadi::CollectionCreateJob * > ( job );
    m_dataEngineCollection =cjob->collection();
    Akonadi::Collection createSource;
    createSource.setName(m_serviceName);
    createSource.setParentCollection(cjob->collection());
}

void StorageAkonadiJob::createSourceCollectionDone(KJob* job)
{
    Akonadi::CollectionCreateJob * cjob = static_cast < Akonadi::CollectionCreateJob * > ( job );
    m_sourceCollection = new Akonadi::Collection(cjob->collection());
}


Plasma::ServiceJob* StorageAkonadi::createJob(const QString &operation, QMap<QString, QVariant> &parameters)
{
    return new StorageAkonadiJob(m_serviceName, operation, parameters, m_dataEngineCollection);
}

//Storage implementation
StorageAkonadi::StorageAkonadi(QObject* parent, const QVariantList& args) : Plasma::Service(parent)
{
    m_serviceName = args[0].toString();
    Akonadi::CollectionFetchJob *job =
        new Akonadi::CollectionFetchJob(Akonadi::Collection::root(),
                Akonadi::CollectionFetchJob::FirstLevel);
    connect(job, SIGNAL( result(KJob*) ), this, SLOT ( foundStorageCollection(KJob*) ) );
    setName("storage_akonadi");
}

StorageAkonadi::~StorageAkonadi()
{
}

#include "storage.moc"
