/*
 *   Copyright 2010 Ryan Rix <ry@n.rix.si>
 * 
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License as
 *   published by the Free Software Foundation; either version 2, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "activities_runner.h"

#include <QDBusConnection>
#include <QDBusConnectionInterface>
#include <QDBusInterface>
#include <QDBusServiceWatcher>

#include <KDebug>
#include <KIcon>

ActivitiesRunner::ActivitiesRunner(QObject *parent, const QVariantList& args)
    : Plasma::AbstractRunner(parent, args),
      m_controller(0),
      m_consumer(0)
{
    Q_UNUSED(args);
    setObjectName("org.kde.activities_runner");
    
    m_controller = new KActivityController( this );
    m_consumer = new KActivityConsumer( this );
    connect(m_controller, SIGNAL(activityAdded(QString)), this, SLOT(addActivity(QString)));
    connect(m_controller, SIGNAL(activityRemoved(QString)), this, SLOT(removeActivity(QString)));

    QStringList idList = m_consumer->availableActivities();
    
    foreach( QString id, idList )
    {
        addActivity(id);
        kDebug() << "Found activity:" << id;
    }
    
    QDBusServiceWatcher* watcher = new QDBusServiceWatcher("org.kde.ActivityManager", QDBusConnection::sessionBus(),
                                                           QDBusServiceWatcher::WatchForOwnerChange, this);
    connect(watcher, SIGNAL(serviceOwnerChanged(QString,QString,QString)),this, SLOT(checkAvailability(QString,QString,QString))); checkAvailability(QString(), QString(), QString());
}

ActivitiesRunner::~ActivitiesRunner()
{
}

void ActivitiesRunner::match(Plasma::RunnerContext &context)
{
    const QString term = context.query();
    if( m_enabled )
    {
        QString id = m_activityNames.key(term);
        if( id != "" )
        {
            KActivityInfo* info = m_activities.value( id );
            QString icon = info->icon();

            Plasma::QueryMatch match(this);
            match.setId("activities_runner");
            match.setType(Plasma::QueryMatch::ExactMatch);
            match.setIcon(KIcon(icon));
            match.setText(i18n("Sets the activity to %1").arg(term));
            match.setData(QVariant(term));
            match.setRelevance(1.0);
            context.addMatch(term, match);

            kDebug() << "Found match:" << term;
        }
        else
        {
            // Okay, so this is pretty failtastic... We're going to iterate through
            // the entire hash. Normally this won't be that bad, unless a user has,
            // like, over 9000 activities. Patches welcome. ;)
            
            // We're looking for partial matches in this block.
            QHashIterator<QString,QString> it( m_activityNames );
            while( it.hasNext() )
            {
                it.next();
                
                QString name = it.value();
                if( name.startsWith(term, Qt::CaseInsensitive) )
                {
                    // We have a match!
                    QString id = it.key();
                    KActivityInfo* info = m_activities.value( id );
                    
                    QString icon = info->icon();
                    
                    Plasma::QueryMatch match(this);
                    match.setId("activities_runner");
                    match.setType(Plasma::QueryMatch::PossibleMatch);
                    match.setIcon(KIcon(icon));
                    match.setText(i18n("Sets the activity to %1").arg(name));
                    match.setData(QVariant(name));
                    match.setRelevance(1.0);
                    context.addMatch(name, match);
                    
                    kDebug() << "Found Partial activity match:" << name << term;
                }
            }
        }
    }
}

void ActivitiesRunner::run(const Plasma::RunnerContext &context, const Plasma::QueryMatch &match)
{
    Q_UNUSED(context)
    
    if( m_enabled )
    {
        QString id = m_activityNames.key(match.data().toString() );
        
        if( id != "" )
        {
            kDebug() << "Setting current Activity to" << id << match.data().toString();
        
            m_controller->setCurrentActivity( id );
        }
        else
        {
            kDebug() << "Failed to find activity that matched the data... :(" << id << match.data().toString();
        }
    }
}

void ActivitiesRunner::checkAvailability(const QString &name, const QString &oldOwner, const QString &newOwner)
{
    Q_UNUSED(oldOwner)
    
    bool enabled = false;
    if (name.isEmpty()) {
        enabled = QDBusConnection::sessionBus().interface()->isServiceRegistered("org.kde.ActivityManager").value();
    } else {
        enabled = !newOwner.isEmpty();
    }

    if (m_enabled != enabled) {
        m_enabled = enabled;

        if (m_enabled) {
            addSyntax(Plasma::RunnerSyntax(i18nc("Note this is a KRunner keyword", ":q:"),
                                           i18n("Switches the workspace to the specified "
                                                "activity.")));
        } else {
            setSyntaxes(QList<Plasma::RunnerSyntax>());
        }
    }
}

void ActivitiesRunner::addActivity( QString id )
{
    KActivityInfo* info = new KActivityInfo(id);

    m_activities.insert( id, info );
    m_activityNames.insert( id, info->name() );

    connect(info, SIGNAL(nameChanged(QString)), this, SLOT(updateActivityName(QString)));

    kDebug() << "Added a new activity:" << id << info->name();
}

void ActivitiesRunner::removeActivity( QString id )
{
    disconnect(m_activities[id], SIGNAL(nameChanged(QString)), this, SLOT(updateActivityName(QString)));

    m_activities.remove( id );
    m_activityNames.remove( id );

    kDebug() << "Deleted a stale activity:" << id;
}

void ActivitiesRunner::updateActivityName(QString newName)
{
    KActivityInfo* info = qobject_cast<KActivityInfo*>(sender());
    if (info) {
        QString id = info->id();

        m_activityNames[id]=newName;

        kDebug() << "Renamed activity" << id << "to" << info->name();
    }
}

K_EXPORT_PLASMA_RUNNER(activities_runner, ActivitiesRunner)

#include "activities_runner.moc"
