/*
 *   Copyright 2008 Aike J Sommer <dev@aikesommer.name>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2,
 *   or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */


#include "screenwidget.h"

#include <QObject>
#include <QPainter>
#include <QGraphicsSceneMouseEvent>

#include <KDebug>


ScreenWidget::ScreenWidget(QGraphicsItem * parent)
    : QObject(),
    QGraphicsItem(parent),
    m_scale(0),
    m_resizing(false),
    m_moving(false),
    m_hovered(false)
{
    setAcceptHoverEvents(true);
    m_resizeHandle = new ResizeHandle(this);
    connect(m_resizeHandle, SIGNAL(dragged(QPointF)), this, SLOT(resizeHandleDragged(QPointF)));
    connect(m_resizeHandle, SIGNAL(released()), this, SLOT(resizeHandleReleased()));
    m_moveHandle = new ResizeHandle(this);
    connect(m_moveHandle, SIGNAL(dragged(QPointF)), this, SLOT(moveHandleDragged(QPointF)));
    connect(m_moveHandle, SIGNAL(released()), this, SLOT(moveHandleReleased()));
}

QRectF ScreenWidget::boundingRect() const {
    return QRectF(QPointF(0, 0), m_screenGeom.size() * m_scale).adjusted(0, 0, (m_resizing ? m_resized.x() : 0), (m_resizing ? m_resized.y() : 0)).united(m_drawnRect);
}

void ScreenWidget::paint(QPainter * p, const QStyleOptionGraphicsItem * o, QWidget * w) {
    Q_UNUSED(o);
    Q_UNUSED(w);
    
    //qreal scale = std::min(((qreal) m_contentsRect.width()) / m_totalSize.width(), ((qreal) m_contentsRect.height()) / m_totalSize.height());
    
    //QRectF rect(m_contentsRect.x() + m_geom.x() * scale, m_contentsRect.y() + m_geom.y() * scale, m_geom.width() * scale, m_geom.height() * scale);
    QSize screenSize = m_screenGeom.size();
    QPointF screenPos(0, 0);
    if (m_resizing) {
        screenSize = draggedToSize();
    } else if (m_moving) {
        screenPos = QPointF(draggedToPosition());
        screenPos *= m_scale;
    }
    
    QRectF rect(screenPos, screenSize * m_scale);
    
        //rect.adjust(0, 0, m_resized.x(), m_resized.y());
        //rect.setSize(draggedToSize() * m_scale);

    rect.adjust(2, 2, -2, -2);
    m_drawnRect = rect.adjusted(-1, -1, 1, 1);
    
    if (m_hovered) {
        m_resizeHandle->setVisible(true);
        m_moveHandle->setVisible(true);
        m_resizeHandle->setPos(rect.bottomRight() + QPointF(-12, -12));
        m_moveHandle->setPos(rect.topLeft() + (rect.bottomRight() - rect.topLeft()) / 2 + QPointF(-6, -6));
    } else {
        m_resizeHandle->setVisible(false);
        m_moveHandle->setVisible(false);
    }
    
    p->save();
    if (m_resizing) {
        p->setPen(Qt::white);
        
        foreach (const QSize& size, m_sizes) {
            QRectF sizeRect(QPointF(0, 0), size * m_scale);
            sizeRect.adjust(2, 2, -2, -2);
            p->drawRect(sizeRect);
        
            m_drawnRect |= sizeRect;
        }
    } else if (m_moving) {
        p->setPen(Qt::white);
        
        foreach (const QPoint& position, m_positions) {
            QPointF pos = QPointF(position - m_screenGeom.topLeft()) * m_scale;
            QRectF posRect(pos, m_screenGeom.size() * m_scale);
            posRect.adjust(2, 2, -2, -2);
            p->drawRect(posRect);
        
            m_drawnRect |= posRect;
        }
    }
    
    if (m_hovered) {
        p->setPen(Qt::red);
    } else {
        p->setPen(Qt::gray);
    }
    
    p->drawRect(rect);
    p->setOpacity(0.5);
    p->fillRect(rect, p->pen().brush());
    p->restore();
    
    p->save();
    p->setPen(Qt::white);
    QFont font = QFont(p->font());
    font.setBold(true);
    font.setPixelSize(14);
    p->setFont(font);
    QRectF textRect = rect.adjusted(5, 5, -5, -5);
    p->drawText(textRect,
            Qt::AlignHCenter | Qt::AlignTop,
            m_id);
    font.setBold(false);
    font.setPixelSize(11);
    p->setFont(font);
    textRect.adjust(0, 20, 0, 0);
    p->drawText(textRect,
            Qt::AlignHCenter | Qt::AlignTop,
            QString::number(screenSize.width()) + "x" + QString::number(screenSize.height()));
    p->restore();
    
    if (m_hovered) {
        
    }
}

void ScreenWidget::setScreenGeom(QRect geom) {
    m_screenGeom = geom;
}

void ScreenWidget::setId(const QString & id) {
    m_id = id;
}

QRect ScreenWidget::screenGeom() {
    return m_screenGeom;
}

void ScreenWidget::setAvailableSizes(QList<QSize> sizes) {
    m_sizes = sizes;
    /*foreach (QSize size, sizes) {
        kDebug() << "available size:" << m_screenGeom << size;
    }*/
}

void ScreenWidget::setAvailablePositions(QList<QPoint> positions) {
    m_positions = positions;
    /*foreach (QSize size, sizes) {
        kDebug() << "available size:" << m_screenGeom << size;
    }*/
}

QSize ScreenWidget::draggedToSize() {
    if (! m_resizing) {
        return m_screenGeom.size();
    }
    
    QPointF scaledDragged = m_resized / m_scale;
    QSize maxSize = m_screenGeom.size() + QSize(scaledDragged.x(), scaledDragged.y());
    
    //int max = 0;
    int min = -1;
    //QSize draggedTo(0, 0);
    QSize minSize(0, 0);
    foreach (const QSize& size, m_sizes) {
        if ((size.width() >= maxSize.width()) && (size.height() >= maxSize.height())) {
            if ((min < 0) || (size.width() * size.height() < min)) {
                min = size.width() * size.height();
                minSize = size;
            }
        }
    }
    
    //if (draggedTo.isEmpty()) {
        if (minSize.isEmpty()) {
            return m_screenGeom.size();
        } else {
            return minSize;
        }
    /*} else {
        return draggedTo;
    }*/
}

QPoint ScreenWidget::draggedToPosition() {
    if (! m_moving) {
        return QPoint(0, 0);
    }
    
    QPointF offset(m_scale * m_screenGeom.width() / 2, m_scale * m_screenGeom.height() / 2);
    QSizeF size(m_scale * m_screenGeom.width(), m_scale * m_screenGeom.height());
    //QPointF scaledDragged = m_moved / m_scale;
    QPointF draggedPos(m_moved + offset);
    
    foreach (const QPoint& position, m_positions) {
        QPointF pos = QPointF(position - m_screenGeom.topLeft()) * m_scale;
        QRectF posRect(pos, m_screenGeom.size() * m_scale);
        if (posRect.contains(draggedPos)) {
            return position - m_screenGeom.topLeft();
        }
    }
    
    return QPoint(0, 0);
}

void ScreenWidget::setScale(qreal scale) {
    if (m_scale != scale) {
        m_scale = scale;
        prepareGeometryChange();
    }
}

void ScreenWidget::resizeHandleDragged(QPointF offset) {
    m_resizing = true;
    m_resized = offset;
    update();
}

void ScreenWidget::resizeHandleReleased() {
    QSize size = draggedToSize();
    if (size != m_screenGeom.size()) {
        kDebug() << "requesting screen resize from" << m_screenGeom.size() << "to" << size;
        emit resizeRequested(size);
    }
    
    m_resizing = false;
    update();
}

void ScreenWidget::moveHandleDragged(QPointF offset) {
    m_moving = true;
    m_moved = offset;
    update();
}

void ScreenWidget::moveHandleReleased() {
    QPoint pos = draggedToPosition() + m_screenGeom.topLeft();
    if (pos != m_screenGeom.topLeft()) {
        kDebug() << "requesting screen move from" << m_screenGeom.topLeft() << "to" << pos;
        emit moveRequested(pos);
    }
    
    m_moving = false;
    update();
}

void ScreenWidget::hoverEnterEvent(QGraphicsSceneHoverEvent * event) {
    Q_UNUSED(event)
    
    m_hovered = true;
    setZValue(1);
    update();
}

void ScreenWidget::hoverLeaveEvent(QGraphicsSceneHoverEvent * event) {
    Q_UNUSED(event)
    
    m_hovered = false;
    setZValue(0);
    update();
}



ScreensWidget::ScreensWidget(QGraphicsItem * parent)
    : QGraphicsWidget(parent)
{
}
