/*
 * Copyright 2009 Petri Damstén <damu@iki.fi>
 * Copyright 2008 Aaron Seigo <aseigo@kde.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "view.h"
#include "backgrounddialog.h"
#include "appletbrowser.h"
#include <Plasma/Containment>
#include <Plasma/Theme>
#include <KDebug>
#include <QAction>
#include <typeinfo>

WindowView::WindowView(bool interactive, QWidget *parent)
    : Plasma::View(0, parent)
    , m_corona(0)
    , m_containment(0)
    , m_configDialog(0)
    , m_appletBrowser(0)
    , m_interactive(true)
{
    Plasma::Theme::defaultTheme()->setUseGlobalSettings(false);
    m_corona = new Plasma::Corona(this);
    m_corona->loadLayout();
    if (m_corona->containments().count() > 0) {
        m_containment = m_corona->containments()[0];
    } else {
        m_containment = m_corona->addContainment("desktop");
        m_containment->init();
        m_containment->setFormFactor(Plasma::Planar);
        m_containment->setWallpaper("image", "SingleImage");
        m_containment->setFormFactor(Plasma::Planar);
    }
    setContainment(m_containment);
    setScene(m_containment->scene());

    KConfigGroup containmentConfig = m_containment->config();
    KConfigGroup pluginsConfig(&containmentConfig, "ActionPlugins");
    KConfigGroup contextConfig(&pluginsConfig, "RightButton;NoModifier");
    contextConfig.writeEntry("_add panel", false);
    contextConfig.writeEntry("_lock_screen", false);
    contextConfig.writeEntry("_logout", false);
    contextConfig.writeEntry("_run_command", false);
    contextConfig.writeEntry("remove", false);
    m_containment->setContainmentActions("wheel:Vertical;NoModifier", 0);

    connect(m_containment, SIGNAL(showAddWidgetsInterface(QPointF)),
            this, SLOT(showAppletBrowser()));
    connect(m_containment, SIGNAL(configureRequested(Plasma::Containment*)),
            this, SLOT(configureContainment(Plasma::Containment*)));

    setFrameStyle(QFrame::NoFrame);
    setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    setInteractive(interactive);
}

WindowView::~WindowView()
{
    m_corona->saveLayout();
    delete m_configDialog;
    delete m_appletBrowser;
}

void WindowView::setInteractive(bool interactive)
{
    m_interactive = interactive;
    setToolboxVisible(interactive);
    setFocusPolicy(interactive ? Qt::StrongFocus : Qt::NoFocus);
    setContextMenuPolicy(interactive ? Qt::DefaultContextMenu : Qt::NoContextMenu);
    setCursor(interactive ? Qt::ArrowCursor : Qt::BlankCursor);
    m_containment->setContainmentActions("MidButton;NoModifier", interactive ? "paste" : 0);
    m_containment->setContainmentActions("RightButton;NoModifier", interactive ? "contextmenu" : 0);
    KConfigGroup coronaConfig(m_corona->config(), "General");
    m_corona->setImmutability(interactive ? Plasma::Mutable : Plasma::SystemImmutable);
    setEnabled(interactive ? true : false);
}

void WindowView::setToolboxVisible(bool visible)
{
    foreach (QGraphicsItem* item, m_containment->childItems()) {
        QString s = typeid(*item).name();
        if (s.contains("DesktopToolBox")) {
            item->setVisible(visible);
            return;
        }
    }
}

void WindowView::resizeEvent(QResizeEvent *event)
{
    QGraphicsView::resizeEvent(event);
    m_containment->resize(size());
    setSceneRect(m_containment->geometry());
}

void WindowView::configureContainment(Plasma::Containment *containment)
{
    const QString id = "plasma_containment_settings";

    if (m_configDialog) {
        m_configDialog->reloadConfig();
    } else {
        KConfigSkeleton *nullManager = new KConfigSkeleton(0);
        m_configDialog = new BackgroundDialog(size(), containment, this, this, id, nullManager);
        m_configDialog->setAttribute(Qt::WA_DeleteOnClose);
        connect(m_configDialog, SIGNAL(destroyed(QObject*)),
                this, SLOT(configDialogDestroyed()));
    }
    m_configDialog->show();
}

void WindowView::configDialogDestroyed()
{
    m_configDialog = 0;
}

void WindowView::showAppletBrowser()
{
    if (!m_appletBrowser) {
        m_appletBrowser = new Plasma::AppletBrowser(this);
        m_appletBrowser->setContainment(m_containment);
        m_appletBrowser->setApplication();
        m_appletBrowser->setAttribute(Qt::WA_DeleteOnClose);
        m_appletBrowser->setWindowTitle(i18n("Add Widgets"));
        m_appletBrowser->setWindowIcon(KIcon("plasmagik"));
        connect(m_appletBrowser, SIGNAL(destroyed(QObject*)),
                this, SLOT(appletBrowserDestroyed()));
    } else {
        m_appletBrowser->setContainment(m_containment);
    }
    m_appletBrowser->show();
}

void WindowView::appletBrowserDestroyed()
{
    m_appletBrowser = 0;
}

bool WindowView::event(QEvent *event)
{
    switch (event->type()) {
        case QEvent::ContextMenu:
        case QEvent::MouseButtonDblClick:
        case QEvent::MouseButtonPress:
        case QEvent::MouseButtonRelease:
        case QEvent::MouseMove:
        case QEvent::KeyPress:
        case QEvent::KeyRelease:
            if (!m_interactive) {
                event->setAccepted(true);
                return true;
            }
            break;

        default:
            break;
    }
    return Plasma::View::event(event);
}

#include "view.moc"
