/*
 *   Copyright 2009 Beat Wolf <asraniel@fryx.ch>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License as
 *   published by the Free Software Foundation; either version 2, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <KDebug>
#include <KConfigGroup>
#include <KTemporaryFile>
#include <QFile>
#include <QFileInfo>
#include <QPair>
#include <QHash>
#include "converter.h"

Converter::Converter(QUrl in, QUrl out)
{
    m_in = in;
    m_out = out;
}
Converter::~Converter()
{

}

void Converter::convert()
{
    QFile data(m_out.path());
    if (data.open(QFile::WriteOnly) ){
        m_outStream.setDevice(&data);
    }
    readPlasma();
}

void Converter::readPlasma()
{
    QFile inputFile(m_in.path());
    KTemporaryFile temp;
    temp.open();
    KSharedConfigPtr c = KSharedConfig::openConfig(temp.fileName());
    config = new Plasma::ConfigLoader(c, &inputFile,this);

    writeActions();
    delete config;
}

QString Converter::jolieType(QString plasmaType)
{
    bool array = false;
    bool unsupported = false;
    QString original = plasmaType;

    if(plasmaType == "QString" ||
        plasmaType == "KUrl") {
        plasmaType = "string";
    }

    if(plasmaType == "QStringList") {
        plasmaType = "string";
        array = true;
    }

    if(plasmaType.startsWith("QList<")) {
        plasmaType.remove("QList<");
        plasmaType.remove(">");
        array = true;
    }

    //HACK: this is wrong, but there are no unsigned types in jolie
    if(plasmaType == "uint" || plasmaType == "udouble") {
        plasmaType.remove(0,1);
        unsupported = true;
    }

    //HACK: this is wrong, but there is no longlong or bool type in jolie
    if(plasmaType.contains("long") || plasmaType == "bool") {
        plasmaType = "int";
        unsupported = true;
    }

    if(plasmaType == "QPoint") {
        plasmaType = "void {\n";
        plasmaType += "\t\t.x: int\n";
        plasmaType += "\t\t.y: int\n";
        plasmaType += "\t}";
    }

    if(plasmaType == "QSize") {
        plasmaType = "void {\n";
        plasmaType += "\t\t.width: int\n";
        plasmaType += "\t\t.height: int\n";
        plasmaType += "\t}";
    }

    if(plasmaType == "QRect") {
        plasmaType = "void {\n";
        plasmaType += "\t\t.topLeft: int\n";
        plasmaType += "\t\t.topRight: int\n";
        plasmaType += "\t\t.bottomLeft: int\n";
        plasmaType += "\t\t.bottomRight: int\n";
        plasmaType += "\t}";
    }

    //TODO: missing are: color, datetime and enum

    if(unsupported) {
        kWarning() << "Type "+original+" not supported by jolie"; 
    }

    QString type;
    if(array) {
       type = "[0,*]";
    }

    type += ": "+plasmaType;
    return type;
}

void Converter::writeActions()
{
    QList <KConfigSkeletonItem*> items = config->items();
    QHash < QString, QList < QPair < QString,QString > > > params;

    //read in all the parameters of all the operations
    foreach(KConfigSkeletonItem* item, items) {
        QPair<QString,QString> temp;
        temp.first = item->name();
        temp.second = jolieType(config->property(item->name()).typeName());
        params[item->group()].append(temp);
    }

    //create a "type" for each operation with its parameters
    foreach(QString group, params.keys()) {
        m_outStream << "type "<< group <<"Type:void {\n";
        QList < QPair < QString,QString > > list = params.value(group);
        for(int i = 0; i < list.size(); i++) {
            m_outStream << "\t." << list.at(i).first << list.at(i).second << "\n";
        }
        m_outStream << "}\n\n";
    }


    //write interface
    QFileInfo info(m_in.path());
    m_outStream << "interface " << info.baseName() << "Interface{\n";
    m_outStream << "RequestResponse:\n";
    const int lastIndex = config->groupList().size() - 1;
    for(int i = 0; i <= lastIndex; i++) {
        QString group = config->groupList().at(i);
        QString param = "void";
        //if an operation has parameters, use it's type (from before)
        if(params.contains(group)) {
            param = group+"Type";
        }

        m_outStream << "        " << group << "(" << param << ")(undefined)";
        
        //append the "," iff it is not the last operation
        if(i < lastIndex) {
            m_outStream << ",";
        }
        m_outStream << "\n";
    }
    m_outStream << "}\n";
}
