/*
 *   Copyright 2009 by Shafqat Bhuiyan <priomsrb@gmail.com>
 *   Copyright 2009 by Egor Tsinko <etsinko@gmail.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License as
 *   published by the Free Software Foundation; either version 2, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "scripted_image.h"

#include <KDebug>
#include <QByteArray>
#include <QFile>
#include <QPainter>
#include <QUrl>

#define MAX_OUTPUT_SIZE 1024

ScriptedImage::ScriptedImage(QObject *parent, const QVariantList &args)
    : Plasma::Wallpaper(parent, args)
    , wallpaperPath(QString())
{
    connect(&scriptProcess, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(setWallpaperFromScriptOutput()));
    connect(&scriptProcess, SIGNAL(error(QProcess::ProcessError)), this, SLOT(setWallpaperToFallbackImage()));
    connect(this, SIGNAL(renderCompleted(QImage)), this, SLOT(setWallpaper(QImage)));
    connect(&scriptTimer, SIGNAL(timeout()), this, SLOT(runScript()));
    scriptTimer.setSingleShot(true);
}

ScriptedImage::~ScriptedImage() {
    if(scriptProcess.state() == QProcess::Starting || scriptProcess.state() == QProcess::Running) {
        scriptProcess.close();
    }
}

void ScriptedImage::init(const KConfigGroup &config)
{
    if(scriptTimer.isActive())
        scriptTimer.stop();

    resizeMethod = (ResizeMethod)config.readEntry("wallpaperposition", (int)ScaledResize);
    scriptPath = config.readEntry("scriptpath", QString());
    secondsToRunScript = config.readEntry("secondstorunscript", float(30 * 60));
    fallbackImagePath = config.readEntry("fallbackimagepath", QString());

    setResizeMethodHint(resizeMethod);
    previousSize = boundingRect().size();

    loadFallbackImage();
    setWallpaperToFallbackImage();
    setupScript();
    runScript();
}

void ScriptedImage::loadFallbackImage() {
    bool loaded = fallbackImagePixmap.load(fallbackImagePath);
    if(!loaded) {
        setFallbackImageToBlack();
    }
}

void ScriptedImage::setFallbackImageToBlack() {
    fallbackImagePixmap = QPixmap(boundingRect().size().toSize());
    fallbackImagePixmap.fill(Qt::black);
}

void ScriptedImage::setWallpaperToFallbackImage() {
    wallpaperPixmap = fallbackImagePixmap.scaled(boundingRect().size().toSize());
    emit update(boundingRect());
}

void ScriptedImage::setupScript() {
    if(scriptProcess.state() == QProcess::Running) {
        scriptProcess.close();
    }
    scriptProcess.setProgram(scriptPath);
    QString scriptDirectory = QUrl().fromLocalFile(scriptPath).resolved(QUrl(".")).toLocalFile();
    scriptProcess.setWorkingDirectory(scriptDirectory);
    scriptProcess.setOutputChannelMode(KProcess::OnlyStdoutChannel);
}

void ScriptedImage::runScript() {
    if(scriptProcess.state() == QProcess::Running || scriptProcess.state() == QProcess::Starting) {
        // Call this function again after 1 second to see if script has finished
        int waitingTime = qMin(int(secondsToRunScript*1000), 1000);
        scriptTimer.start(waitingTime);
    } else if(scriptProcess.state() == QProcess::NotRunning) {
        scriptProcess.close();
        scriptProcess.start();
        scriptTimer.start(secondsToRunScript * 1000);
    }
}

void ScriptedImage::setWallpaperFromScriptOutput() {
    setWallpaperPathToScriptOutput();
    redrawWallpaper();
}

void ScriptedImage::redrawWallpaper() {
    if(QFile::exists(wallpaperPath)) {
        render(wallpaperPath, boundingRect().size().toSize(), resizeMethod, QColor(0, 0, 0));
    } else {
        setWallpaperToFallbackImage();
    }
}

void ScriptedImage::setWallpaper(const QImage &image) {
    wallpaperPixmap = QPixmap::fromImage(image);
    emit update(boundingRect());
}

void ScriptedImage::setWallpaperPathToScriptOutput() {
    wallpaperPath = scriptProcess.read(MAX_OUTPUT_SIZE);
    wallpaperPath = wallpaperPath.trimmed();

    if(!wallpaperPath.isEmpty()) {
        QUrl wallpaperUrl = QUrl().fromLocalFile(wallpaperPath);
        QUrl scriptUrl = QUrl().fromLocalFile(scriptPath);

        // Deal with relative wallpaper urls
        wallpaperPath = scriptUrl.resolved(wallpaperUrl).toLocalFile();
    }
}

void ScriptedImage::paint(QPainter *painter, const QRectF& exposedRect)
{
    // Check if geometry changed
    if (previousSize != boundingRect().size()) {
        redrawWallpaper();
        previousSize = boundingRect().size();
        return;
    }

    painter->drawPixmap(exposedRect, wallpaperPixmap, exposedRect.translated(-boundingRect().topLeft()));
}

QWidget* ScriptedImage::createConfigurationInterface(QWidget* parent)
{
    QWidget *widget = new QWidget(parent);
    ui.setupUi(widget);

    QTime timeForRefresh = QTime().addSecs(int(secondsToRunScript));

    ui.m_scriptPath->setText(scriptPath);
    ui.m_refreshInterval->setTime(timeForRefresh);
    ui.m_fallbackImagePath->setText(fallbackImagePath);
    
    // etsinko: Taken from Plasma/Wallpaper/Image.cpp
    ui.m_Positioning->addItem ( i18n ( "Scaled & Cropped" ), ScaledAndCroppedResize );
    ui.m_Positioning->addItem ( i18n ( "Scaled" ), ScaledResize );
    ui.m_Positioning->addItem ( i18n ( "Scaled, keep proportions" ), MaxpectResize );
    ui.m_Positioning->addItem ( i18n ( "Centered" ), CenteredResize );
    ui.m_Positioning->addItem ( i18n ( "Tiled" ), TiledResize );
    ui.m_Positioning->addItem ( i18n ( "Center Tiled" ), CenterTiledResize );
    for ( int i = 0; i <ui.m_Positioning->count(); ++i )
    {
        if ( resizeMethod == ui.m_Positioning->itemData ( i ).value<int>() )
        {
            ui.m_Positioning->setCurrentIndex ( i );
            break;
        }
    }

    connect(this, SIGNAL(settingsChanged(bool)), parent, SLOT(settingsChanged(bool)));

    connect(ui.m_scriptPath, SIGNAL(textChanged(QString)), this, SLOT(settingsModified()));
    connect(ui.m_refreshInterval, SIGNAL(timeChanged(QTime)), this, SLOT(settingsModified()));
    connect(ui.m_fallbackImagePath, SIGNAL(textChanged(QString)), this, SLOT(settingsModified()));
    connect ( ui.m_Positioning, SIGNAL ( currentIndexChanged ( int ) ), this, SLOT ( settingsModified() ) );

    return widget;
}

void ScriptedImage::settingsModified()
{
    bool modified = false;

    if ( resizeMethod != ( ResizeMethod ) ui.m_Positioning->itemData ( ui.m_Positioning->currentIndex() ).value<int>() )
    {
        resizeMethod = ( ResizeMethod ) ui.m_Positioning->itemData ( ui.m_Positioning->currentIndex() ).value<int>();
        setResizeMethodHint ( resizeMethod );
        
	// etsinko: Update image in the preview
	redrawWallpaper();
	modified = true;
    }
    
    if(ui.m_scriptPath->text() != scriptPath) {
        scriptPath = ui.m_scriptPath->text();
	
	// etsinko: Update image in the preview
	setupScript();
	runScript();
        modified = true;
    }

    int secondsToRunScriptInUi = qAbs(ui.m_refreshInterval->time().secsTo(QTime()));
    if(secondsToRunScriptInUi != secondsToRunScript) {
        secondsToRunScript = secondsToRunScriptInUi;
        if(secondsToRunScript == 0) {
            secondsToRunScript = 1;
        }
        modified = true;
    }

    if(ui.m_fallbackImagePath->text() != fallbackImagePath) {
        fallbackImagePath = ui.m_fallbackImagePath->text();
        modified = true;
    }

    if(modified) {
        emit settingsChanged(true);
        emit update(boundingRect());
    }
}

void ScriptedImage::save(KConfigGroup &config)
{
    config.writeEntry("wallpaperposition", (int)resizeMethod);
    config.writeEntry("scriptpath", scriptPath);
    config.writeEntry("secondstorunscript", secondsToRunScript);
    config.writeEntry("fallbackimagepath", fallbackImagePath);
}

#include "scripted_image.moc"

