/***************************************************************************
 *   Copyright (C) 2009 by Wang Hoi <zealot.hoi@gmail.com>                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

#include "koratbar.h"
#include "koratgroup.h"
#include "korattab.h"

#include <plasma/animator.h>
#include <plasma/framesvg.h>
#include <plasma/widgets/frame.h>
#include <plasma/widgets/tabbar.h>
#include <plasma/widgets/lineedit.h>
#include <plasma/widgets/iconwidget.h>
#include <plasma/widgets/toolbutton.h>

#include <KIcon>

#include <QGraphicsLinearLayout>
#include <QGraphicsGridLayout>
#include <QPainter>
#include <QColor>
#include <QSizePolicy>
#include <QLinearGradient>
#include <QToolButton>
#include <QMenu>
#include <QAction>

namespace Korat
{

class ToolBarPrivate
{
public:
    ToolBarPrivate(ToolBar *bar)
        : q(bar)
        , background(new Plasma::FrameSvg)
        , layout(new QGraphicsLinearLayout(Qt::Vertical))
        , upperLayout(new QGraphicsLinearLayout)
        , lowerLayout(new QGraphicsLinearLayout)
        , tabbar(new Plasma::TabBar)
        , filterOptionButton(new Plasma::ToolButton)
        , filterEdit(new Plasma::LineEdit)
        , newPageAnimId(-1)
        , oldPageAnimId(-1)
        , currentIndex(0)
        , newPage(0)
        , oldPage(0)
    {
    }
    
    void slidingCompleted(QGraphicsItem *item)
    {
        if (item == oldPage || item == newPage) {
            if (item == newPage) {
                lowerLayout->addItem(newPage);
                newPageAnimId = -1;
            } else {
                lowerLayout->removeItem(oldPage);
                oldPageAnimId = -1;
                item->hide();
            }
            q->setFlags(0);
        }
    }

    ToolBar *q;

    Plasma::FrameSvg *background;

    QGraphicsLinearLayout *layout;
    QGraphicsLinearLayout *upperLayout;
    QGraphicsLinearLayout *lowerLayout;

    Plasma::TabBar *tabbar;
    Plasma::ToolButton *filterOptionButton;
    Plasma::LineEdit *filterEdit;

    int newPageAnimId;
    int oldPageAnimId;

    int currentIndex;
    Tab *newPage;
    Tab *oldPage;

    QList<Tab *> tabList;
};

ToolBar::ToolBar(QGraphicsItem *parent)
    : QGraphicsWidget(parent)
    , d(new ToolBarPrivate(this))
{
    d->background->setImagePath("widgets/panel-background");
    d->background->setElementPrefix("north");

#if KDE_VERSION_MAJOR == 4 && KDE_VERSION_MINOR > 2
    d->filterOptionButton->setIcon(KIcon("view-filter"));
#endif
    d->filterEdit->setZValue(1000);

    d->upperLayout->setSpacing(0);
    d->upperLayout->setContentsMargins(0,0,0,0);
    d->lowerLayout->setSpacing(0);
    d->lowerLayout->setContentsMargins(0,0,0,0);

    d->upperLayout->addItem(d->tabbar);
    d->upperLayout->addItem(d->filterOptionButton);
    d->upperLayout->addItem(d->filterEdit);

    d->layout->setSpacing(0);
    d->layout->setContentsMargins(0,0,0,0);

    d->layout->addItem(d->upperLayout);
    d->layout->addItem(d->lowerLayout);

    setLayout(d->layout);

    connect(d->tabbar,SIGNAL(currentChanged(int)),this,SLOT(setCurrentTab(int)));
    connect(Plasma::Animator::self(),SIGNAL(movementFinished(QGraphicsItem *)),
            this,SLOT(slidingCompleted(QGraphicsItem *)));

    resize(600,200);
}

ToolBar::~ToolBar()
{
    delete d;
}

void ToolBar::addTab(Tab *tab)
{
    d->tabbar->addTab(tab->icon(),tab->title());
    if (!d->tabList.contains(tab)) {
        d->tabList << tab;
    }
    if (d->tabList.size() == 1) {
        d->lowerLayout->addItem(tab);
        tab->setVisible(true);
    } else {
        tab->setVisible(false);
    }
}

void ToolBar::removeTab(Tab *tab)
{
    int idx = d->tabList.indexOf(tab);
    if (idx != -1) {
        d->tabbar->removeTab(idx);
        d->tabList.removeAt(idx);
    }
}

Tab *ToolBar::tabAt(int index)
{
    return d->tabList.at(index);
}

void ToolBar::resizeEvent(QGraphicsSceneResizeEvent * event)
{
    QGraphicsWidget::resizeEvent(event);
    d->background->resizeFrame(event->newSize());
}

void ToolBar::paint(QPainter *painter,const QStyleOptionGraphicsItem *option, QWidget *widget)
{
    Q_UNUSED(option);
    Q_UNUSED(widget);

    QColor bg = QColor(Qt::blue);
    QLinearGradient linearGrad(boundingRect().topLeft(),boundingRect().bottomLeft());
    linearGrad.setColorAt(0, QColor(Qt::blue).lighter());
    linearGrad.setColorAt(0.2, Qt::blue);
    linearGrad.setColorAt(1, Qt::transparent);
    painter->fillRect(boundingRect(),linearGrad);
    d->background->paintFrame(painter);
}

void ToolBar::setCurrentTab(int index)
{
    d->oldPage = d->tabList.at(d->currentIndex);
    d->newPage = d->tabList.at(index);
    if (d->newPage) {
        setFlags(QGraphicsItem::ItemClipsChildrenToShape);

        if (d->newPageAnimId != -1 || d->oldPageAnimId != -1) {
            foreach (QGraphicsWidget *page, d->tabList) {
                page->hide();
            }
            if (d->newPageAnimId != -1) {
                Plasma::Animator::self()->stopItemMovement(d->newPageAnimId);
            }
            if (d->oldPageAnimId != -1) {
                Plasma::Animator::self()->stopItemMovement(d->oldPageAnimId);
            }
        }

//        d->newPage->setParentItem(this);
        d->newPage->show();
        if (d->oldPage) {
            QRect beforeCurrentGeom(d->oldPage->geometry().toRect());
            beforeCurrentGeom.moveTopRight(beforeCurrentGeom.topLeft());

            if (index > d->currentIndex) {
                d->newPage->setPos(d->oldPage->geometry().topRight());
                d->newPageAnimId = Plasma::Animator::self()->moveItem(
                    d->newPage,
                    Plasma::Animator::SlideOutMovement,
                    d->oldPage->pos().toPoint());
                d->oldPageAnimId = Plasma::Animator::self()->moveItem(
                    d->oldPage, Plasma::Animator::SlideOutMovement,
                    beforeCurrentGeom.topLeft());

            } else {

                d->newPage->setPos(beforeCurrentGeom.topLeft());
                d->newPageAnimId = Plasma::Animator::self()->moveItem(
                    d->newPage,
                    Plasma::Animator::SlideOutMovement,
                    d->oldPage->pos().toPoint());
                d->oldPageAnimId = Plasma::Animator::self()->moveItem(
                    d->oldPage, Plasma::Animator::SlideOutMovement,
                    d->oldPage->geometry().topRight().toPoint());
            }
//            d->layout->removeItem(layoutItem);
        }
//        d->newPage->show();
    }
    d->currentIndex = index;
}

}

#include "koratbar.moc"

// vim: sw=4 sts=4 et tw=100
