/*
    This file is part of the KDE project.
    Copyright (c) 2005 Friedrich W. H. Kossebau <Friedrich.W.H@Kossebau.de>
    Based upon kdmlib files.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License version 2 as published by the Free Software Foundation.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301  USA
*/


// c specific
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
// qt specific
// kde specific
#include <kdebug.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <dcopclient.h>
// applet specific
#include "klocalsessionmanager.h"
#include "klocalsessionmanager.moc"


enum KDMType { DMUnknown, NoDM, NewKDM, OldKDM/*, GDM*/ };

class KDMData
{
  public:
    KDMData();
    KDMType type()                             const { return DMType; }
    const char *const displayName()            const { return DisplayName; }
    const char *const displayManagerSettings() const { return DisplayManagerSettings; }
  protected:
    KDMType DMType;
    const char *DisplayName;
    const char *DisplayManagerSettings;
};


KDMData::KDMData()
{
  if( (DisplayName = ::getenv("DISPLAY")) == 0 )
    DMType = NoDM;
  else if( (DisplayManagerSettings = ::getenv( "DM_CONTROL" )) != 0 )
    DMType = NewKDM;
  else if( (DisplayManagerSettings = ::getenv( "XDM_MANAGED")) != 0  && DisplayManagerSettings[0] == '/' )
    DMType = OldKDM;
  else
    DMType = NoDM;
}

static KDMData DM;


class KDisplayManagerPipe
{
  public:
    KDisplayManagerPipe();
    ~KDisplayManagerPipe();
  public:
    /**
    * Send a command to the display manager
    * @param Command the command to execute.
    * @param Buffer the result buffer.
    * @return result:
    *  @li If true, the command was successfully executed.
    *   @p Buffer might contain addional results.
    *  @li If false and @p Buffer is empty, a communication error occurred
    *   (most probably KDM is not running).
    *  @li If false and @p Buffer is non-empty, it contains the error message
    *   from KDM.
    */
    bool send( const char *Command, QCString &Buffer );
    /**
     * Send a command to the display manager
     */
    bool send( const char *Command );

  protected:
    int Pipe;
};

KDisplayManagerPipe::KDisplayManagerPipe()
 : Pipe( -1 )
{
  // connect to display manager
  switch( DM.type() )
  {
    case NewKDM:
      // cannot create socket for connection?
      if( (Pipe = ::socket( PF_UNIX, SOCK_STREAM, 0 )) < 0 )
        break;
      {
        struct sockaddr_un Socket;
        Socket.sun_family = AF_UNIX;
        // get end of display name
        char *N;
        const char* const DisplayName = DM.displayName();
        if( (N = ::strchr( DisplayName, ':' )) != 0 )
          N = ::strchr( N, '.' );
        ::snprintf( Socket.sun_path, sizeof(Socket.sun_path),
                  "%s/dmctl-%.*s/socket", DM.displayManagerSettings(), N>0 ? N - DisplayName : 512, DisplayName );
        // cannot create connection?
        if( ::connect(Pipe,(struct sockaddr *)&Socket,sizeof( Socket )) != 0 )
        {
          ::close( Pipe );
          Pipe = -1;
        }
      }
      break;
    case OldKDM:
      {
        QString PipeName( DM.displayManagerSettings() );
        PipeName.truncate( PipeName.find( ',' ) );
        Pipe = ::open( PipeName.latin1(), O_WRONLY );
      }
      break;
    default:;
  }
}


bool KDisplayManagerPipe::send( const char *Command )
{
  QCString Buffer;
  return send( Command, Buffer );
}

bool KDisplayManagerPipe::send( const char *Command, QCString &Buffer )
{
  bool Result = false;

  int tl;

  if( Pipe < 0 )
    goto busted;

  tl = ::strlen( Command );
  if( ::write(Pipe,Command,tl) != tl )
  {
    bust:
      ::close( Pipe );
      Pipe = -1;
    busted:
      Buffer.resize( 0 );
      return false;
  }
  if( DM.type() == OldKDM )
  {
    Buffer.resize( 0 );
    return true;
  }

  unsigned Length = 0;
  for( ;; )
  {
    if( Buffer.size() < 128 )
      Buffer.resize( 128 );
    else if( Buffer.size() < Length * 2 )
      Buffer.resize( Length * 2 );
    if( (tl = ::read( Pipe, Buffer.data()+Length, Buffer.size()-Length)) <= 0 )
    {
      if( tl < 0 && errno == EINTR )
        continue;
      goto bust;
    }
    Length += tl;
    if( Buffer[Length-1] == '\n' )
    {
      Buffer[Length-1] = 0;
      if( Length > 2 && Buffer[0] == 'o' && Buffer[1] == 'k' && Buffer[2] < 32 )
        Result = true;
      break;
    }
  }
  return Result;
}


KDisplayManagerPipe::~KDisplayManagerPipe()
{
  if( Pipe >= 0 )
    close( Pipe );
}


KLocalSessionManager::KLocalSessionManager( QWidget *S, int SN )
 : Screen( S ), ScreenNumber( SN ), LockOnSwitch( false )
{
}


void KLocalSessionManager::setLockOnSwitch( bool Lock )
{
  LockOnSwitch = Lock;
}


bool KLocalSessionManager::doesLockOnSwitch() const { return LockOnSwitch; }

bool KLocalSessionManager::canCreateSession() const { return kapp->authorize( "start_new_session" ); }
bool KLocalSessionManager::canLogoutSession() const { return kapp->authorize( "logout" ); }
bool KLocalSessionManager::canLockSession()   const { return kapp->authorize( "lock_screen" ); }

bool KLocalSessionManager::canSaveSession()   const
{
  KConfig SMServerConfig( "ksmserverrc", false, false );
  SMServerConfig.setGroup( "General" );
  return SMServerConfig.readEntry("loginMode") == "restoreSavedSession";
}

bool KLocalSessionManager::canShutdown() const
{
  if( DM.type() == OldKDM )
    return strstr( DM.displayManagerSettings(), ",maysd" ) != 0;

  QCString Response;
  return KDisplayManagerPipe().send( "caps\n", Response ) && Response.find( "\tshutdown" ) >= 0;
}

bool KLocalSessionManager::canSwitchSession() const
{
  if( DM.type() == OldKDM )
    return DM.displayName()[0] == ':';

  QCString Response;
  return KDisplayManagerPipe().send( "caps\n", Response ) && Response.find( "\tlocal" ) >= 0;
}


void KLocalSessionManager::createSession() const
{
  if( !canCreateSession() )
    return;

  int Result = KMessageBox::warningContinueCancel(
      Screen,
      i18n("<p>You have chosen to open another desktop session.<br>"
              "The Current session will be hidden "
              "and a new login screen will be displayed.<br>"
              "An F-key is assigned to each session; "
              "F%1 is usually assigned to the first session, "
              "F%2 to the second session and so on. "
              "You can switch between sessions by pressing "
              "Ctrl, Alt and the appropriate F-key at the same time. "
              "Additionally, the KDE Panel and Desktop menus have "
              "actions for switching between sessions.</p>")
                          .arg(7).arg(8),
      i18n("Warning - New Session"),
      KGuiItem(i18n("&Start New Session"), "fork"),
      ":confirmNewSession",
#if KDE_IS_VERSION(3,4,0)
      KMessageBox::PlainCaption|
#endif
KMessageBox::Notify );

  if( Result == KMessageBox::Cancel )
    return;

  if( LockOnSwitch )
    lockSession();

  KDisplayManagerPipe().send( "reserve\n" );
}


void KLocalSessionManager::switchToSession( int SessionId ) const
{
  bool Switched = KDisplayManagerPipe().send( QString("activate\tvt%1\n").arg(SessionId).latin1() );
  if( Switched && LockOnSwitch )
    lockSession();
}


void KLocalSessionManager::saveSession() const
{
  if( canSaveSession() )
    kapp->dcopClient()->send( "SMServerConfig", "default",
                              "saveCurrentSession()", QByteArray() );
}

void KLocalSessionManager::lockSession() const
{
  if( !canLockSession() )
    return;

  QCString DesktopAppName( "kdesktop" );
  if( ScreenNumber != 0 )
    DesktopAppName.sprintf( "kdesktop-screen-%d", ScreenNumber );
  kapp->dcopClient()->send( DesktopAppName, "KScreensaverIface", "lock()", QByteArray() );

//TODO: what is the difference to:
//  KDisplayManagerPipe().send( "lock\n" );
}


void KLocalSessionManager::logoutSession() const
{
  if( canLogoutSession() )
    kapp->requestShutDown();
}


void KLocalSessionManager::shutdown( KApplication::ShutdownType ShutdownType,
              KApplication::ShutdownMode ShutdownMode,
              const QString &BootOption )
{
  if( canShutdown() && ShutdownType != KApplication::ShutdownTypeNone )
  {
    QCString Command( "shutdown\t" );
    Command.append( ShutdownType == KApplication::ShutdownTypeReboot ? "reboot\t" : "halt\t" );
    if( !BootOption.isEmpty() && DM.type() == NewKDM )
      Command.append( "=" ).append( BootOption.local8Bit() ).append( "\t" );
    Command.append( ShutdownMode == KApplication::ShutdownModeInteractive ? "ask\n" :
                ShutdownMode == KApplication::ShutdownModeForceNow ?        "forcenow\n" :
                ShutdownMode == KApplication::ShutdownModeTryNow ?          "trynow\n" :
                "schedule\n" );
    KDisplayManagerPipe().send( Command.data() );
  }
}

bool KLocalSessionManager::getBootOptions( QStringList &Options, int &DefaultOption, int &CurrentOption ) const
{
  if( DM.type() != NewKDM )
    return false;

  bool Success = false;
  QCString Response;
  if( KDisplayManagerPipe().send("listbootoptions\n",Response) )
  {
    Options = QStringList::split( '\t', QString::fromLocal8Bit(Response.data()) );
    if( Options.size() >= 4 )
    {
      bool Ok;
      DefaultOption = Options[2].toInt( &Ok );
      if( Ok )
      {
        CurrentOption = Options[3].toInt( &Ok );
        if( Ok )
        {
          Options = QStringList::split( ' ', Options[1] );
          for( QStringList::Iterator it = Options.begin(); it != Options.end(); ++it )
            (*it).replace( "\\s", " " );
          Success = true;
        }
      }
    }
  }

  return Success;
}


int KLocalSessionManager::numberOfReservedSessions() const
{
  if( DM.type() == OldKDM )
    return strstr( DM.displayManagerSettings(), ",rsvd" ) ? 1 : -1;

  QCString Response;
  int p = 0;
  if( !(KDisplayManagerPipe().send( "caps\n", Response ) && (p = Response.find( "\treserve " )) >= 0) )
          return -1;
  return atoi( Response.data() + p + 9 );
}

KSessionList KLocalSessionManager::localSessions()
{
  KSessionList List;
  getLocalSessions( List );
  return List;
}

bool KLocalSessionManager::getLocalSessions( KSessionList &List )
{
  if( DM.type() == OldKDM )
    return false;

  bool Success = false;
  QCString Response;
  if( KDisplayManagerPipe().send("list\talllocal\n", Response) )
  {
    const QStringList Sessions = QStringList::split( QChar('\t'), Response.data()+3 );
    for( QStringList::ConstIterator it = Sessions.begin(); it != Sessions.end(); ++it )
    {
      const QStringList SessionDetails = QStringList::split( QChar(','), *it, true );
      KSession Session;
      Session.DisplayName = SessionDetails[0];
      // remote?
      if( SessionDetails[1][0] == '@' )
      {
        Session.Origin = SessionDetails[1].mid( 1 );
        Session.VTId = 0;
      }
      else
        Session.VTId = SessionDetails[1].mid( 2 ).toInt();
      Session.LoginName = SessionDetails[2];
      Session.Session = SessionDetails[3];
      Session.Current = ( SessionDetails[4].find('*') >= 0 );
      Session.IsTty = ( SessionDetails[4].find('t') >= 0 );
      List.append( Session );
    }
    Success = true;
  }
  return Success;
}
